
module Make

# 42 "asllib/Parser.mly"
          (Config : ParserConfig.CONFIG)
# 7 "asllib/Parser.ml"

= struct
  
  module MenhirBasics = struct
    
    exception Error
    
    let _eRR =
      fun _s ->
        raise Error
    
    type token = Tokens.token
    
  end
  
  include MenhirBasics
  
# 50 "asllib/Parser.mly"
  

open AST
open ASTUtils
open Desugar

let version = V1

let t_bit =
  T_Bits (E_Literal (L_Int Z.one) |> add_dummy_annotation ~version, [])

let zero = E_Literal (L_Int Z.zero) |> add_dummy_annotation ~version

let make_ldi_vars (xs, ty) =
  let make_one x =
    S_Decl (LDK_Var, LDI_Var x, Some ty, None)
    |> add_dummy_annotation ~version
  in
  List.map make_one xs |> stmt_from_list |> desc

let make_ty_decl_subtype (x, s) =
  let name, _fields = s.desc in
  let ty = ASTUtils.add_pos_from s (T_Named name) in
  D_TypeDecl (x, ty, Some s.desc)

let prec =
  let open AST in
  function
  | BOR | BAND | IMPL | BEQ -> 1
  | EQ_OP | NEQ -> 2
  | PLUS | MINUS | OR | EOR | AND -> 3
  | MUL | DIV | DIVRM | RDIV | MOD | SHL | SHR -> 4
  | POW -> 5
  | _ -> 0 (* Non assoc *)

let check_not_same_prec loc op op' =
  if prec op = prec op' then Error.(fatal_from loc CannotParse)

let check_not_binop_same_prec op e =
  match e.desc with
  | E_Binop (op', _, _) when op != op' -> check_not_same_prec e op op'
  | _ -> ()

let e_binop (e1, op, e2) =
  let () = check_not_binop_same_prec op e1
  and () = check_not_binop_same_prec op e2
  in
  E_Binop (op, e1, e2)

let e_call call = E_Call { call with call_type = ST_Function }
let s_call call = S_Call { call with call_type = ST_Procedure }

let le_var x = LE_Var x.desc |> add_pos_from x
let lhs_access ~base ~index ~fields = { base; index; fields; slices=add_dummy_annotation ~version [] }


# 82 "asllib/Parser.ml"
  
  type ('s, 'r) _menhir_state = 
    | MenhirState000 : ('s, _menhir_box_opn) _menhir_state
      (** State 000.
        Stack shape : .
        Start symbol: opn. *)

    | MenhirState001 : (('s, 'r) _menhir_cell1_WHILE, 'r) _menhir_state
      (** State 001.
        Stack shape : WHILE.
        Start symbol: <undetermined>. *)

    | MenhirState004 : (('s, 'r) _menhir_cell1_NOT, 'r) _menhir_state
      (** State 004.
        Stack shape : NOT.
        Start symbol: <undetermined>. *)

    | MenhirState005 : (('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 005.
        Stack shape : MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState006 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 006.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState008 : (('s, 'r) _menhir_cell1_IF, 'r) _menhir_state
      (** State 008.
        Stack shape : IF.
        Start symbol: <undetermined>. *)

    | MenhirState009 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 009.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState010 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 010.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState012 : (('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_state
      (** State 012.
        Stack shape : BNOT.
        Start symbol: <undetermined>. *)

    | MenhirState015 : (('s, 'r) _menhir_cell1_ARBITRARY, 'r) _menhir_state
      (** State 015.
        Stack shape : ARBITRARY.
        Start symbol: <undetermined>. *)

    | MenhirState018 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 018.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState019 : (('s, 'r) _menhir_cell1_INTEGER, 'r) _menhir_state
      (** State 019.
        Stack shape : INTEGER.
        Start symbol: <undetermined>. *)

    | MenhirState020 : (('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 020.
        Stack shape : LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState021 : ((('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 021.
        Stack shape : LBRACE MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState025 : ((('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 025.
        Stack shape : MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState026 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_state
      (** State 026.
        Stack shape : expr LLBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState027 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 027.
        Stack shape : expr LLBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState028 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHR, 'r) _menhir_state
      (** State 028.
        Stack shape : expr SHR.
        Start symbol: <undetermined>. *)

    | MenhirState029 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 029.
        Stack shape : expr SHR expr.
        Start symbol: <undetermined>. *)

    | MenhirState030 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_POW, 'r) _menhir_state
      (** State 030.
        Stack shape : expr POW.
        Start symbol: <undetermined>. *)

    | MenhirState031 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_POW, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 031.
        Stack shape : expr POW expr.
        Start symbol: <undetermined>. *)

    | MenhirState032 : (('s, 'r) _menhir_cell1_LBRACKET, 'r) _menhir_state
      (** State 032.
        Stack shape : LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState033 : (('s, 'r) _menhir_cell1_COLON, 'r) _menhir_state
      (** State 033.
        Stack shape : COLON.
        Start symbol: <undetermined>. *)

    | MenhirState034 : ((('s, 'r) _menhir_cell1_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 034.
        Stack shape : COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState035 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHL, 'r) _menhir_state
      (** State 035.
        Stack shape : expr SHL.
        Start symbol: <undetermined>. *)

    | MenhirState036 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 036.
        Stack shape : expr SHL expr.
        Start symbol: <undetermined>. *)

    | MenhirState037 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_IN, 'r) _menhir_state
      (** State 037.
        Stack shape : expr IN.
        Start symbol: <undetermined>. *)

    | MenhirState038 : (('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 038.
        Stack shape : LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState039 : (('s, 'r) _menhir_cell1_NOT, 'r) _menhir_state
      (** State 039.
        Stack shape : NOT.
        Start symbol: <undetermined>. *)

    | MenhirState040 : ((('s, 'r) _menhir_cell1_NOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 040.
        Stack shape : NOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState042 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DOT _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 042.
        Stack shape : expr DOT LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState044 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 044.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState051 : (('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 051.
        Stack shape : MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState052 : ((('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 052.
        Stack shape : MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState054 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 054.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState055 : (('s, 'r) _menhir_cell1_LEQ, 'r) _menhir_state
      (** State 055.
        Stack shape : LEQ.
        Start symbol: <undetermined>. *)

    | MenhirState056 : ((('s, 'r) _menhir_cell1_LEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 056.
        Stack shape : LEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState057 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RDIV, 'r) _menhir_state
      (** State 057.
        Stack shape : expr RDIV.
        Start symbol: <undetermined>. *)

    | MenhirState058 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RDIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 058.
        Stack shape : expr RDIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState059 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_CONCAT, 'r) _menhir_state
      (** State 059.
        Stack shape : expr CONCAT.
        Start symbol: <undetermined>. *)

    | MenhirState060 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_CONCAT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 060.
        Stack shape : expr CONCAT expr.
        Start symbol: <undetermined>. *)

    | MenhirState061 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS, 'r) _menhir_state
      (** State 061.
        Stack shape : expr PLUS.
        Start symbol: <undetermined>. *)

    | MenhirState062 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 062.
        Stack shape : expr PLUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState063 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MUL, 'r) _menhir_state
      (** State 063.
        Stack shape : expr MUL.
        Start symbol: <undetermined>. *)

    | MenhirState064 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MUL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 064.
        Stack shape : expr MUL expr.
        Start symbol: <undetermined>. *)

    | MenhirState065 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MOD, 'r) _menhir_state
      (** State 065.
        Stack shape : expr MOD.
        Start symbol: <undetermined>. *)

    | MenhirState066 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MOD, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 066.
        Stack shape : expr MOD expr.
        Start symbol: <undetermined>. *)

    | MenhirState067 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIVRM, 'r) _menhir_state
      (** State 067.
        Stack shape : expr DIVRM.
        Start symbol: <undetermined>. *)

    | MenhirState068 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIVRM, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 068.
        Stack shape : expr DIVRM expr.
        Start symbol: <undetermined>. *)

    | MenhirState069 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIV, 'r) _menhir_state
      (** State 069.
        Stack shape : expr DIV.
        Start symbol: <undetermined>. *)

    | MenhirState070 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 070.
        Stack shape : expr DIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState071 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OR, 'r) _menhir_state
      (** State 071.
        Stack shape : expr OR.
        Start symbol: <undetermined>. *)

    | MenhirState072 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 072.
        Stack shape : expr OR expr.
        Start symbol: <undetermined>. *)

    | MenhirState073 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_NEQ, 'r) _menhir_state
      (** State 073.
        Stack shape : expr NEQ.
        Start symbol: <undetermined>. *)

    | MenhirState075 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_NEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 075.
        Stack shape : expr NEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState076 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 076.
        Stack shape : expr MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState077 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 077.
        Stack shape : expr MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState078 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LT, 'r) _menhir_state
      (** State 078.
        Stack shape : expr LT.
        Start symbol: <undetermined>. *)

    | MenhirState079 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 079.
        Stack shape : expr LT expr.
        Start symbol: <undetermined>. *)

    | MenhirState080 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_EOR, 'r) _menhir_state
      (** State 080.
        Stack shape : expr EOR.
        Start symbol: <undetermined>. *)

    | MenhirState081 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_EOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 081.
        Stack shape : expr EOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState082 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_state
      (** State 082.
        Stack shape : expr COLON_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState083 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 083.
        Stack shape : expr COLON_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState084 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_AND, 'r) _menhir_state
      (** State 084.
        Stack shape : expr AND.
        Start symbol: <undetermined>. *)

    | MenhirState085 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_AND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 085.
        Stack shape : expr AND expr.
        Start symbol: <undetermined>. *)

    | MenhirState086 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LEQ, 'r) _menhir_state
      (** State 086.
        Stack shape : expr LEQ.
        Start symbol: <undetermined>. *)

    | MenhirState087 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 087.
        Stack shape : expr LEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState088 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GT, 'r) _menhir_state
      (** State 088.
        Stack shape : expr GT.
        Start symbol: <undetermined>. *)

    | MenhirState089 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 089.
        Stack shape : expr GT expr.
        Start symbol: <undetermined>. *)

    | MenhirState090 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GEQ, 'r) _menhir_state
      (** State 090.
        Stack shape : expr GEQ.
        Start symbol: <undetermined>. *)

    | MenhirState091 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 091.
        Stack shape : expr GEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState092 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_IMPL, 'r) _menhir_state
      (** State 092.
        Stack shape : expr IMPL.
        Start symbol: <undetermined>. *)

    | MenhirState093 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_IMPL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 093.
        Stack shape : expr IMPL expr.
        Start symbol: <undetermined>. *)

    | MenhirState094 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_state
      (** State 094.
        Stack shape : expr EQ_OP.
        Start symbol: <undetermined>. *)

    | MenhirState096 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 096.
        Stack shape : expr EQ_OP expr.
        Start symbol: <undetermined>. *)

    | MenhirState097 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BOR, 'r) _menhir_state
      (** State 097.
        Stack shape : expr BOR.
        Start symbol: <undetermined>. *)

    | MenhirState098 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 098.
        Stack shape : expr BOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState099 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BEQ, 'r) _menhir_state
      (** State 099.
        Stack shape : expr BEQ.
        Start symbol: <undetermined>. *)

    | MenhirState100 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 100.
        Stack shape : expr BEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState101 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BAND, 'r) _menhir_state
      (** State 101.
        Stack shape : expr BAND.
        Start symbol: <undetermined>. *)

    | MenhirState102 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BAND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 102.
        Stack shape : expr BAND expr.
        Start symbol: <undetermined>. *)

    | MenhirState103 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_AS, 'r) _menhir_state
      (** State 103.
        Stack shape : expr AS.
        Start symbol: <undetermined>. *)

    | MenhirState107 : (('s, 'r) _menhir_cell1_BITS _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 107.
        Stack shape : BITS LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState108 : ((('s, 'r) _menhir_cell1_BITS _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 108.
        Stack shape : BITS LPAR expr.
        Start symbol: <undetermined>. *)

    | MenhirState109 : (((('s, 'r) _menhir_cell1_BITS _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR, 'r) _menhir_state
      (** State 109.
        Stack shape : BITS LPAR expr RPAR.
        Start symbol: <undetermined>. *)

    | MenhirState110 : (('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 110.
        Stack shape : LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState114 : (('s, 'r) _menhir_cell1_slices _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 114.
        Stack shape : slices IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState115 : (('s, 'r) _menhir_cell1_COLON, 'r) _menhir_state
      (** State 115.
        Stack shape : COLON.
        Start symbol: <undetermined>. *)

    | MenhirState118 : (('s, 'r) _menhir_cell1_ARRAY, 'r) _menhir_state
      (** State 118.
        Stack shape : ARRAY.
        Start symbol: <undetermined>. *)

    | MenhirState119 : ((('s, 'r) _menhir_cell1_ARRAY, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 119.
        Stack shape : ARRAY expr.
        Start symbol: <undetermined>. *)

    | MenhirState121 : (((('s, 'r) _menhir_cell1_ARRAY, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RRBRACKET, 'r) _menhir_state
      (** State 121.
        Stack shape : ARRAY expr RRBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState129 : (('s, 'r) _menhir_cell1_bitfield, 'r) _menhir_state
      (** State 129.
        Stack shape : bitfield.
        Start symbol: <undetermined>. *)

    | MenhirState135 : (('s, 'r) _menhir_cell1_IF, 'r) _menhir_state
      (** State 135.
        Stack shape : IF.
        Start symbol: <undetermined>. *)

    | MenhirState136 : ((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 136.
        Stack shape : IF expr.
        Start symbol: <undetermined>. *)

    | MenhirState137 : (((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 137.
        Stack shape : IF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState138 : ((((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 138.
        Stack shape : IF expr THEN expr.
        Start symbol: <undetermined>. *)

    | MenhirState139 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_state
      (** State 139.
        Stack shape : expr THEN expr ELSIF.
        Start symbol: <undetermined>. *)

    | MenhirState140 : (((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 140.
        Stack shape : expr THEN expr ELSIF expr.
        Start symbol: <undetermined>. *)

    | MenhirState141 : ((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 141.
        Stack shape : expr THEN expr ELSIF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState142 : (((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 142.
        Stack shape : expr THEN expr ELSIF expr THEN expr.
        Start symbol: <undetermined>. *)

    | MenhirState143 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSE, 'r) _menhir_state
      (** State 143.
        Stack shape : expr THEN expr ELSE.
        Start symbol: <undetermined>. *)

    | MenhirState144 : (((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSE, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 144.
        Stack shape : expr THEN expr ELSE expr.
        Start symbol: <undetermined>. *)

    | MenhirState147 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 147.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState148 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 148.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState150 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 150.
        Stack shape : IDENTIFIER LBRACE IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState151 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 151.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState156 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_nclist_expr_ _menhir_cell0_RBRACE, 'r) _menhir_state
      (** State 156.
        Stack shape : IDENTIFIER LBRACE nclist(expr) RBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState159 : (('s, 'r) _menhir_cell1_field_assign, 'r) _menhir_state
      (** State 159.
        Stack shape : field_assign.
        Start symbol: <undetermined>. *)

    | MenhirState160 : ((('s, 'r) _menhir_cell1_field_assign, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 160.
        Stack shape : field_assign IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState161 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 161.
        Stack shape : IDENTIFIER EQ.
        Start symbol: <undetermined>. *)

    | MenhirState162 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_EQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 162.
        Stack shape : IDENTIFIER EQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState164 : (('s, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 164.
        Stack shape : expr.
        Start symbol: <undetermined>. *)

    | MenhirState165 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COMMA, 'r) _menhir_state
      (** State 165.
        Stack shape : expr COMMA.
        Start symbol: <undetermined>. *)

    | MenhirState170 : (('s, 'r) _menhir_cell1_GEQ, 'r) _menhir_state
      (** State 170.
        Stack shape : GEQ.
        Start symbol: <undetermined>. *)

    | MenhirState171 : ((('s, 'r) _menhir_cell1_GEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 171.
        Stack shape : GEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState172 : (('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_state
      (** State 172.
        Stack shape : BNOT.
        Start symbol: <undetermined>. *)

    | MenhirState173 : ((('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 173.
        Stack shape : BNOT LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState175 : (('s, 'r) _menhir_cell1_ARBITRARY, 'r) _menhir_state
      (** State 175.
        Stack shape : ARBITRARY.
        Start symbol: <undetermined>. *)

    | MenhirState183 : (('s, 'r) _menhir_cell1_pattern, 'r) _menhir_state
      (** State 183.
        Stack shape : pattern.
        Start symbol: <undetermined>. *)

    | MenhirState185 : (('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_state
      (** State 185.
        Stack shape : expr_pattern.
        Start symbol: <undetermined>. *)

    | MenhirState186 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SLICING, 'r) _menhir_state
      (** State 186.
        Stack shape : expr_pattern SLICING.
        Start symbol: <undetermined>. *)

    | MenhirState187 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SLICING, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 187.
        Stack shape : expr_pattern SLICING expr.
        Start symbol: <undetermined>. *)

    | MenhirState188 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHR, 'r) _menhir_state
      (** State 188.
        Stack shape : expr_pattern SHR.
        Start symbol: <undetermined>. *)

    | MenhirState189 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 189.
        Stack shape : expr_pattern SHR expr.
        Start symbol: <undetermined>. *)

    | MenhirState190 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHL, 'r) _menhir_state
      (** State 190.
        Stack shape : expr_pattern SHL.
        Start symbol: <undetermined>. *)

    | MenhirState191 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 191.
        Stack shape : expr_pattern SHL expr.
        Start symbol: <undetermined>. *)

    | MenhirState192 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_RDIV, 'r) _menhir_state
      (** State 192.
        Stack shape : expr_pattern RDIV.
        Start symbol: <undetermined>. *)

    | MenhirState193 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_RDIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 193.
        Stack shape : expr_pattern RDIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState194 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_POW, 'r) _menhir_state
      (** State 194.
        Stack shape : expr_pattern POW.
        Start symbol: <undetermined>. *)

    | MenhirState195 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_POW, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 195.
        Stack shape : expr_pattern POW expr.
        Start symbol: <undetermined>. *)

    | MenhirState196 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_PLUS, 'r) _menhir_state
      (** State 196.
        Stack shape : expr_pattern PLUS.
        Start symbol: <undetermined>. *)

    | MenhirState197 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_PLUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 197.
        Stack shape : expr_pattern PLUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState198 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_OR, 'r) _menhir_state
      (** State 198.
        Stack shape : expr_pattern OR.
        Start symbol: <undetermined>. *)

    | MenhirState199 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_OR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 199.
        Stack shape : expr_pattern OR expr.
        Start symbol: <undetermined>. *)

    | MenhirState200 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_NEQ, 'r) _menhir_state
      (** State 200.
        Stack shape : expr_pattern NEQ.
        Start symbol: <undetermined>. *)

    | MenhirState202 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_NEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 202.
        Stack shape : expr_pattern NEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState203 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MUL, 'r) _menhir_state
      (** State 203.
        Stack shape : expr_pattern MUL.
        Start symbol: <undetermined>. *)

    | MenhirState204 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MUL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 204.
        Stack shape : expr_pattern MUL expr.
        Start symbol: <undetermined>. *)

    | MenhirState205 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MOD, 'r) _menhir_state
      (** State 205.
        Stack shape : expr_pattern MOD.
        Start symbol: <undetermined>. *)

    | MenhirState206 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MOD, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 206.
        Stack shape : expr_pattern MOD expr.
        Start symbol: <undetermined>. *)

    | MenhirState207 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 207.
        Stack shape : expr_pattern MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState208 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 208.
        Stack shape : expr_pattern MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState209 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LT, 'r) _menhir_state
      (** State 209.
        Stack shape : expr_pattern LT.
        Start symbol: <undetermined>. *)

    | MenhirState210 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 210.
        Stack shape : expr_pattern LT expr.
        Start symbol: <undetermined>. *)

    | MenhirState211 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_state
      (** State 211.
        Stack shape : expr_pattern LLBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState212 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 212.
        Stack shape : expr_pattern LLBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState214 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LEQ, 'r) _menhir_state
      (** State 214.
        Stack shape : expr_pattern LEQ.
        Start symbol: <undetermined>. *)

    | MenhirState215 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 215.
        Stack shape : expr_pattern LEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState216 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_IN, 'r) _menhir_state
      (** State 216.
        Stack shape : expr_pattern IN.
        Start symbol: <undetermined>. *)

    | MenhirState217 : ((('s, 'r) _menhir_cell1_IN, 'r) _menhir_cell1_BNOT, 'r) _menhir_state
      (** State 217.
        Stack shape : IN BNOT.
        Start symbol: <undetermined>. *)

    | MenhirState219 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_IMPL, 'r) _menhir_state
      (** State 219.
        Stack shape : expr_pattern IMPL.
        Start symbol: <undetermined>. *)

    | MenhirState220 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_IMPL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 220.
        Stack shape : expr_pattern IMPL expr.
        Start symbol: <undetermined>. *)

    | MenhirState221 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GT, 'r) _menhir_state
      (** State 221.
        Stack shape : expr_pattern GT.
        Start symbol: <undetermined>. *)

    | MenhirState222 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 222.
        Stack shape : expr_pattern GT expr.
        Start symbol: <undetermined>. *)

    | MenhirState223 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GEQ, 'r) _menhir_state
      (** State 223.
        Stack shape : expr_pattern GEQ.
        Start symbol: <undetermined>. *)

    | MenhirState224 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 224.
        Stack shape : expr_pattern GEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState225 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_state
      (** State 225.
        Stack shape : expr_pattern EQ_OP.
        Start symbol: <undetermined>. *)

    | MenhirState227 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 227.
        Stack shape : expr_pattern EQ_OP expr.
        Start symbol: <undetermined>. *)

    | MenhirState228 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_EOR, 'r) _menhir_state
      (** State 228.
        Stack shape : expr_pattern EOR.
        Start symbol: <undetermined>. *)

    | MenhirState229 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_EOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 229.
        Stack shape : expr_pattern EOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState231 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DOT _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 231.
        Stack shape : expr_pattern DOT LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState235 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIVRM, 'r) _menhir_state
      (** State 235.
        Stack shape : expr_pattern DIVRM.
        Start symbol: <undetermined>. *)

    | MenhirState236 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIVRM, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 236.
        Stack shape : expr_pattern DIVRM expr.
        Start symbol: <undetermined>. *)

    | MenhirState237 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIV, 'r) _menhir_state
      (** State 237.
        Stack shape : expr_pattern DIV.
        Start symbol: <undetermined>. *)

    | MenhirState238 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 238.
        Stack shape : expr_pattern DIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState239 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_CONCAT, 'r) _menhir_state
      (** State 239.
        Stack shape : expr_pattern CONCAT.
        Start symbol: <undetermined>. *)

    | MenhirState240 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_CONCAT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 240.
        Stack shape : expr_pattern CONCAT expr.
        Start symbol: <undetermined>. *)

    | MenhirState241 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_state
      (** State 241.
        Stack shape : expr_pattern COLON_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState242 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 242.
        Stack shape : expr_pattern COLON_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState243 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BOR, 'r) _menhir_state
      (** State 243.
        Stack shape : expr_pattern BOR.
        Start symbol: <undetermined>. *)

    | MenhirState244 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 244.
        Stack shape : expr_pattern BOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState245 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BEQ, 'r) _menhir_state
      (** State 245.
        Stack shape : expr_pattern BEQ.
        Start symbol: <undetermined>. *)

    | MenhirState246 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 246.
        Stack shape : expr_pattern BEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState247 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BAND, 'r) _menhir_state
      (** State 247.
        Stack shape : expr_pattern BAND.
        Start symbol: <undetermined>. *)

    | MenhirState248 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BAND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 248.
        Stack shape : expr_pattern BAND expr.
        Start symbol: <undetermined>. *)

    | MenhirState249 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_AS, 'r) _menhir_state
      (** State 249.
        Stack shape : expr_pattern AS.
        Start symbol: <undetermined>. *)

    | MenhirState252 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_AND, 'r) _menhir_state
      (** State 252.
        Stack shape : expr_pattern AND.
        Start symbol: <undetermined>. *)

    | MenhirState253 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_AND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 253.
        Stack shape : expr_pattern AND expr.
        Start symbol: <undetermined>. *)

    | MenhirState257 : ((('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 257.
        Stack shape : BNOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState259 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_pattern, 'r) _menhir_state
      (** State 259.
        Stack shape : LPAR pattern.
        Start symbol: <undetermined>. *)

    | MenhirState262 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_state
      (** State 262.
        Stack shape : LPAR expr_pattern.
        Start symbol: <undetermined>. *)

    | MenhirState268 : (('s, 'r) _menhir_cell1_slice, 'r) _menhir_state
      (** State 268.
        Stack shape : slice.
        Start symbol: <undetermined>. *)

    | MenhirState270 : (('s, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 270.
        Stack shape : expr.
        Start symbol: <undetermined>. *)

    | MenhirState271 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_STAR_COLON, 'r) _menhir_state
      (** State 271.
        Stack shape : expr STAR_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState272 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_STAR_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 272.
        Stack shape : expr STAR_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState273 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS_COLON, 'r) _menhir_state
      (** State 273.
        Stack shape : expr PLUS_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState274 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 274.
        Stack shape : expr PLUS_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState275 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON, 'r) _menhir_state
      (** State 275.
        Stack shape : expr COLON.
        Start symbol: <undetermined>. *)

    | MenhirState276 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 276.
        Stack shape : expr COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState283 : (('s, 'r) _menhir_cell1_int_constraint, 'r) _menhir_state
      (** State 283.
        Stack shape : int_constraint.
        Start symbol: <undetermined>. *)

    | MenhirState285 : (('s, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 285.
        Stack shape : expr.
        Start symbol: <undetermined>. *)

    | MenhirState286 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SLICING, 'r) _menhir_state
      (** State 286.
        Stack shape : expr SLICING.
        Start symbol: <undetermined>. *)

    | MenhirState287 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SLICING, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 287.
        Stack shape : expr SLICING expr.
        Start symbol: <undetermined>. *)

    | MenhirState291 : (('s, 'r) _menhir_cell1_ty, 'r) _menhir_state
      (** State 291.
        Stack shape : ty.
        Start symbol: <undetermined>. *)

    | MenhirState297 : ((('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 297.
        Stack shape : BNOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState301 : ((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 301.
        Stack shape : IF expr.
        Start symbol: <undetermined>. *)

    | MenhirState302 : (((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 302.
        Stack shape : IF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState303 : ((((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 303.
        Stack shape : IF expr THEN expr.
        Start symbol: <undetermined>. *)

    | MenhirState305 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 305.
        Stack shape : LPAR expr.
        Start symbol: <undetermined>. *)

    | MenhirState307 : (((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COMMA, 'r) _menhir_state
      (** State 307.
        Stack shape : LPAR expr COMMA.
        Start symbol: <undetermined>. *)

    | MenhirState310 : ((('s, 'r) _menhir_cell1_NOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 310.
        Stack shape : NOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState311 : ((('s, 'r) _menhir_cell1_WHILE, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 311.
        Stack shape : WHILE expr.
        Start symbol: <undetermined>. *)

    | MenhirState312 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LOOPLIMIT, 'r) _menhir_state
      (** State 312.
        Stack shape : expr LOOPLIMIT.
        Start symbol: <undetermined>. *)

    | MenhirState313 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LOOPLIMIT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 313.
        Stack shape : expr LOOPLIMIT expr.
        Start symbol: <undetermined>. *)

    | MenhirState315 : (((('s, 'r) _menhir_cell1_WHILE, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_loop_limit, 'r) _menhir_state
      (** State 315.
        Stack shape : WHILE expr loop_limit.
        Start symbol: <undetermined>. *)

    | MenhirState316 : (('s, 'r) _menhir_cell1_VAR, 'r) _menhir_state
      (** State 316.
        Stack shape : VAR.
        Start symbol: <undetermined>. *)

    | MenhirState318 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 318.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState321 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_discard_or_identifier, 'r) _menhir_state
      (** State 321.
        Stack shape : LPAR discard_or_identifier.
        Start symbol: <undetermined>. *)

    | MenhirState325 : ((('s, 'r) _menhir_cell1_discard_or_identifier, 'r) _menhir_cell1_discard_or_identifier, 'r) _menhir_state
      (** State 325.
        Stack shape : discard_or_identifier discard_or_identifier.
        Start symbol: <undetermined>. *)

    | MenhirState328 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 328.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState332 : ((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_state
      (** State 332.
        Stack shape : VAR decl_item.
        Start symbol: <undetermined>. *)

    | MenhirState333 : (((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 333.
        Stack shape : VAR decl_item EQ.
        Start symbol: <undetermined>. *)

    | MenhirState334 : ((('s, 'r) _menhir_cell1_EQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 334.
        Stack shape : EQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState337 : (((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_state
      (** State 337.
        Stack shape : VAR decl_item as_ty.
        Start symbol: <undetermined>. *)

    | MenhirState338 : ((((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 338.
        Stack shape : VAR decl_item as_ty EQ.
        Start symbol: <undetermined>. *)

    | MenhirState339 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 339.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState340 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 340.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState341 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_RBRACE, 'r) _menhir_state
      (** State 341.
        Stack shape : IDENTIFIER LBRACE RBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState343 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_COMMA, 'r) _menhir_state
      (** State 343.
        Stack shape : IDENTIFIER LBRACE COMMA.
        Start symbol: <undetermined>. *)

    | MenhirState345 : ((((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_COMMA, 'r) _menhir_cell1_nclist_expr_ _menhir_cell0_RBRACE, 'r) _menhir_state
      (** State 345.
        Stack shape : IDENTIFIER LBRACE COMMA nclist(expr) RBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState351 : ((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_clist2_IDENTIFIER_, 'r) _menhir_state
      (** State 351.
        Stack shape : VAR clist2(IDENTIFIER).
        Start symbol: <undetermined>. *)

    | MenhirState358 : (('s, 'r) _menhir_cell1_TRY, 'r) _menhir_state
      (** State 358.
        Stack shape : TRY.
        Start symbol: <undetermined>. *)

    | MenhirState359 : (('s, 'r) _menhir_cell1_THROW, 'r) _menhir_state
      (** State 359.
        Stack shape : THROW.
        Start symbol: <undetermined>. *)

    | MenhirState361 : ((('s, 'r) _menhir_cell1_THROW, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 361.
        Stack shape : THROW expr.
        Start symbol: <undetermined>. *)

    | MenhirState363 : (('s, 'r) _menhir_cell1_RETURN, 'r) _menhir_state
      (** State 363.
        Stack shape : RETURN.
        Start symbol: <undetermined>. *)

    | MenhirState366 : ((('s, 'r) _menhir_cell1_RETURN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 366.
        Stack shape : RETURN expr.
        Start symbol: <undetermined>. *)

    | MenhirState367 : (('s, 'r) _menhir_cell1_REPEAT, 'r) _menhir_state
      (** State 367.
        Stack shape : REPEAT.
        Start symbol: <undetermined>. *)

    | MenhirState368 : (('s, 'r) _menhir_cell1_PRINTLN, 'r) _menhir_state
      (** State 368.
        Stack shape : PRINTLN.
        Start symbol: <undetermined>. *)

    | MenhirState371 : (('s, 'r) _menhir_cell1_PRINT, 'r) _menhir_state
      (** State 371.
        Stack shape : PRINT.
        Start symbol: <undetermined>. *)

    | MenhirState375 : (('s, 'r) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 375.
        Stack shape : PRAGMA IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState381 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 381.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState383 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 383.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState384 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_state
      (** State 384.
        Stack shape : IDENTIFIER LLBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState385 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 385.
        Stack shape : IDENTIFIER LLBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState386 : ((((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RRBRACKET, 'r) _menhir_state
      (** State 386.
        Stack shape : IDENTIFIER LLBRACKET expr RRBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState388 : (('s, 'r) _menhir_cell1_DOT _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 388.
        Stack shape : DOT IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState394 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr, 'r) _menhir_state
      (** State 394.
        Stack shape : LPAR discard_or_sliced_basic_lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState397 : ((('s, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr, 'r) _menhir_state
      (** State 397.
        Stack shape : discard_or_sliced_basic_lexpr discard_or_sliced_basic_lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState399 : (('s, 'r) _menhir_cell1_basic_lexpr, 'r) _menhir_state
      (** State 399.
        Stack shape : basic_lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState404 : (('s, 'r) _menhir_cell1_IF, 'r) _menhir_state
      (** State 404.
        Stack shape : IF.
        Start symbol: <undetermined>. *)

    | MenhirState405 : ((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 405.
        Stack shape : IF expr.
        Start symbol: <undetermined>. *)

    | MenhirState406 : (((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 406.
        Stack shape : IF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState407 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 407.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState408 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 408.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState410 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_DOT _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 410.
        Stack shape : IDENTIFIER DOT LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState414 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_DOT _menhir_cell0_LPAR, 'r) _menhir_cell1_discard_or_field, 'r) _menhir_state
      (** State 414.
        Stack shape : IDENTIFIER DOT LPAR discard_or_field.
        Start symbol: <undetermined>. *)

    | MenhirState417 : ((('s, 'r) _menhir_cell1_discard_or_field, 'r) _menhir_cell1_discard_or_field, 'r) _menhir_state
      (** State 417.
        Stack shape : discard_or_field discard_or_field.
        Start symbol: <undetermined>. *)

    | MenhirState421 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_DOT _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 421.
        Stack shape : IDENTIFIER DOT LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState427 : (('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 427.
        Stack shape : FOR IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState428 : ((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 428.
        Stack shape : FOR IDENTIFIER expr.
        Start symbol: <undetermined>. *)

    | MenhirState431 : (((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_direction, 'r) _menhir_state
      (** State 431.
        Stack shape : FOR IDENTIFIER expr direction.
        Start symbol: <undetermined>. *)

    | MenhirState432 : ((((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_direction, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 432.
        Stack shape : FOR IDENTIFIER expr direction expr.
        Start symbol: <undetermined>. *)

    | MenhirState434 : (((((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_direction, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_loop_limit, 'r) _menhir_state
      (** State 434.
        Stack shape : FOR IDENTIFIER expr direction expr loop_limit.
        Start symbol: <undetermined>. *)

    | MenhirState435 : (('s, 'r) _menhir_cell1_DEBUG, 'r) _menhir_state
      (** State 435.
        Stack shape : DEBUG.
        Start symbol: <undetermined>. *)

    | MenhirState439 : (('s, 'r) _menhir_cell1_CASE, 'r) _menhir_state
      (** State 439.
        Stack shape : CASE.
        Start symbol: <undetermined>. *)

    | MenhirState440 : ((('s, 'r) _menhir_cell1_CASE, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 440.
        Stack shape : CASE expr.
        Start symbol: <undetermined>. *)

    | MenhirState441 : (((('s, 'r) _menhir_cell1_CASE, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OF, 'r) _menhir_state
      (** State 441.
        Stack shape : CASE expr OF.
        Start symbol: <undetermined>. *)

    | MenhirState442 : (('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_state
      (** State 442.
        Stack shape : WHEN.
        Start symbol: <undetermined>. *)

    | MenhirState444 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_state
      (** State 444.
        Stack shape : WHEN pattern_list.
        Start symbol: <undetermined>. *)

    | MenhirState445 : (((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 445.
        Stack shape : WHEN pattern_list expr.
        Start symbol: <undetermined>. *)

    | MenhirState446 : ((((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ARROW, 'r) _menhir_state
      (** State 446.
        Stack shape : WHEN pattern_list expr ARROW.
        Start symbol: <undetermined>. *)

    | MenhirState447 : (('s, 'r) _menhir_cell1_ASSERT, 'r) _menhir_state
      (** State 447.
        Stack shape : ASSERT.
        Start symbol: <undetermined>. *)

    | MenhirState448 : ((('s, 'r) _menhir_cell1_ASSERT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 448.
        Stack shape : ASSERT expr.
        Start symbol: <undetermined>. *)

    | MenhirState451 : (('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 451.
        Stack shape : ARROBASE_LOOPLIMIT LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState452 : ((('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 452.
        Stack shape : ARROBASE_LOOPLIMIT LPAR expr.
        Start symbol: <undetermined>. *)

    | MenhirState454 : (((('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR _menhir_cell0_WHILE, 'r) _menhir_state
      (** State 454.
        Stack shape : ARROBASE_LOOPLIMIT LPAR expr RPAR WHILE.
        Start symbol: <undetermined>. *)

    | MenhirState455 : ((((('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR _menhir_cell0_WHILE, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 455.
        Stack shape : ARROBASE_LOOPLIMIT LPAR expr RPAR WHILE expr.
        Start symbol: <undetermined>. *)

    | MenhirState456 : (((((('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR _menhir_cell0_WHILE, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DO, 'r) _menhir_state
      (** State 456.
        Stack shape : ARROBASE_LOOPLIMIT LPAR expr RPAR WHILE expr DO.
        Start symbol: <undetermined>. *)

    | MenhirState460 : (('s, 'r) _menhir_cell1_stmt, 'r) _menhir_state
      (** State 460.
        Stack shape : stmt.
        Start symbol: <undetermined>. *)

    | MenhirState463 : (('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_state
      (** State 463.
        Stack shape : local_decl_keyword_non_var.
        Start symbol: <undetermined>. *)

    | MenhirState464 : ((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_state
      (** State 464.
        Stack shape : local_decl_keyword_non_var decl_item.
        Start symbol: <undetermined>. *)

    | MenhirState465 : (((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 465.
        Stack shape : local_decl_keyword_non_var decl_item EQ.
        Start symbol: <undetermined>. *)

    | MenhirState466 : ((((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_EQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 466.
        Stack shape : local_decl_keyword_non_var decl_item EQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState469 : (((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_state
      (** State 469.
        Stack shape : local_decl_keyword_non_var decl_item as_ty.
        Start symbol: <undetermined>. *)

    | MenhirState470 : ((((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 470.
        Stack shape : local_decl_keyword_non_var decl_item as_ty expr.
        Start symbol: <undetermined>. *)

    | MenhirState475 : (('s, 'r) _menhir_cell1_lexpr, 'r) _menhir_state
      (** State 475.
        Stack shape : lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState476 : ((('s, 'r) _menhir_cell1_lexpr, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 476.
        Stack shape : lexpr expr.
        Start symbol: <undetermined>. *)

    | MenhirState480 : (('s, 'r) _menhir_cell1_call, 'r) _menhir_state
      (** State 480.
        Stack shape : call.
        Start symbol: <undetermined>. *)

    | MenhirState481 : ((('s, 'r) _menhir_cell1_call, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 481.
        Stack shape : call expr.
        Start symbol: <undetermined>. *)

    | MenhirState484 : (('s, 'r) _menhir_cell1_call _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 484.
        Stack shape : call LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState487 : ((('s, 'r) _menhir_cell1_call _menhir_cell0_LBRACKET, 'r) _menhir_cell1_clist2_IDENTIFIER_ _menhir_cell0_RBRACKET, 'r) _menhir_state
      (** State 487.
        Stack shape : call LBRACKET clist2(IDENTIFIER) RBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState488 : (((('s, 'r) _menhir_cell1_call _menhir_cell0_LBRACKET, 'r) _menhir_cell1_clist2_IDENTIFIER_ _menhir_cell0_RBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 488.
        Stack shape : call LBRACKET clist2(IDENTIFIER) RBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState491 : (('s, 'r) _menhir_cell1_call _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 491.
        Stack shape : call IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState492 : ((('s, 'r) _menhir_cell1_call _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 492.
        Stack shape : call IDENTIFIER expr.
        Start symbol: <undetermined>. *)

    | MenhirState495 : (((('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR _menhir_cell0_REPEAT, 'r) _menhir_state
      (** State 495.
        Stack shape : ARROBASE_LOOPLIMIT LPAR expr RPAR REPEAT.
        Start symbol: <undetermined>. *)

    | MenhirState497 : ((((('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR _menhir_cell0_REPEAT, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 497.
        Stack shape : ARROBASE_LOOPLIMIT LPAR expr RPAR REPEAT stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState498 : (((((('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR _menhir_cell0_REPEAT, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 498.
        Stack shape : ARROBASE_LOOPLIMIT LPAR expr RPAR REPEAT stmt_list expr.
        Start symbol: <undetermined>. *)

    | MenhirState501 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_state
      (** State 501.
        Stack shape : WHEN pattern_list.
        Start symbol: <undetermined>. *)

    | MenhirState506 : ((((('s, 'r) _menhir_cell1_CASE, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OF, 'r) _menhir_cell1_case_alt_list, 'r) _menhir_state
      (** State 506.
        Stack shape : CASE expr OF case_alt_list.
        Start symbol: <undetermined>. *)

    | MenhirState513 : (('s, 'r) _menhir_cell1_case_alt, 'r) _menhir_state
      (** State 513.
        Stack shape : case_alt.
        Start symbol: <undetermined>. *)

    | MenhirState518 : ((((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 518.
        Stack shape : IF expr THEN stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState519 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_state
      (** State 519.
        Stack shape : expr THEN stmt_list ELSIF.
        Start symbol: <undetermined>. *)

    | MenhirState520 : (((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 520.
        Stack shape : expr THEN stmt_list ELSIF expr.
        Start symbol: <undetermined>. *)

    | MenhirState521 : ((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 521.
        Stack shape : expr THEN stmt_list ELSIF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState522 : (((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 522.
        Stack shape : expr THEN stmt_list ELSIF expr THEN stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState523 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSE, 'r) _menhir_state
      (** State 523.
        Stack shape : expr THEN stmt_list ELSE.
        Start symbol: <undetermined>. *)

    | MenhirState530 : ((('s, 'r) _menhir_cell1_REPEAT, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 530.
        Stack shape : REPEAT stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState531 : (((('s, 'r) _menhir_cell1_REPEAT, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 531.
        Stack shape : REPEAT stmt_list expr.
        Start symbol: <undetermined>. *)

    | MenhirState535 : ((('s, 'r) _menhir_cell1_TRY, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 535.
        Stack shape : TRY stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState536 : (('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_state
      (** State 536.
        Stack shape : WHEN.
        Start symbol: <undetermined>. *)

    | MenhirState538 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 538.
        Stack shape : WHEN IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState540 : (((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_ty, 'r) _menhir_state
      (** State 540.
        Stack shape : WHEN IDENTIFIER ty.
        Start symbol: <undetermined>. *)

    | MenhirState543 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_ty, 'r) _menhir_state
      (** State 543.
        Stack shape : WHEN ty.
        Start symbol: <undetermined>. *)

    | MenhirState547 : (((('s, 'r) _menhir_cell1_TRY, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_nlist_catcher_, 'r) _menhir_state
      (** State 547.
        Stack shape : TRY stmt_list nlist(catcher).
        Start symbol: <undetermined>. *)

    | MenhirState552 : (('s, 'r) _menhir_cell1_catcher, 'r) _menhir_state
      (** State 552.
        Stack shape : catcher.
        Start symbol: <undetermined>. *)

    | MenhirState560 : ('s, _menhir_box_spec) _menhir_state
      (** State 560.
        Stack shape : .
        Start symbol: spec. *)

    | MenhirState561 : (('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_state
      (** State 561.
        Stack shape : VAR.
        Start symbol: spec. *)

    | MenhirState564 : ((('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_state
      (** State 564.
        Stack shape : VAR ignored_or_identifier.
        Start symbol: spec. *)

    | MenhirState566 : (((('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_state
      (** State 566.
        Stack shape : VAR ignored_or_identifier option(as_ty).
        Start symbol: spec. *)

    | MenhirState567 : ((((('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_cell1_expr, _menhir_box_spec) _menhir_state
      (** State 567.
        Stack shape : VAR ignored_or_identifier option(as_ty) expr.
        Start symbol: spec. *)

    | MenhirState572 : (('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 572.
        Stack shape : TYPE IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState575 : (('s, _menhir_box_spec) _menhir_cell1_SUBTYPES _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 575.
        Stack shape : SUBTYPES IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState576 : (('s, _menhir_box_spec) _menhir_cell1_LBRACE, _menhir_box_spec) _menhir_state
      (** State 576.
        Stack shape : LBRACE.
        Start symbol: spec. *)

    | MenhirState577 : (('s, _menhir_box_spec) _menhir_cell1_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 577.
        Stack shape : IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState580 : (('s, _menhir_box_spec) _menhir_cell1_typed_identifier, _menhir_box_spec) _menhir_state
      (** State 580.
        Stack shape : typed_identifier.
        Start symbol: spec. *)

    | MenhirState586 : ((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_state
      (** State 586.
        Stack shape : TYPE IDENTIFIER OF.
        Start symbol: spec. *)

    | MenhirState587 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_RECORD, _menhir_box_spec) _menhir_state
      (** State 587.
        Stack shape : TYPE IDENTIFIER OF RECORD.
        Start symbol: spec. *)

    | MenhirState590 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_EXCEPTION, _menhir_box_spec) _menhir_state
      (** State 590.
        Stack shape : TYPE IDENTIFIER OF EXCEPTION.
        Start symbol: spec. *)

    | MenhirState593 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ENUMERATION _menhir_cell0_LBRACE, _menhir_box_spec) _menhir_state
      (** State 593.
        Stack shape : TYPE IDENTIFIER OF ENUMERATION LBRACE.
        Start symbol: spec. *)

    | MenhirState595 : (('s, _menhir_box_spec) _menhir_cell1_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 595.
        Stack shape : IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState599 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ty_decl, _menhir_box_spec) _menhir_state
      (** State 599.
        Stack shape : TYPE IDENTIFIER OF ty_decl.
        Start symbol: spec. *)

    | MenhirState607 : (('s, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 607.
        Stack shape : SETTER IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState608 : (('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_LBRACE, _menhir_box_spec) _menhir_state
      (** State 608.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: spec. *)

    | MenhirState609 : (('s, _menhir_box_spec) _menhir_cell1_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 609.
        Stack shape : IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState612 : (('s, _menhir_box_spec) _menhir_cell1_opt_typed_identifier, _menhir_box_spec) _menhir_state
      (** State 612.
        Stack shape : opt_typed_identifier.
        Start symbol: spec. *)

    | MenhirState616 : ((('s, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 616.
        Stack shape : SETTER IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState617 : ((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_LPAR, _menhir_box_spec) _menhir_state
      (** State 617.
        Stack shape : IDENTIFIER params_opt LPAR.
        Start symbol: spec. *)

    | MenhirState619 : (('s, _menhir_box_spec) _menhir_cell1_typed_identifier, _menhir_box_spec) _menhir_state
      (** State 619.
        Stack shape : typed_identifier.
        Start symbol: spec. *)

    | MenhirState626 : (((('s, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 626.
        Stack shape : SETTER IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState627 : ((((('s, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_typed_identifier, _menhir_box_spec) _menhir_state
      (** State 627.
        Stack shape : SETTER IDENTIFIER params_opt func_args typed_identifier.
        Start symbol: spec. *)

    | MenhirState628 : (('s, _menhir_box_spec) _menhir_cell1_BEGIN, _menhir_box_spec) _menhir_state
      (** State 628.
        Stack shape : BEGIN.
        Start symbol: spec. *)

    | MenhirState635 : (('s, _menhir_box_spec) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 635.
        Stack shape : PRAGMA IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState640 : (('s, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 640.
        Stack shape : GETTER IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState641 : ((('s, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 641.
        Stack shape : GETTER IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState642 : (((('s, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 642.
        Stack shape : GETTER IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState643 : (((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ARROW, _menhir_box_spec) _menhir_state
      (** State 643.
        Stack shape : IDENTIFIER params_opt func_args ARROW.
        Start symbol: spec. *)

    | MenhirState645 : ((((('s, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_state
      (** State 645.
        Stack shape : GETTER IDENTIFIER params_opt func_args return_type.
        Start symbol: spec. *)

    | MenhirState648 : (('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 648.
        Stack shape : FUNC IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState649 : ((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 649.
        Stack shape : FUNC IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState650 : (((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 650.
        Stack shape : FUNC IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState652 : ((((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_state
      (** State 652.
        Stack shape : FUNC IDENTIFIER params_opt func_args return_type.
        Start symbol: spec. *)

    | MenhirState653 : (((((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_expr, _menhir_box_spec) _menhir_state
      (** State 653.
        Stack shape : FUNC IDENTIFIER params_opt func_args return_type expr.
        Start symbol: spec. *)

    | MenhirState654 : ((((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type _menhir_cell0_recurse_limit, _menhir_box_spec) _menhir_state
      (** State 654.
        Stack shape : FUNC IDENTIFIER params_opt func_args return_type recurse_limit.
        Start symbol: spec. *)

    | MenhirState662 : (('s, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var, _menhir_box_spec) _menhir_state
      (** State 662.
        Stack shape : global_decl_keyword_non_var.
        Start symbol: spec. *)

    | MenhirState663 : ((('s, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_state
      (** State 663.
        Stack shape : global_decl_keyword_non_var ignored_or_identifier.
        Start symbol: spec. *)

    | MenhirState665 : (((('s, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_state
      (** State 665.
        Stack shape : global_decl_keyword_non_var ignored_or_identifier option(as_ty).
        Start symbol: spec. *)

    | MenhirState666 : ((((('s, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_cell1_expr, _menhir_box_spec) _menhir_state
      (** State 666.
        Stack shape : global_decl_keyword_non_var ignored_or_identifier option(as_ty) expr.
        Start symbol: spec. *)

    | MenhirState669 : (('s, _menhir_box_spec) _menhir_cell1_decl, _menhir_box_spec) _menhir_state
      (** State 669.
        Stack shape : decl.
        Start symbol: spec. *)

    | MenhirState671 : ('s, _menhir_box_stmts) _menhir_state
      (** State 671.
        Stack shape : .
        Start symbol: stmts. *)

  
  and ('s, 'r) _menhir_cell1_as_ty = 
    | MenhirCell1_as_ty of 's * ('s, 'r) _menhir_state * (AST.ty)
  
  and ('s, 'r) _menhir_cell1_basic_lexpr = 
    | MenhirCell1_basic_lexpr of 's * ('s, 'r) _menhir_state * (Desugar.lhs_access) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_bitfield = 
    | MenhirCell1_bitfield of 's * ('s, 'r) _menhir_state * (AST.bitfield)
  
  and ('s, 'r) _menhir_cell1_call = 
    | MenhirCell1_call of 's * ('s, 'r) _menhir_state * (AST.call) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_case_alt = 
    | MenhirCell1_case_alt of 's * ('s, 'r) _menhir_state * (AST.case_alt_desc AST.annotated)
  
  and ('s, 'r) _menhir_cell1_case_alt_list = 
    | MenhirCell1_case_alt_list of 's * ('s, 'r) _menhir_state * (AST.case_alt_desc AST.annotated list)
  
  and ('s, 'r) _menhir_cell1_catcher = 
    | MenhirCell1_catcher of 's * ('s, 'r) _menhir_state * (AST.catcher)
  
  and ('s, 'r) _menhir_cell1_clist2_IDENTIFIER_ = 
    | MenhirCell1_clist2_IDENTIFIER_ of 's * ('s, 'r) _menhir_state * (string list)
  
  and ('s, 'r) _menhir_cell1_decl = 
    | MenhirCell1_decl of 's * ('s, 'r) _menhir_state * (AST.decl)
  
  and ('s, 'r) _menhir_cell1_decl_item = 
    | MenhirCell1_decl_item of 's * ('s, 'r) _menhir_state * (AST.local_decl_item)
  
  and ('s, 'r) _menhir_cell1_direction = 
    | MenhirCell1_direction of 's * ('s, 'r) _menhir_state * (AST.for_direction)
  
  and ('s, 'r) _menhir_cell1_discard_or_field = 
    | MenhirCell1_discard_or_field of 's * ('s, 'r) _menhir_state * (Desugar.lhs_field option)
  
  and ('s, 'r) _menhir_cell1_discard_or_identifier = 
    | MenhirCell1_discard_or_identifier of 's * ('s, 'r) _menhir_state * (string)
  
  and ('s, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr = 
    | MenhirCell1_discard_or_sliced_basic_lexpr of 's * ('s, 'r) _menhir_state * (Desugar.lhs_access option)
  
  and ('s, 'r) _menhir_cell1_expr = 
    | MenhirCell1_expr of 's * ('s, 'r) _menhir_state * (AST.expr) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_expr_pattern = 
    | MenhirCell1_expr_pattern of 's * ('s, 'r) _menhir_state * (AST.expr) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_field_assign = 
    | MenhirCell1_field_assign of 's * ('s, 'r) _menhir_state * (string * AST.expr)
  
  and ('s, 'r) _menhir_cell1_func_args = 
    | MenhirCell1_func_args of 's * ('s, 'r) _menhir_state * (AST.typed_identifier list)
  
  and ('s, 'r) _menhir_cell1_global_decl_keyword_non_var = 
    | MenhirCell1_global_decl_keyword_non_var of 's * ('s, 'r) _menhir_state * (AST.global_decl_keyword) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ignored_or_identifier = 
    | MenhirCell1_ignored_or_identifier of 's * ('s, 'r) _menhir_state * (string)
  
  and ('s, 'r) _menhir_cell1_int_constraint = 
    | MenhirCell1_int_constraint of 's * ('s, 'r) _menhir_state * (AST.int_constraint)
  
  and ('s, 'r) _menhir_cell1_lexpr = 
    | MenhirCell1_lexpr of 's * ('s, 'r) _menhir_state * (AST.lexpr) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_local_decl_keyword_non_var = 
    | MenhirCell1_local_decl_keyword_non_var of 's * ('s, 'r) _menhir_state * (AST.local_decl_keyword) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_loop_limit = 
    | MenhirCell1_loop_limit of 's * ('s, 'r) _menhir_state * (AST.expr option)
  
  and ('s, 'r) _menhir_cell1_nclist_expr_ = 
    | MenhirCell1_nclist_expr_ of 's * ('s, 'r) _menhir_state * (AST.expr list)
  
  and ('s, 'r) _menhir_cell1_nlist_catcher_ = 
    | MenhirCell1_nlist_catcher_ of 's * ('s, 'r) _menhir_state * (AST.catcher list)
  
  and ('s, 'r) _menhir_cell1_opt_typed_identifier = 
    | MenhirCell1_opt_typed_identifier of 's * ('s, 'r) _menhir_state * (string * AST.ty option)
  
  and ('s, 'r) _menhir_cell1_option_as_ty_ = 
    | MenhirCell1_option_as_ty_ of 's * ('s, 'r) _menhir_state * (AST.ty option)
  
  and ('s, 'r) _menhir_cell1_params_opt = 
    | MenhirCell1_params_opt of 's * ('s, 'r) _menhir_state * ((string * AST.ty option) list)
  
  and ('s, 'r) _menhir_cell1_pattern = 
    | MenhirCell1_pattern of 's * ('s, 'r) _menhir_state * (AST.pattern) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_pattern_list = 
    | MenhirCell1_pattern_list of 's * ('s, 'r) _menhir_state * (AST.pattern)
  
  and 's _menhir_cell0_recurse_limit = 
    | MenhirCell0_recurse_limit of 's * (AST.expr option)
  
  and ('s, 'r) _menhir_cell1_return_type = 
    | MenhirCell1_return_type of 's * ('s, 'r) _menhir_state * (AST.ty)
  
  and ('s, 'r) _menhir_cell1_slice = 
    | MenhirCell1_slice of 's * ('s, 'r) _menhir_state * (AST.slice)
  
  and ('s, 'r) _menhir_cell1_slices = 
    | MenhirCell1_slices of 's * ('s, 'r) _menhir_state * (AST.slice list) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_stmt = 
    | MenhirCell1_stmt of 's * ('s, 'r) _menhir_state * (AST.stmt_desc AST.annotated) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_stmt_list = 
    | MenhirCell1_stmt_list of 's * ('s, 'r) _menhir_state * (AST.stmt) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ty = 
    | MenhirCell1_ty of 's * ('s, 'r) _menhir_state * (AST.ty) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ty_decl = 
    | MenhirCell1_ty_decl of 's * ('s, 'r) _menhir_state * (AST.ty)
  
  and ('s, 'r) _menhir_cell1_typed_identifier = 
    | MenhirCell1_typed_identifier of 's * ('s, 'r) _menhir_state * (AST.field)
  
  and ('s, 'r) _menhir_cell1_AND = 
    | MenhirCell1_AND of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ARBITRARY = 
    | MenhirCell1_ARBITRARY of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ARRAY = 
    | MenhirCell1_ARRAY of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ARROBASE_LOOPLIMIT = 
    | MenhirCell1_ARROBASE_LOOPLIMIT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ARROW = 
    | MenhirCell1_ARROW of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_AS = 
    | MenhirCell1_AS of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ASSERT = 
    | MenhirCell1_ASSERT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_BAND = 
    | MenhirCell1_BAND of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_BEGIN = 
    | MenhirCell1_BEGIN of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_BEQ = 
    | MenhirCell1_BEQ of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_BITS = 
    | MenhirCell1_BITS of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_BNOT = 
    | MenhirCell1_BNOT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_BOR = 
    | MenhirCell1_BOR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_CASE = 
    | MenhirCell1_CASE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_COLON = 
    | MenhirCell1_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_COLON_COLON = 
    | MenhirCell1_COLON_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_COMMA = 
    | MenhirCell1_COMMA of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_CONCAT = 
    | MenhirCell1_CONCAT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_DEBUG = 
    | MenhirCell1_DEBUG of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_DIV = 
    | MenhirCell1_DIV of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_DIVRM = 
    | MenhirCell1_DIVRM of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_DO = 
    | MenhirCell1_DO of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_DOT = 
    | MenhirCell1_DOT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ELSE = 
    | MenhirCell1_ELSE of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ELSIF = 
    | MenhirCell1_ELSIF of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ENUMERATION = 
    | MenhirCell1_ENUMERATION of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_EOR = 
    | MenhirCell1_EOR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_EQ = 
    | MenhirCell1_EQ of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_EQ_OP = 
    | MenhirCell1_EQ_OP of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_EXCEPTION = 
    | MenhirCell1_EXCEPTION of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_FOR = 
    | MenhirCell1_FOR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_FUNC = 
    | MenhirCell1_FUNC of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_GEQ = 
    | MenhirCell1_GEQ of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_GETTER = 
    | MenhirCell1_GETTER of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_GT = 
    | MenhirCell1_GT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_IDENTIFIER = 
    | MenhirCell1_IDENTIFIER of 's * ('s, 'r) _menhir_state * (
# 41 "asllib/Tokens.mly"
       (string)
# 2060 "asllib/Parser.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_IDENTIFIER = 
    | MenhirCell0_IDENTIFIER of 's * (
# 41 "asllib/Tokens.mly"
       (string)
# 2067 "asllib/Parser.ml"
  ) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_IF = 
    | MenhirCell1_IF of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_IMPL = 
    | MenhirCell1_IMPL of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_IN = 
    | MenhirCell1_IN of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_INTEGER = 
    | MenhirCell1_INTEGER of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LBRACE = 
    | MenhirCell1_LBRACE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_LBRACE = 
    | MenhirCell0_LBRACE of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LBRACKET = 
    | MenhirCell1_LBRACKET of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_LBRACKET = 
    | MenhirCell0_LBRACKET of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LEQ = 
    | MenhirCell1_LEQ of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LLBRACKET = 
    | MenhirCell1_LLBRACKET of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_LOOPLIMIT = 
    | MenhirCell1_LOOPLIMIT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_LPAR = 
    | MenhirCell1_LPAR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_LPAR = 
    | MenhirCell0_LPAR of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LT = 
    | MenhirCell1_LT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_MINUS = 
    | MenhirCell1_MINUS of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_MOD = 
    | MenhirCell1_MOD of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_MUL = 
    | MenhirCell1_MUL of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_NEQ = 
    | MenhirCell1_NEQ of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_NOT = 
    | MenhirCell1_NOT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_OF = 
    | MenhirCell1_OF of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_OR = 
    | MenhirCell1_OR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_PLUS = 
    | MenhirCell1_PLUS of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_PLUS_COLON = 
    | MenhirCell1_PLUS_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_POW = 
    | MenhirCell1_POW of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_PRAGMA = 
    | MenhirCell1_PRAGMA of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_PRINT = 
    | MenhirCell1_PRINT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_PRINTLN = 
    | MenhirCell1_PRINTLN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RBRACE = 
    | MenhirCell1_RBRACE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_RBRACE = 
    | MenhirCell0_RBRACE of 's * Lexing.position
  
  and 's _menhir_cell0_RBRACKET = 
    | MenhirCell0_RBRACKET of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RDIV = 
    | MenhirCell1_RDIV of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_RECORD = 
    | MenhirCell1_RECORD of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_REPEAT = 
    | MenhirCell1_REPEAT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_REPEAT = 
    | MenhirCell0_REPEAT of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RETURN = 
    | MenhirCell1_RETURN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RPAR = 
    | MenhirCell1_RPAR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RRBRACKET = 
    | MenhirCell1_RRBRACKET of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_SETTER = 
    | MenhirCell1_SETTER of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_SHL = 
    | MenhirCell1_SHL of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SHR = 
    | MenhirCell1_SHR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SLICING = 
    | MenhirCell1_SLICING of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_STAR_COLON = 
    | MenhirCell1_STAR_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SUBTYPES = 
    | MenhirCell1_SUBTYPES of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_THEN = 
    | MenhirCell1_THEN of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_THROW = 
    | MenhirCell1_THROW of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_TRY = 
    | MenhirCell1_TRY of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_TYPE = 
    | MenhirCell1_TYPE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_VAR = 
    | MenhirCell1_VAR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_WHEN = 
    | MenhirCell1_WHEN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_WHILE = 
    | MenhirCell1_WHILE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_WHILE = 
    | MenhirCell0_WHILE of 's * Lexing.position
  
  and _menhir_box_stmts = 
    | MenhirBox_stmts of (AST.stmt) [@@unboxed]
  
  and _menhir_box_spec = 
    | MenhirBox_spec of (AST.t) [@@unboxed]
  
  and _menhir_box_opn = 
    | MenhirBox_opn of (AST.t) [@@unboxed]
  
  let _menhir_action_003 =
    fun _1 _2 ->
      (_2 : (AST.ty))
  
  let _menhir_action_004 =
    fun _endpos_desc_ _startpos_desc_ desc fields ->
      let base =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2244 "asllib/Parser.ml"
        
      in
      (
# 423 "asllib/Parser.mly"
    ( lhs_access ~base ~index:None ~fields )
# 2250 "asllib/Parser.ml"
       : (Desugar.lhs_access))
  
  let _menhir_action_005 =
    fun _2 _4 _endpos_desc_ _startpos_desc_ desc fields idx ->
      let base =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2261 "asllib/Parser.ml"
        
      in
      (
# 425 "asllib/Parser.mly"
    ( lhs_access ~base ~index:(Some idx) ~fields )
# 2267 "asllib/Parser.ml"
       : (Desugar.lhs_access))
  
  let _menhir_action_006 =
    fun s x ->
      (
# 376 "asllib/Parser.mly"
                                              ( BitField_Simple (x, s)     )
# 2275 "asllib/Parser.ml"
       : (AST.bitfield))
  
  let _menhir_action_007 =
    fun bf s x ->
      (
# 377 "asllib/Parser.mly"
                                              ( BitField_Nested (x, s, bf) )
# 2283 "asllib/Parser.ml"
       : (AST.bitfield))
  
  let _menhir_action_008 =
    fun s ty x ->
      (
# 378 "asllib/Parser.mly"
                                              ( BitField_Type   (x, s, ty) )
# 2291 "asllib/Parser.ml"
       : (AST.bitfield))
  
  let _menhir_action_009 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 2299 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_010 =
    fun () ->
      (
# 373 "asllib/Parser.mly"
                     ( [] )
# 2307 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_011 =
    fun _1 ->
      (_1 : (AST.bitfield list))
  
  let _menhir_action_012 =
    fun args name ->
      (
# 582 "asllib/Parser.mly"
    ( { name; params=[]; args; call_type = ST_Function } )
# 2319 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_013 =
    fun name x ->
      let args = 
# 579 "asllib/Parser.mly"
                     ( [] )
# 2327 "asllib/Parser.ml"
       in
      let params = 
# 205 "<standard.mly>"
    ( x )
# 2332 "asllib/Parser.ml"
       in
      (
# 584 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 2337 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_014 =
    fun _1_inlined1 name x ->
      let args = _1_inlined1 in
      let params = 
# 205 "<standard.mly>"
    ( x )
# 2346 "asllib/Parser.ml"
       in
      (
# 584 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 2351 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_015 =
    fun _1 _4 _endpos_stmt_ _startpos__1_ pattern stmt ->
      let desc =
        let where = 
# 123 "<standard.mly>"
    ( None )
# 2360 "asllib/Parser.ml"
         in
        
# 491 "asllib/Parser.mly"
      ( { pattern; where; stmt } )
# 2365 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_stmt_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2374 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated))
  
  let _menhir_action_016 =
    fun _1 _1_inlined1 _2 _4 _endpos_stmt_ _startpos__1_ pattern stmt ->
      let desc =
        let where =
          let _1 = _1_inlined1 in
          let x = _2 in
          
# 126 "<standard.mly>"
    ( Some x )
# 2386 "asllib/Parser.ml"
          
        in
        
# 491 "asllib/Parser.mly"
      ( { pattern; where; stmt } )
# 2392 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_stmt_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2401 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated))
  
  let _menhir_action_017 =
    fun _1 ->
      (_1 : (AST.case_alt_desc AST.annotated list))
  
  let _menhir_action_018 =
    fun _1 _2 otherwise_stmt ->
      (
# 496 "asllib/Parser.mly"
      ( otherwise_stmt )
# 2413 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_019 =
    fun _1 _4 stmt_list ty ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 2421 "asllib/Parser.ml"
       in
      let _2_1 = _2 in
      ((_2_1, ty, stmt_list) : (AST.catcher))
  
  let _menhir_action_020 =
    fun _1 _4 stmt_list ty x ->
      let _2 =
        let x = 
# 196 "<standard.mly>"
    ( x )
# 2432 "asllib/Parser.ml"
         in
        
# 126 "<standard.mly>"
    ( Some x )
# 2437 "asllib/Parser.ml"
        
      in
      let _2_1 = _2 in
      ((_2_1, ty, stmt_list) : (AST.catcher))
  
  let _menhir_action_021 =
    fun () ->
      (
# 154 "asllib/Parser.mly"
                ( [] )
# 2448 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_022 =
    fun _1 ->
      (_1 : (AST.expr list))
  
  let _menhir_action_023 =
    fun () ->
      (
# 154 "asllib/Parser.mly"
                ( [] )
# 2460 "asllib/Parser.ml"
       : (AST.ty list))
  
  let _menhir_action_024 =
    fun _1 ->
      (_1 : (AST.ty list))
  
  let _menhir_action_025 =
    fun () ->
      (
# 154 "asllib/Parser.mly"
                ( [] )
# 2472 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_026 =
    fun _1 ->
      (_1 : (AST.typed_identifier list))
  
  let _menhir_action_027 =
    fun _2 li x ->
      (
# 157 "asllib/Parser.mly"
                                           ( x :: li )
# 2484 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_028 =
    fun _2 li x ->
      (
# 157 "asllib/Parser.mly"
                                           ( x :: li )
# 2492 "asllib/Parser.ml"
       : (Desugar.lhs_field option list))
  
  let _menhir_action_029 =
    fun _2 li x ->
      (
# 157 "asllib/Parser.mly"
                                           ( x :: li )
# 2500 "asllib/Parser.ml"
       : (Desugar.lhs_access option list))
  
  let _menhir_action_030 =
    fun x ->
      let _1 = 
# 205 "<standard.mly>"
    ( x )
# 2508 "asllib/Parser.ml"
       in
      let _1_1 = _1 in
      (
# 299 "asllib/Parser.mly"
                                        WellConstrained  
# 2514 "asllib/Parser.ml"
       _1_1 : (AST.constraint_kind))
  
  let _menhir_action_031 =
    fun x ->
      let _1 = 
# 205 "<standard.mly>"
    ( x )
# 2522 "asllib/Parser.ml"
       in
      (
# 300 "asllib/Parser.mly"
                   ( PendingConstrained )
# 2527 "asllib/Parser.ml"
       : (AST.constraint_kind))
  
  let _menhir_action_032 =
    fun _1 ->
      (_1 : (AST.constraint_kind))
  
  let _menhir_action_033 =
    fun () ->
      (
# 297 "asllib/Parser.mly"
                                             ( UnConstrained )
# 2539 "asllib/Parser.ml"
       : (AST.constraint_kind))
  
  let _menhir_action_034 =
    fun _1 _endpos_body_ _startpos__1_ body func_args name params_opt recurse_limit return_type ->
      let desc = 
# 602 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = Some return_type;
            subprogram_type = ST_Function;
            recurse_limit;
            builtin = false;
          }
        )
# 2558 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_body_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2566 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_035 =
    fun _1 _endpos_body_ _startpos__1_ body func_args name params_opt ->
      let desc = 
# 617 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            builtin = false;
          }
        )
# 2585 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_body_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2593 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_036 =
    fun _1 _endpos_func_body_ _startpos__1_ func_args func_body name params_opt return_type ->
      let desc = 
# 633 "asllib/Parser.mly"
        (
          D_Func
            {
              name;
              parameters = params_opt;
              args = func_args;
              return_type = Some return_type;
              body = SB_ASL func_body;
              subprogram_type = ST_Getter;
              recurse_limit = None;
              builtin = false;
            }
        )
# 2613 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_func_body_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2621 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_037 =
    fun _1 _5 _endpos_func_body_ _startpos__1_ func_args func_body name params_opt v ->
      let desc = 
# 650 "asllib/Parser.mly"
        (
          D_Func
            {
              name;
              parameters = params_opt;
              args = v :: func_args;
              return_type = None;
              body = SB_ASL func_body;
              subprogram_type = ST_Setter;
              recurse_limit = None;
              builtin = false;
            }
        )
# 2641 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_func_body_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2649 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_038 =
    fun _1 _3 _endpos__2_ _startpos__1_ subtype_opt t x ->
      let desc =
        let x = 
# 666 "asllib/Parser.mly"
                                                            D_TypeDecl            
# 2658 "asllib/Parser.ml"
         (x, t, subtype_opt) in
        
# 196 "<standard.mly>"
    ( x )
# 2663 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2672 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_039 =
    fun _1 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc x ->
      let desc =
        let x =
          let s =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2685 "asllib/Parser.ml"
            
          in
          
# 669 "asllib/Parser.mly"
                                                            make_ty_decl_subtype  
# 2691 "asllib/Parser.ml"
           (x, s)
        in
        
# 196 "<standard.mly>"
    ( x )
# 2697 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2706 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_040 =
    fun _4 _endpos__2_ _startpos_keyword_ keyword name ty x ->
      let desc =
        let x =
          let initial_value = 
# 136 "asllib/Parser.mly"
                        Some 
# 2716 "asllib/Parser.ml"
           x in
          
# 674 "asllib/Parser.mly"
        ( D_GlobalStorage { keyword; name; ty; initial_value } )
# 2721 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 2727 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_keyword_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2736 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_041 =
    fun _1 _4 _endpos__2_ _startpos__1_ name ty x ->
      let desc =
        let x =
          let initial_value = 
# 136 "asllib/Parser.mly"
                        Some 
# 2746 "asllib/Parser.ml"
           x in
          
# 677 "asllib/Parser.mly"
        ( D_GlobalStorage { keyword=GDK_Var; name; ty; initial_value } )
# 2751 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 2757 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2766 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_042 =
    fun _1 _endpos__2_ _startpos__1_ name x ->
      let desc =
        let x =
          let ty = 
# 136 "asllib/Parser.mly"
                        Some 
# 2776 "asllib/Parser.ml"
           x in
          
# 681 "asllib/Parser.mly"
        ( D_GlobalStorage { keyword=GDK_Var; name; ty; initial_value=None})
# 2781 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 2787 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2796 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_043 =
    fun _1 _endpos__2_ _startpos__1_ e x ->
      let desc =
        let x = 
# 684 "asllib/Parser.mly"
                                               D_Pragma  
# 2805 "asllib/Parser.ml"
         (x, e) in
        
# 196 "<standard.mly>"
    ( x )
# 2810 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2819 "asllib/Parser.ml"
       : (AST.decl))
  
  let _menhir_action_044 =
    fun discard_or_identifier ->
      (
# 463 "asllib/Parser.mly"
                                         LDI_Var  
# 2827 "asllib/Parser.ml"
       discard_or_identifier : (AST.local_decl_item))
  
  let _menhir_action_045 =
    fun _1 ->
      let _1_1 = _1 in
      (
# 464 "asllib/Parser.mly"
                                         LDI_Tuple  
# 2836 "asllib/Parser.ml"
       _1_1 : (AST.local_decl_item))
  
  let _menhir_action_046 =
    fun _1 ->
      (
# 486 "asllib/Parser.mly"
                       ( AST.Up )
# 2844 "asllib/Parser.ml"
       : (AST.for_direction))
  
  let _menhir_action_047 =
    fun _1 ->
      (
# 486 "asllib/Parser.mly"
                                            ( AST.Down )
# 2852 "asllib/Parser.ml"
       : (AST.for_direction))
  
  let _menhir_action_048 =
    fun _1 ->
      (
# 440 "asllib/Parser.mly"
                             ( None )
# 2860 "asllib/Parser.ml"
       : (Desugar.lhs_field option))
  
  let _menhir_action_049 =
    fun _endpos_desc_ _startpos_desc_ desc ->
      let _1 =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2871 "asllib/Parser.ml"
        
      in
      let _1_1 = _1 in
      (
# 441 "asllib/Parser.mly"
                               Some  
# 2878 "asllib/Parser.ml"
       _1_1 : (Desugar.lhs_field option))
  
  let _menhir_action_050 =
    fun _1 ->
      (
# 459 "asllib/Parser.mly"
                   ( fresh_var "__ldi_discard" )
# 2886 "asllib/Parser.ml"
       : (string))
  
  let _menhir_action_051 =
    fun _1 ->
      (_1 : (string))
  
  let _menhir_action_052 =
    fun _1 ->
      (
# 436 "asllib/Parser.mly"
                          ( None )
# 2898 "asllib/Parser.ml"
       : (Desugar.lhs_access option))
  
  let _menhir_action_053 =
    fun sliced_basic_lexpr ->
      (
# 437 "asllib/Parser.mly"
                            Some  
# 2906 "asllib/Parser.ml"
       sliced_basic_lexpr : (Desugar.lhs_access option))
  
  let _menhir_action_054 =
    fun _1 _2 ->
      (_2 : (AST.expr))
  
  let _menhir_action_055 =
    fun _1 _3 _endpos_e_else_ _startpos__1_ c e e_else ->
      let desc = 
# 254 "asllib/Parser.mly"
                                                        E_Cond 
# 2918 "asllib/Parser.ml"
       (c, e, e_else) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_else_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2926 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_056 =
    fun _2 _3 args name ->
      (
# 587 "asllib/Parser.mly"
    ( { name; params=[]; args; call_type = ST_Function } )
# 2934 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_057 =
    fun _2 _3 _5 name params ->
      let args = 
# 579 "asllib/Parser.mly"
                     ( [] )
# 2942 "asllib/Parser.ml"
       in
      (
# 589 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 2947 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_058 =
    fun _1 _2 _3 _5 name params ->
      let args = _1 in
      (
# 589 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 2956 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_059 =
    fun _endpos_value_ _startpos_value_ value ->
      let desc = 
# 260 "asllib/Parser.mly"
                                                                    E_Literal             
# 2964 "asllib/Parser.ml"
       value in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_value_, _startpos_value_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2972 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_060 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 261 "asllib/Parser.mly"
                                                                    E_Var                 
# 2982 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2991 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_061 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 216 "asllib/Parser.mly"
                  ( AND    )
# 3000 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3005 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3014 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_062 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 217 "asllib/Parser.mly"
                  ( BAND   )
# 3023 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3028 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3037 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_063 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 218 "asllib/Parser.mly"
                  ( BOR    )
# 3046 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3051 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3060 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_064 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 219 "asllib/Parser.mly"
                  ( EQ_OP  )
# 3069 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3074 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3083 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_065 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 220 "asllib/Parser.mly"
                  ( DIV    )
# 3092 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3097 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3106 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_066 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 221 "asllib/Parser.mly"
                  ( DIVRM  )
# 3115 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3120 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3129 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_067 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 222 "asllib/Parser.mly"
                  ( EOR    )
# 3138 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3143 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3152 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_068 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 223 "asllib/Parser.mly"
                  ( EQ_OP  )
# 3161 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3166 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3175 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_069 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 224 "asllib/Parser.mly"
                  ( NEQ    )
# 3184 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3189 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3198 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_070 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 225 "asllib/Parser.mly"
                  ( GT     )
# 3207 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3212 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3221 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_071 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 226 "asllib/Parser.mly"
                  ( GEQ    )
# 3230 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3235 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3244 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_072 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 227 "asllib/Parser.mly"
                  ( IMPL   )
# 3253 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3258 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3267 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_073 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 228 "asllib/Parser.mly"
                  ( LT     )
# 3276 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3281 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3290 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_074 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 229 "asllib/Parser.mly"
                  ( LEQ    )
# 3299 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3304 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3313 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_075 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 230 "asllib/Parser.mly"
                  ( PLUS   )
# 3322 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3327 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3336 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_076 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 231 "asllib/Parser.mly"
                  ( MINUS  )
# 3345 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3350 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3359 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_077 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 232 "asllib/Parser.mly"
                  ( MOD    )
# 3368 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3373 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3382 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_078 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 233 "asllib/Parser.mly"
                  ( MUL    )
# 3391 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3396 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3405 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_079 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 234 "asllib/Parser.mly"
                  ( OR     )
# 3414 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3419 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3428 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_080 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 235 "asllib/Parser.mly"
                  ( RDIV   )
# 3437 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3442 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3451 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_081 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 236 "asllib/Parser.mly"
                  ( SHL    )
# 3460 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3465 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3474 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_082 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 237 "asllib/Parser.mly"
                  ( SHR    )
# 3483 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3488 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3497 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_083 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 238 "asllib/Parser.mly"
                  ( POW    )
# 3506 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3511 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3520 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_084 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 239 "asllib/Parser.mly"
                  ( BV_CONCAT )
# 3529 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3534 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3543 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_085 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op =
          let _1 = () in
          
# 213 "asllib/Parser.mly"
                                  ( PLUS )
# 3554 "asllib/Parser.ml"
          
        in
        
# 262 "asllib/Parser.mly"
                                                                e_binop               
# 3560 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3569 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_086 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 209 "asllib/Parser.mly"
            ( BNOT )
# 3578 "asllib/Parser.ml"
         in
        
# 263 "asllib/Parser.mly"
                                                                    E_Unop                
# 3583 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3592 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_087 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 210 "asllib/Parser.mly"
            ( NEG )
# 3601 "asllib/Parser.ml"
         in
        
# 263 "asllib/Parser.mly"
                                                                    E_Unop                
# 3606 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3615 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_088 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 211 "asllib/Parser.mly"
            ( NOT )
# 3624 "asllib/Parser.ml"
         in
        
# 263 "asllib/Parser.mly"
                                                                    E_Unop                
# 3629 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3638 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_089 =
    fun _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else ->
      let desc = 
# 264 "asllib/Parser.mly"
                                                                    E_Cond                
# 3646 "asllib/Parser.ml"
       (e1, e2, e_else) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_else_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3654 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_090 =
    fun _endpos_call_ _startpos_call_ call ->
      let desc = 
# 265 "asllib/Parser.mly"
                                                                    e_call                
# 3662 "asllib/Parser.ml"
       call in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_call_, _startpos_call_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3670 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_091 =
    fun _endpos_slices_ _startpos_e_ e slices ->
      let desc = 
# 266 "asllib/Parser.mly"
                                                                E_Slice               
# 3678 "asllib/Parser.ml"
       (e, slices) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_slices_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3686 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_092 =
    fun _2 _4 _endpos__4_ _startpos_e1_ e1 e2 ->
      let desc = 
# 267 "asllib/Parser.mly"
                                                                E_GetArray            
# 3694 "asllib/Parser.ml"
       (e1, e2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__4_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3702 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_093 =
    fun _2 _endpos_x_ _startpos_e_ e x ->
      let desc = 
# 268 "asllib/Parser.mly"
                                                                E_GetField            
# 3710 "asllib/Parser.ml"
       (e, x) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_x_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3718 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_094 =
    fun _2 _endpos__3_ _startpos_e_ e x ->
      let desc =
        let fs = 
# 205 "<standard.mly>"
    ( x )
# 3727 "asllib/Parser.ml"
         in
        
# 269 "asllib/Parser.mly"
                                                                E_GetFields           
# 3732 "asllib/Parser.ml"
         (e, fs)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3741 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_095 =
    fun _2 _endpos_ty_ _startpos_expr_ expr ty ->
      let desc = 
# 271 "asllib/Parser.mly"
                                                                E_ATC                 
# 3749 "asllib/Parser.ml"
       (expr, ty) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3757 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_096 =
    fun _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_ constraint_kind expr ->
      let desc =
        let implicit_t_int =
          let desc = 
# 411 "asllib/Parser.mly"
                                                       T_Int 
# 3767 "asllib/Parser.ml"
           constraint_kind in
          let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_constraint_kind_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3775 "asllib/Parser.ml"
          
        in
        
# 272 "asllib/Parser.mly"
                                                                E_ATC                 
# 3781 "asllib/Parser.ml"
         (expr, implicit_t_int)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3790 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_097 =
    fun _2 _endpos_pattern_set_ _startpos_expr_ expr pattern_set ->
      let desc = 
# 274 "asllib/Parser.mly"
                                                                E_Pattern             
# 3798 "asllib/Parser.ml"
       (expr, pattern_set) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_pattern_set_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3806 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_098 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_ expr ->
      let desc =
        let pattern_mask =
          let desc =
            let _1_1 = _1 in
            
# 339 "asllib/Parser.mly"
                                            Pattern_Mask  
# 3818 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3827 "asllib/Parser.ml"
          
        in
        
# 275 "asllib/Parser.mly"
                                                                E_Pattern             
# 3833 "asllib/Parser.ml"
         (expr, pattern_mask)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3842 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_099 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e ->
      let desc =
        let p =
          let desc =
            let _1_1 = _1 in
            
# 339 "asllib/Parser.mly"
                                            Pattern_Mask  
# 3854 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3863 "asllib/Parser.ml"
          
        in
        
# 276 "asllib/Parser.mly"
                                                              ( E_Pattern (e, Pattern_Not (p) |> add_pos_from p) )
# 3869 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3878 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_100 =
    fun _1 _2 _endpos_ty_ _startpos__1_ ty ->
      let desc = 
# 277 "asllib/Parser.mly"
                                                                E_Arbitrary         
# 3886 "asllib/Parser.ml"
       ty in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3894 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_101 =
    fun _endpos__3_ _startpos__1_ x ->
      let desc =
        let e = 
# 205 "<standard.mly>"
    ( x )
# 3903 "asllib/Parser.ml"
         in
        
# 278 "asllib/Parser.mly"
                                                              ( E_Tuple [ e ]        )
# 3908 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3917 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_102 =
    fun _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc ->
      let desc =
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3929 "asllib/Parser.ml"
          
        in
        
# 282 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), []) )
# 3935 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3944 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_103 =
    fun _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let fields = 
# 205 "<standard.mly>"
    ( x )
# 3953 "asllib/Parser.ml"
         in
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3961 "asllib/Parser.ml"
          
        in
        
# 284 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), fields) )
# 3967 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3976 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_104 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 286 "asllib/Parser.mly"
                                                                    E_Tuple               
# 3986 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3995 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_105 =
    fun _endpos_value_ _startpos_value_ value ->
      let desc = 
# 311 "asllib/Parser.mly"
                                                                    E_Literal             
# 4003 "asllib/Parser.ml"
       value in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_value_, _startpos_value_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4011 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_106 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 312 "asllib/Parser.mly"
                                                                    E_Var                 
# 4021 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4030 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_107 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 216 "asllib/Parser.mly"
                  ( AND    )
# 4039 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4044 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4053 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_108 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 217 "asllib/Parser.mly"
                  ( BAND   )
# 4062 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4067 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4076 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_109 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 218 "asllib/Parser.mly"
                  ( BOR    )
# 4085 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4090 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4099 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_110 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 219 "asllib/Parser.mly"
                  ( EQ_OP  )
# 4108 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4113 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4122 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_111 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 220 "asllib/Parser.mly"
                  ( DIV    )
# 4131 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4136 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4145 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_112 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 221 "asllib/Parser.mly"
                  ( DIVRM  )
# 4154 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4159 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4168 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_113 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 222 "asllib/Parser.mly"
                  ( EOR    )
# 4177 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4182 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4191 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_114 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 223 "asllib/Parser.mly"
                  ( EQ_OP  )
# 4200 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4205 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4214 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_115 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 224 "asllib/Parser.mly"
                  ( NEQ    )
# 4223 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4228 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4237 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_116 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 225 "asllib/Parser.mly"
                  ( GT     )
# 4246 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4251 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4260 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_117 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 226 "asllib/Parser.mly"
                  ( GEQ    )
# 4269 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4274 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4283 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_118 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 227 "asllib/Parser.mly"
                  ( IMPL   )
# 4292 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4297 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4306 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_119 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 228 "asllib/Parser.mly"
                  ( LT     )
# 4315 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4320 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4329 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_120 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 229 "asllib/Parser.mly"
                  ( LEQ    )
# 4338 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4343 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4352 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_121 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 230 "asllib/Parser.mly"
                  ( PLUS   )
# 4361 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4366 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4375 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_122 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 231 "asllib/Parser.mly"
                  ( MINUS  )
# 4384 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4389 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4398 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_123 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 232 "asllib/Parser.mly"
                  ( MOD    )
# 4407 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4412 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4421 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_124 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 233 "asllib/Parser.mly"
                  ( MUL    )
# 4430 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4435 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4444 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_125 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 234 "asllib/Parser.mly"
                  ( OR     )
# 4453 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4458 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4467 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_126 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 235 "asllib/Parser.mly"
                  ( RDIV   )
# 4476 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4481 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4490 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_127 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 236 "asllib/Parser.mly"
                  ( SHL    )
# 4499 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4504 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4513 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_128 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 237 "asllib/Parser.mly"
                  ( SHR    )
# 4522 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4527 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4536 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_129 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 238 "asllib/Parser.mly"
                  ( POW    )
# 4545 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4550 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4559 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_130 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 239 "asllib/Parser.mly"
                  ( BV_CONCAT )
# 4568 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4573 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4582 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_131 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op =
          let _1 = () in
          
# 213 "asllib/Parser.mly"
                                  ( PLUS )
# 4593 "asllib/Parser.ml"
          
        in
        
# 313 "asllib/Parser.mly"
                                                                        e_binop               
# 4599 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4608 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_132 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 209 "asllib/Parser.mly"
            ( BNOT )
# 4617 "asllib/Parser.ml"
         in
        
# 314 "asllib/Parser.mly"
                                                                    E_Unop                
# 4622 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4631 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_133 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 210 "asllib/Parser.mly"
            ( NEG )
# 4640 "asllib/Parser.ml"
         in
        
# 314 "asllib/Parser.mly"
                                                                    E_Unop                
# 4645 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4654 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_134 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 211 "asllib/Parser.mly"
            ( NOT )
# 4663 "asllib/Parser.ml"
         in
        
# 314 "asllib/Parser.mly"
                                                                    E_Unop                
# 4668 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4677 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_135 =
    fun _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else ->
      let desc = 
# 315 "asllib/Parser.mly"
                                                                    E_Cond                
# 4685 "asllib/Parser.ml"
       (e1, e2, e_else) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_else_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4693 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_136 =
    fun _endpos_call_ _startpos_call_ call ->
      let desc = 
# 316 "asllib/Parser.mly"
                                                                    e_call                
# 4701 "asllib/Parser.ml"
       call in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_call_, _startpos_call_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4709 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_137 =
    fun _endpos_slices_ _startpos_e_ e slices ->
      let desc = 
# 317 "asllib/Parser.mly"
                                                                        E_Slice               
# 4717 "asllib/Parser.ml"
       (e, slices) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_slices_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4725 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_138 =
    fun _2 _4 _endpos__4_ _startpos_e1_ e1 e2 ->
      let desc = 
# 318 "asllib/Parser.mly"
                                                                        E_GetArray            
# 4733 "asllib/Parser.ml"
       (e1, e2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__4_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4741 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_139 =
    fun _2 _endpos_x_ _startpos_e_ e x ->
      let desc = 
# 319 "asllib/Parser.mly"
                                                                        E_GetField            
# 4749 "asllib/Parser.ml"
       (e, x) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_x_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4757 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_140 =
    fun _2 _endpos__3_ _startpos_e_ e x ->
      let desc =
        let fs = 
# 205 "<standard.mly>"
    ( x )
# 4766 "asllib/Parser.ml"
         in
        
# 320 "asllib/Parser.mly"
                                                                        E_GetFields           
# 4771 "asllib/Parser.ml"
         (e, fs)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4780 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_141 =
    fun _2 _endpos_ty_ _startpos_expr_pattern_ expr_pattern ty ->
      let desc = 
# 322 "asllib/Parser.mly"
                                                                        E_ATC                 
# 4788 "asllib/Parser.ml"
       (expr_pattern, ty) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4796 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_142 =
    fun _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_pattern_ constraint_kind expr_pattern ->
      let desc =
        let implicit_t_int =
          let desc = 
# 411 "asllib/Parser.mly"
                                                       T_Int 
# 4806 "asllib/Parser.ml"
           constraint_kind in
          let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_constraint_kind_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4814 "asllib/Parser.ml"
          
        in
        
# 323 "asllib/Parser.mly"
                                                                        E_ATC                 
# 4820 "asllib/Parser.ml"
         (expr_pattern, implicit_t_int)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4829 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_143 =
    fun _2 _endpos_pattern_set_ _startpos_expr_pattern_ expr_pattern pattern_set ->
      let desc = 
# 325 "asllib/Parser.mly"
                                                                        E_Pattern             
# 4837 "asllib/Parser.ml"
       (expr_pattern, pattern_set) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_pattern_set_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4845 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_144 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_pattern_ expr_pattern ->
      let desc =
        let pattern_mask =
          let desc =
            let _1_1 = _1 in
            
# 339 "asllib/Parser.mly"
                                            Pattern_Mask  
# 4857 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4866 "asllib/Parser.ml"
          
        in
        
# 326 "asllib/Parser.mly"
                                                                        E_Pattern             
# 4872 "asllib/Parser.ml"
         (expr_pattern, pattern_mask)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4881 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_145 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e ->
      let desc =
        let p =
          let desc =
            let _1_1 = _1 in
            
# 339 "asllib/Parser.mly"
                                            Pattern_Mask  
# 4893 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4902 "asllib/Parser.ml"
          
        in
        
# 327 "asllib/Parser.mly"
                                                                      ( E_Pattern (e, Pattern_Not (p) |> add_pos_from p) )
# 4908 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4917 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_146 =
    fun _1 _2 _endpos_ty_ _startpos__1_ ty ->
      let desc = 
# 329 "asllib/Parser.mly"
                                                                        E_Arbitrary         
# 4925 "asllib/Parser.ml"
       ty in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4933 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_147 =
    fun _endpos__3_ _startpos__1_ x ->
      let desc =
        let e = 
# 205 "<standard.mly>"
    ( x )
# 4942 "asllib/Parser.ml"
         in
        
# 330 "asllib/Parser.mly"
                                                                      ( E_Tuple [ e ]        )
# 4947 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4956 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_148 =
    fun _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc ->
      let desc =
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4968 "asllib/Parser.ml"
          
        in
        
# 334 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), []) )
# 4974 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4983 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_149 =
    fun _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let fields = 
# 205 "<standard.mly>"
    ( x )
# 4992 "asllib/Parser.ml"
         in
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5000 "asllib/Parser.ml"
          
        in
        
# 336 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), fields) )
# 5006 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5015 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_150 =
    fun x y ->
      (
# 180 "<standard.mly>"
    ( (x, y) )
# 5023 "asllib/Parser.ml"
       : (string * AST.expr))
  
  let _menhir_action_151 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 5031 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_152 =
    fun () ->
      (
# 361 "asllib/Parser.mly"
                  ( [] )
# 5039 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_153 =
    fun _1 ->
      (_1 : (AST.field list))
  
  let _menhir_action_154 =
    fun _1 ->
      (_1 : (AST.typed_identifier list))
  
  let _menhir_action_155 =
    fun _1_inlined1 _2 x ->
      let _3 =
        let _1 = _1_inlined1 in
        ()
      in
      (
# 205 "<standard.mly>"
    ( x )
# 5059 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_156 =
    fun _1_inlined1 _endpos__1_inlined1_ _startpos__1_inlined1_ x ->
      let _3 =
        let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
        let _endpos = _endpos__1_ in
        let _startpos = _startpos__1_ in
        
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 5074 "asllib/Parser.ml"
        
      in
      (
# 205 "<standard.mly>"
    ( x )
# 5080 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_157 =
    fun _1 ->
      (
# 477 "asllib/Parser.mly"
                ( GDK_Let      )
# 5088 "asllib/Parser.ml"
       : (AST.global_decl_keyword))
  
  let _menhir_action_158 =
    fun _1 ->
      (
# 478 "asllib/Parser.mly"
                ( GDK_Constant )
# 5096 "asllib/Parser.ml"
       : (AST.global_decl_keyword))
  
  let _menhir_action_159 =
    fun _1 ->
      (
# 479 "asllib/Parser.mly"
                ( GDK_Config   )
# 5104 "asllib/Parser.ml"
       : (AST.global_decl_keyword))
  
  let _menhir_action_160 =
    fun _1 ->
      (
# 595 "asllib/Parser.mly"
           ( global_ignored () )
# 5112 "asllib/Parser.ml"
       : (string))
  
  let _menhir_action_161 =
    fun _1 ->
      (_1 : (string))
  
  let _menhir_action_162 =
    fun expr ->
      (
# 303 "asllib/Parser.mly"
                                  Constraint_Exact  
# 5124 "asllib/Parser.ml"
       expr : (AST.int_constraint))
  
  let _menhir_action_163 =
    fun _2 e1 e2 ->
      (
# 304 "asllib/Parser.mly"
                                  Constraint_Range  
# 5132 "asllib/Parser.ml"
       (e1, e2) : (AST.int_constraint))
  
  let _menhir_action_164 =
    fun sliced_basic_lexpr ->
      (
# 444 "asllib/Parser.mly"
                            desugar_lhs_access  
# 5140 "asllib/Parser.ml"
       sliced_basic_lexpr : (AST.lexpr))
  
  let _menhir_action_165 =
    fun _endpos__3_ _startpos__1_ x ->
      let _1 =
        let desc = 
# 205 "<standard.mly>"
    ( x )
# 5149 "asllib/Parser.ml"
         in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5157 "asllib/Parser.ml"
        
      in
      let _1_1 = _1 in
      (
# 445 "asllib/Parser.mly"
                                                                 desugar_lhs_tuple  
# 5164 "asllib/Parser.ml"
       _1_1 : (AST.lexpr))
  
  let _menhir_action_166 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 447 "asllib/Parser.mly"
             ( LE_Discard )
# 5172 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5180 "asllib/Parser.ml"
       : (AST.lexpr))
  
  let _menhir_action_167 =
    fun _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let flds = 
# 205 "<standard.mly>"
    ( x )
# 5189 "asllib/Parser.ml"
         in
        let x =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5197 "asllib/Parser.ml"
          
        in
        
# 449 "asllib/Parser.mly"
      ( LE_SetFields (le_var x, flds, []) )
# 5203 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5212 "asllib/Parser.ml"
       : (AST.lexpr))
  
  let _menhir_action_168 =
    fun _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let flds = 
# 205 "<standard.mly>"
    ( x )
# 5221 "asllib/Parser.ml"
         in
        let x =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5229 "asllib/Parser.ml"
          
        in
        
# 451 "asllib/Parser.mly"
      ( desugar_lhs_fields_tuple x flds )
# 5235 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5244 "asllib/Parser.ml"
       : (AST.lexpr))
  
  let _menhir_action_169 =
    fun () ->
      (
# 216 "<standard.mly>"
    ( [] )
# 5252 "asllib/Parser.ml"
       : (AST.t))
  
  let _menhir_action_170 =
    fun x xs ->
      (
# 219 "<standard.mly>"
    ( x :: xs )
# 5260 "asllib/Parser.ml"
       : (AST.t))
  
  let _menhir_action_171 =
    fun _1 ->
      (
# 470 "asllib/Parser.mly"
                ( LDK_Let       )
# 5268 "asllib/Parser.ml"
       : (AST.local_decl_keyword))
  
  let _menhir_action_172 =
    fun _1 ->
      (
# 471 "asllib/Parser.mly"
                ( LDK_Constant  )
# 5276 "asllib/Parser.ml"
       : (AST.local_decl_keyword))
  
  let _menhir_action_173 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 5284 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_174 =
    fun _1 _2 ->
      let x = _2 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 5293 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_175 =
    fun _1 ->
      (_1 : (AST.stmt))
  
  let _menhir_action_176 =
    fun _endpos__0_ ->
      let desc = 
# 591 "asllib/Parser.mly"
                                                   ( S_Pass )
# 5305 "asllib/Parser.ml"
       in
      let _endpos_desc_ = _endpos__0_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _endpos in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5313 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_177 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5321 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_178 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5329 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_179 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5337 "asllib/Parser.ml"
       : (Desugar.lhs_field option list))
  
  let _menhir_action_180 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5345 "asllib/Parser.ml"
       : (Desugar.lhs_field option list))
  
  let _menhir_action_181 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5353 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_182 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5361 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_183 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5369 "asllib/Parser.ml"
       : (Desugar.lhs_access option list))
  
  let _menhir_action_184 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5377 "asllib/Parser.ml"
       : (Desugar.lhs_access option list))
  
  let _menhir_action_185 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5385 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_186 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5393 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_187 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5401 "asllib/Parser.ml"
       : ((string * AST.expr) list))
  
  let _menhir_action_188 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5409 "asllib/Parser.ml"
       : ((string * AST.expr) list))
  
  let _menhir_action_189 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5417 "asllib/Parser.ml"
       : (AST.int_constraint list))
  
  let _menhir_action_190 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5425 "asllib/Parser.ml"
       : (AST.int_constraint list))
  
  let _menhir_action_191 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5433 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_192 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5441 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_193 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5449 "asllib/Parser.ml"
       : (AST.pattern list))
  
  let _menhir_action_194 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5457 "asllib/Parser.ml"
       : (AST.pattern list))
  
  let _menhir_action_195 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5465 "asllib/Parser.ml"
       : (AST.slice list))
  
  let _menhir_action_196 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5473 "asllib/Parser.ml"
       : (AST.slice list))
  
  let _menhir_action_197 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5481 "asllib/Parser.ml"
       : (AST.ty list))
  
  let _menhir_action_198 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5489 "asllib/Parser.ml"
       : (AST.ty list))
  
  let _menhir_action_199 =
    fun x ->
      (
# 150 "asllib/Parser.mly"
         ( [ x ] )
# 5497 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_200 =
    fun _2 h t ->
      (
# 151 "asllib/Parser.mly"
                             ( h :: t )
# 5505 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_201 =
    fun () ->
      (
# 428 "asllib/Parser.mly"
    ( [] )
# 5513 "asllib/Parser.ml"
       : (Desugar.lhs_field list))
  
  let _menhir_action_202 =
    fun _1 _endpos_desc_ _startpos_desc_ desc t ->
      let h =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5524 "asllib/Parser.ml"
        
      in
      (
# 429 "asllib/Parser.mly"
                                                   ( h :: t )
# 5530 "asllib/Parser.ml"
       : (Desugar.lhs_field list))
  
  let _menhir_action_203 =
    fun x ->
      (
# 184 "asllib/Parser.mly"
         ( [ x ] )
# 5538 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated list))
  
  let _menhir_action_204 =
    fun l x ->
      (
# 185 "asllib/Parser.mly"
                     ( x :: l )
# 5546 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated list))
  
  let _menhir_action_205 =
    fun x ->
      (
# 184 "asllib/Parser.mly"
         ( [ x ] )
# 5554 "asllib/Parser.ml"
       : (AST.catcher list))
  
  let _menhir_action_206 =
    fun l x ->
      (
# 185 "asllib/Parser.mly"
                     ( x :: l )
# 5562 "asllib/Parser.ml"
       : (AST.catcher list))
  
  let _menhir_action_207 =
    fun x ->
      (
# 184 "asllib/Parser.mly"
         ( [ x ] )
# 5570 "asllib/Parser.ml"
       : (AST.stmt list))
  
  let _menhir_action_208 =
    fun l x ->
      (
# 185 "asllib/Parser.mly"
                     ( x :: l )
# 5578 "asllib/Parser.ml"
       : (AST.stmt list))
  
  let _menhir_action_209 =
    fun x ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 5586 "asllib/Parser.ml"
       in
      (
# 164 "asllib/Parser.mly"
                         ( [ x ] )
# 5591 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_210 =
    fun x x_inlined1 ->
      let _2 =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 5601 "asllib/Parser.ml"
        
      in
      (
# 164 "asllib/Parser.mly"
                         ( [ x ] )
# 5607 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_211 =
    fun _2 h t ->
      (
# 165 "asllib/Parser.mly"
                              ( h :: t )
# 5615 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_212 =
    fun x ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 5623 "asllib/Parser.ml"
       in
      (
# 164 "asllib/Parser.mly"
                         ( [ x ] )
# 5628 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_213 =
    fun x x_inlined1 ->
      let _2 =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 5638 "asllib/Parser.ml"
        
      in
      (
# 164 "asllib/Parser.mly"
                         ( [ x ] )
# 5644 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_214 =
    fun _2 h t ->
      (
# 165 "asllib/Parser.mly"
                              ( h :: t )
# 5652 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_215 =
    fun x ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 5660 "asllib/Parser.ml"
       in
      (
# 164 "asllib/Parser.mly"
                         ( [ x ] )
# 5665 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_216 =
    fun x x_inlined1 ->
      let _2 =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 5675 "asllib/Parser.ml"
        
      in
      (
# 164 "asllib/Parser.mly"
                         ( [ x ] )
# 5681 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_217 =
    fun _2 h t ->
      (
# 165 "asllib/Parser.mly"
                              ( h :: t )
# 5689 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_218 =
    fun _2 body ->
      (
# 694 "asllib/Parser.mly"
    (
      [
        D_Func
          {
            name = "main";
            args = [];
            parameters = [];
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            builtin = false;
          }
        |> ASTUtils.add_pos_from body
      ]
    )
# 5712 "asllib/Parser.ml"
       : (AST.t))
  
  let _menhir_action_219 =
    fun x ->
      let y = 
# 123 "<standard.mly>"
    ( None )
# 5720 "asllib/Parser.ml"
       in
      (
# 172 "<standard.mly>"
    ( (x, y) )
# 5725 "asllib/Parser.ml"
       : (string * AST.ty option))
  
  let _menhir_action_220 =
    fun x x_inlined1 ->
      let y =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 5735 "asllib/Parser.ml"
        
      in
      (
# 172 "<standard.mly>"
    ( (x, y) )
# 5741 "asllib/Parser.ml"
       : (string * AST.ty option))
  
  let _menhir_action_221 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 5749 "asllib/Parser.ml"
       : (AST.ty option))
  
  let _menhir_action_222 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 5757 "asllib/Parser.ml"
       : (AST.ty option))
  
  let _menhir_action_223 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 5765 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_224 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 5773 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_225 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 5781 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_226 =
    fun _1 _2 ->
      let x = _2 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 5790 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_227 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 5798 "asllib/Parser.ml"
       : (AST.stmt option))
  
  let _menhir_action_228 =
    fun _1 _2 _3 ->
      let x = _3 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 5807 "asllib/Parser.ml"
       : (AST.stmt option))
  
  let _menhir_action_229 =
    fun () ->
      (
# 577 "asllib/Parser.mly"
                  ( [] )
# 5815 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_230 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 5823 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_231 =
    fun _endpos_expr_pattern_ _startpos_expr_pattern_ expr_pattern ->
      let desc = 
# 344 "asllib/Parser.mly"
                        Pattern_Single  
# 5831 "asllib/Parser.ml"
       expr_pattern in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_expr_pattern_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5839 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_232 =
    fun _2 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc = 
# 345 "asllib/Parser.mly"
                                           Pattern_Range  
# 5847 "asllib/Parser.ml"
       (e1, e2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5855 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_233 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 346 "asllib/Parser.mly"
             ( Pattern_All )
# 5863 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5871 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_234 =
    fun _1 _endpos_expr_ _startpos__1_ expr ->
      let desc = 
# 347 "asllib/Parser.mly"
                     Pattern_Leq  
# 5879 "asllib/Parser.ml"
       expr in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_expr_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5887 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_235 =
    fun _1 _endpos_expr_ _startpos__1_ expr ->
      let desc = 
# 348 "asllib/Parser.mly"
                     Pattern_Geq  
# 5895 "asllib/Parser.ml"
       expr in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_expr_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5903 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_236 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 349 "asllib/Parser.mly"
                           Pattern_Tuple  
# 5913 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5922 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_237 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 339 "asllib/Parser.mly"
                                            Pattern_Mask  
# 5932 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5941 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_238 =
    fun _1 ->
      (_1 : (AST.pattern))
  
  let _menhir_action_239 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 340 "asllib/Parser.mly"
                                                   Pattern_Any  
# 5955 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5964 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_240 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 5972 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_241 =
    fun _1 _endpos__3_ _startpos__1_ x ->
      let desc =
        let _2 = 
# 205 "<standard.mly>"
    ( x )
# 5981 "asllib/Parser.ml"
         in
        let _2_1 = _2 in
        
# 357 "asllib/Parser.mly"
                                      Pattern_Not  
# 5987 "asllib/Parser.ml"
         _2_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5996 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_242 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 6004 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_243 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 6012 "asllib/Parser.ml"
       : (AST.ty list))
  
  let _menhir_action_244 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 6020 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_245 =
    fun _2 li x ->
      let x = 
# 172 "asllib/Parser.mly"
                            ( x :: li )
# 6028 "asllib/Parser.ml"
       in
      (
# 205 "<standard.mly>"
    ( x )
# 6033 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_246 =
    fun _2 li x ->
      let x = 
# 172 "asllib/Parser.mly"
                            ( x :: li )
# 6041 "asllib/Parser.ml"
       in
      (
# 205 "<standard.mly>"
    ( x )
# 6046 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_247 =
    fun _2 li x ->
      let x = 
# 172 "asllib/Parser.mly"
                            ( x :: li )
# 6054 "asllib/Parser.ml"
       in
      (
# 205 "<standard.mly>"
    ( x )
# 6059 "asllib/Parser.ml"
       : (AST.pattern list))
  
  let _menhir_action_248 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 6067 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_249 =
    fun _1 _2 ->
      let x = _2 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 6076 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_250 =
    fun _1 _2 ->
      (_2 : (AST.ty))
  
  let _menhir_action_251 =
    fun _1 _3 _endpos_s2_ _startpos__1_ e s1 s2 ->
      let desc = 
# 559 "asllib/Parser.mly"
                                                     S_Cond 
# 6088 "asllib/Parser.ml"
       (e, s1, s2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_s2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = if _startpos_desc_ != _endpos_desc_ then
        _startpos_desc_
      else
        _endpos in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6099 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_252 =
    fun _endpos__0_ ->
      let desc = 
# 484 "asllib/Parser.mly"
            ( S_Pass )
# 6107 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__0_, _endpos__0_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = if _startpos_desc_ != _endpos_desc_ then
        _startpos_desc_
      else
        _endpos in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6118 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_253 =
    fun _1 _2 ->
      (_2 : (AST.stmt))
  
  let _menhir_action_254 =
    fun expr ->
      (
# 366 "asllib/Parser.mly"
                                    Slice_Single   
# 6130 "asllib/Parser.ml"
       expr : (AST.slice))
  
  let _menhir_action_255 =
    fun _2 e1 e2 ->
      (
# 367 "asllib/Parser.mly"
                                    Slice_Range    
# 6138 "asllib/Parser.ml"
       (e1, e2) : (AST.slice))
  
  let _menhir_action_256 =
    fun _2 e1 e2 ->
      (
# 368 "asllib/Parser.mly"
                                    Slice_Length   
# 6146 "asllib/Parser.ml"
       (e1, e2) : (AST.slice))
  
  let _menhir_action_257 =
    fun _1 e ->
      (
# 369 "asllib/Parser.mly"
                                  ( Slice_Length(zero, e) )
# 6154 "asllib/Parser.ml"
       : (AST.slice))
  
  let _menhir_action_258 =
    fun _2 e1 e2 ->
      (
# 370 "asllib/Parser.mly"
                                    Slice_Star     
# 6162 "asllib/Parser.ml"
       (e1, e2) : (AST.slice))
  
  let _menhir_action_259 =
    fun basic_lexpr ->
      (basic_lexpr : (Desugar.lhs_access))
  
  let _menhir_action_260 =
    fun _endpos_desc_ _startpos_desc_ basic desc ->
      let slices =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6177 "asllib/Parser.ml"
        
      in
      (
# 433 "asllib/Parser.mly"
                                                 ( {basic with slices} )
# 6183 "asllib/Parser.ml"
       : (Desugar.lhs_access))
  
  let _menhir_action_261 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 6191 "asllib/Parser.ml"
       : (AST.slice list))
  
  let _menhir_action_262 =
    fun x ->
      (
# 196 "<standard.mly>"
    ( x )
# 6199 "asllib/Parser.ml"
       : (AST.t))
  
  let _menhir_action_263 =
    fun _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ e s1 s2 ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 510 "asllib/Parser.mly"
                                                       S_Cond 
# 6212 "asllib/Parser.ml"
         (e, s1, s2) in
        
# 196 "<standard.mly>"
    ( x )
# 6217 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6226 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_264 =
    fun _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ e s1 s2 ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 6242 "asllib/Parser.ml"
          
        in
        let x = 
# 510 "asllib/Parser.mly"
                                                       S_Cond 
# 6248 "asllib/Parser.ml"
         (e, s1, s2) in
        
# 196 "<standard.mly>"
    ( x )
# 6253 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6262 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_265 =
    fun _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ alt expr ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 512 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt (S_Unreachable |> ASTUtils.add_pos_from expr))
# 6275 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 6280 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6289 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_266 =
    fun _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ alt expr ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 6305 "asllib/Parser.ml"
          
        in
        let x = 
# 512 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt (S_Unreachable |> ASTUtils.add_pos_from expr))
# 6311 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 6316 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6325 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_267 =
    fun _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ alt case_otherwise expr ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 514 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt case_otherwise )
# 6338 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 6343 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6352 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_268 =
    fun _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ alt case_otherwise expr ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 6368 "asllib/Parser.ml"
          
        in
        let x = 
# 514 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt case_otherwise )
# 6374 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 6379 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6388 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_269 =
    fun _1 _1_inlined1 _2 _4 _endpos__2_ _startpos__1_ expr loop_limit stmt_list ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 515 "asllib/Parser.mly"
                                                       S_While 
# 6401 "asllib/Parser.ml"
         (expr, loop_limit, stmt_list) in
        
# 196 "<standard.mly>"
    ( x )
# 6406 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6415 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_270 =
    fun _1 _1_inlined1 _4 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ expr loop_limit stmt_list ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 6431 "asllib/Parser.ml"
          
        in
        let x = 
# 515 "asllib/Parser.mly"
                                                       S_While 
# 6437 "asllib/Parser.ml"
         (expr, loop_limit, stmt_list) in
        
# 196 "<standard.mly>"
    ( x )
# 6442 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6451 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_271 =
    fun _1 _1_inlined1 _2 _3 _8 _endpos__2_ _startpos__1_ body dir end_e index_name limit start_e ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 518 "asllib/Parser.mly"
          ( S_For { index_name; start_e; end_e; dir; body; limit } )
# 6464 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 6469 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6478 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_272 =
    fun _1 _1_inlined1 _3 _8 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ body dir end_e index_name limit start_e ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 6494 "asllib/Parser.ml"
          
        in
        let x = 
# 518 "asllib/Parser.mly"
          ( S_For { index_name; start_e; end_e; dir; body; limit } )
# 6500 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 6505 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6514 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_273 =
    fun _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ c o s ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 519 "asllib/Parser.mly"
                                                                      S_Try  
# 6527 "asllib/Parser.ml"
         (s, c, o) in
        
# 196 "<standard.mly>"
    ( x )
# 6532 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6541 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_274 =
    fun _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ c o s ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 6557 "asllib/Parser.ml"
          
        in
        let x = 
# 519 "asllib/Parser.mly"
                                                                      S_Try  
# 6563 "asllib/Parser.ml"
         (s, c, o) in
        
# 196 "<standard.mly>"
    ( x )
# 6568 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6577 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_275 =
    fun _1 _1_inlined2 _2 _3 _5 _endpos__2_ _startpos__1_ body cond x ->
      let desc =
        let _2 =
          let _1 = _1_inlined2 in
          ()
        in
        let x =
          let looplimit = 
# 205 "<standard.mly>"
    ( x )
# 6591 "asllib/Parser.ml"
           in
          
# 521 "asllib/Parser.mly"
          ( S_While (cond, Some looplimit, body) )
# 6596 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 6602 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6611 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_276 =
    fun _1 _1_inlined2 _3 _5 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ body cond x ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 189 "asllib/Parser.mly"
         (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 6627 "asllib/Parser.ml"
          
        in
        let x =
          let looplimit = 
# 205 "<standard.mly>"
    ( x )
# 6634 "asllib/Parser.ml"
           in
          
# 521 "asllib/Parser.mly"
          ( S_While (cond, Some looplimit, body) )
# 6639 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 6645 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6654 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_277 =
    fun _1 _endpos__2_ _startpos__1_ ->
      let desc =
        let x = 
# 484 "asllib/Parser.mly"
            ( S_Pass )
# 6663 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 6668 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6677 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_278 =
    fun _1 _2_inlined1 _endpos__2_ _startpos__1_ ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let _2_1 = _2 in
          
# 525 "asllib/Parser.mly"
                                                              S_Return  
# 6689 "asllib/Parser.ml"
           _2_1
        in
        
# 196 "<standard.mly>"
    ( x )
# 6695 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6704 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_279 =
    fun _endpos__2_ _startpos_call_ call ->
      let desc =
        let x = 
# 526 "asllib/Parser.mly"
                                                               s_call  
# 6713 "asllib/Parser.ml"
         call in
        
# 196 "<standard.mly>"
    ( x )
# 6718 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_call_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6727 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_280 =
    fun _1 _endpos__2_ _startpos__1_ e ->
      let desc =
        let x = 
# 527 "asllib/Parser.mly"
                                                               S_Assert  
# 6736 "asllib/Parser.ml"
         e in
        
# 196 "<standard.mly>"
    ( x )
# 6741 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6750 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_281 =
    fun _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x ->
      let desc =
        let x =
          let _5 = 
# 136 "asllib/Parser.mly"
                        Some 
# 6760 "asllib/Parser.ml"
           x in
          let ty_opt = 
# 123 "<standard.mly>"
    ( None )
# 6765 "asllib/Parser.ml"
           in
          let _5_1 = _5 in
          
# 528 "asllib/Parser.mly"
                                                                                 S_Decl    
# 6771 "asllib/Parser.ml"
           (local_decl_keyword_non_var, decl_item, ty_opt, _5_1)
        in
        
# 196 "<standard.mly>"
    ( x )
# 6777 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_local_decl_keyword_non_var_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6786 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_282 =
    fun _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x x_inlined1 ->
      let desc =
        let x =
          let _5 =
            let x = x_inlined1 in
            
# 136 "asllib/Parser.mly"
                        Some 
# 6798 "asllib/Parser.ml"
             x
          in
          let ty_opt = 
# 126 "<standard.mly>"
    ( Some x )
# 6804 "asllib/Parser.ml"
           in
          let _5_1 = _5 in
          
# 528 "asllib/Parser.mly"
                                                                                 S_Decl    
# 6810 "asllib/Parser.ml"
           (local_decl_keyword_non_var, decl_item, ty_opt, _5_1)
        in
        
# 196 "<standard.mly>"
    ( x )
# 6816 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_local_decl_keyword_non_var_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6825 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_283 =
    fun _2_inlined1 _endpos__2_ _startpos_le_ e le ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          
# 529 "asllib/Parser.mly"
                                                               S_Assign  
# 6836 "asllib/Parser.ml"
           (le, e)
        in
        
# 196 "<standard.mly>"
    ( x )
# 6842 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_le_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6851 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_284 =
    fun _2_inlined1 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6865 "asllib/Parser.ml"
            
          in
          
# 531 "asllib/Parser.mly"
        ( desugar_setter call [] rhs )
# 6871 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 6877 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__2_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6886 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_285 =
    fun _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc fld rhs ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6900 "asllib/Parser.ml"
            
          in
          
# 533 "asllib/Parser.mly"
        ( desugar_setter call [fld] rhs )
# 6906 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 6912 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__2_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6921 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_286 =
    fun _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs x ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let flds = 
# 205 "<standard.mly>"
    ( x )
# 6932 "asllib/Parser.ml"
           in
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6940 "asllib/Parser.ml"
            
          in
          
# 535 "asllib/Parser.mly"
        ( desugar_setter call flds rhs )
# 6946 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 6952 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__2_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6961 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_287 =
    fun _4 _endpos__2_ _endpos_desc_ _startpos_desc_ _startpos_ldk_ desc ldk lhs ty ->
      let desc =
        let x =
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6974 "asllib/Parser.ml"
            
          in
          
# 537 "asllib/Parser.mly"
        ( desugar_elided_parameter ldk lhs ty call)
# 6980 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 6986 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_ldk_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6995 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_288 =
    fun _1 _endpos__2_ _startpos__1_ e ldi ->
      let desc =
        let x =
          let ty = 
# 123 "<standard.mly>"
    ( None )
# 7005 "asllib/Parser.ml"
           in
          
# 538 "asllib/Parser.mly"
                                                              ( S_Decl (LDK_Var, ldi, ty, e) )
# 7010 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7016 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7025 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_289 =
    fun _1 _endpos__2_ _startpos__1_ e ldi x ->
      let desc =
        let x =
          let ty = 
# 126 "<standard.mly>"
    ( Some x )
# 7035 "asllib/Parser.ml"
           in
          
# 538 "asllib/Parser.mly"
                                                              ( S_Decl (LDK_Var, ldi, ty, e) )
# 7040 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7046 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7055 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_290 =
    fun _1 _2_inlined1 _endpos__2_ _startpos__1_ as_ty ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let _2_1 = _2 in
          
# 539 "asllib/Parser.mly"
                                                               make_ldi_vars  
# 7067 "asllib/Parser.ml"
           (_2_1, as_ty)
        in
        
# 196 "<standard.mly>"
    ( x )
# 7073 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7082 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_291 =
    fun _1 _4 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc lhs ty ->
      let desc =
        let x =
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7095 "asllib/Parser.ml"
            
          in
          
# 541 "asllib/Parser.mly"
        ( desugar_elided_parameter LDK_Var lhs ty call)
# 7101 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7107 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7116 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_292 =
    fun _1 _endpos__2_ _startpos__1_ args ->
      let desc =
        let x = 
# 542 "asllib/Parser.mly"
                                                             ( S_Print { args; newline = true; debug = false } )
# 7125 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7130 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7139 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_293 =
    fun _1 _endpos__2_ _startpos__1_ args ->
      let desc =
        let x = 
# 543 "asllib/Parser.mly"
                                                             ( S_Print { args; newline = false; debug = false } )
# 7148 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7153 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7162 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_294 =
    fun _1 _endpos__2_ _startpos__1_ args ->
      let desc =
        let x = 
# 544 "asllib/Parser.mly"
                                                             ( S_Print { args; newline = true; debug = true } )
# 7171 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7176 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7185 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_295 =
    fun _1 _2_inlined1 _3 _endpos__2_ _startpos__1_ ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          
# 545 "asllib/Parser.mly"
                                                             ( S_Unreachable )
# 7196 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7202 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7211 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_296 =
    fun _1 _3 _endpos__2_ _startpos__1_ expr loop_limit stmt_list ->
      let desc =
        let x = 
# 546 "asllib/Parser.mly"
                                                               S_Repeat  
# 7220 "asllib/Parser.ml"
         (stmt_list, expr, loop_limit) in
        
# 196 "<standard.mly>"
    ( x )
# 7225 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7234 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_297 =
    fun _1 _3 _5 _endpos__2_ _startpos__1_ body cond x ->
      let desc =
        let x =
          let looplimit = 
# 205 "<standard.mly>"
    ( x )
# 7244 "asllib/Parser.ml"
           in
          
# 548 "asllib/Parser.mly"
          ( S_Repeat (body, cond, Some looplimit) )
# 7249 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7255 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7264 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_298 =
    fun _1 _endpos__2_ _startpos__1_ e ->
      let desc =
        let x = 
# 549 "asllib/Parser.mly"
                                                             ( S_Throw (Some (e, None)) )
# 7273 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7278 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7287 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_299 =
    fun _1 _endpos__2_ _startpos__1_ ->
      let desc =
        let x = 
# 550 "asllib/Parser.mly"
                                                             ( S_Throw None             )
# 7296 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7301 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7310 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_300 =
    fun _1 _endpos__2_ _startpos__1_ e x ->
      let desc =
        let x = 
# 551 "asllib/Parser.mly"
                                                               S_Pragma  
# 7319 "asllib/Parser.ml"
         (x, e) in
        
# 196 "<standard.mly>"
    ( x )
# 7324 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7333 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_301 =
    fun _1 ->
      let _1_1 = _1 in
      (
# 555 "asllib/Parser.mly"
                                    stmt_from_list 
# 7342 "asllib/Parser.ml"
       _1_1 : (AST.stmt))
  
  let _menhir_action_302 =
    fun x ->
      (
# 196 "<standard.mly>"
    ( x )
# 7350 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_303 =
    fun _1 _2 ->
      let with_opt = 
# 570 "asllib/Parser.mly"
                ( [] )
# 7358 "asllib/Parser.ml"
       in
      let _2_1 = _2 in
      ((_2_1, with_opt) : (string * AST.field list))
  
  let _menhir_action_304 =
    fun _1 _1_inlined1 _2 fields ->
      let with_opt =
        let _1 = _1_inlined1 in
        fields
      in
      let _2_1 = _2 in
      ((_2_1, with_opt) : (string * AST.field list))
  
  let _menhir_action_305 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 7377 "asllib/Parser.ml"
       : ((string * AST.field list) option))
  
  let _menhir_action_306 =
    fun x ->
      (
# 126 "<standard.mly>"
    ( Some x )
# 7385 "asllib/Parser.ml"
       : ((string * AST.field list) option))
  
  let _menhir_action_307 =
    fun () ->
      (
# 160 "asllib/Parser.mly"
                 ( [] )
# 7393 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_308 =
    fun _1 ->
      (_1 : (AST.bitfield list))
  
  let _menhir_action_309 =
    fun () ->
      (
# 160 "asllib/Parser.mly"
                 ( [] )
# 7405 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_310 =
    fun _1 ->
      (_1 : (AST.field list))
  
  let _menhir_action_311 =
    fun _1 _endpos_c_ _startpos__1_ c ->
      let desc = 
# 383 "asllib/Parser.mly"
                                                          T_Int        
# 7417 "asllib/Parser.ml"
       c in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_c_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7425 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_312 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 384 "asllib/Parser.mly"
                                                        ( T_Real      )
# 7433 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7441 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_313 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 385 "asllib/Parser.mly"
                                                        ( T_Bool      )
# 7449 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7457 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_314 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 386 "asllib/Parser.mly"
                                                        ( T_String    )
# 7465 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7473 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_315 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 387 "asllib/Parser.mly"
                                                        ( t_bit       )
# 7481 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7489 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_316 =
    fun _1 _endpos_bitfields_opt_ _startpos__1_ bitfields_opt x ->
      let desc =
        let _2 = 
# 205 "<standard.mly>"
    ( x )
# 7498 "asllib/Parser.ml"
         in
        let _2_1 = _2 in
        
# 388 "asllib/Parser.mly"
                                                          T_Bits       
# 7504 "asllib/Parser.ml"
         (_2_1, bitfields_opt)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_bitfields_opt_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7513 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_317 =
    fun _endpos_l_ _startpos_l_ l ->
      let desc = 
# 389 "asllib/Parser.mly"
                                                          T_Tuple      
# 7521 "asllib/Parser.ml"
       l in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_l_, _startpos_l_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7529 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_318 =
    fun _endpos_name_ _startpos_name_ name ->
      let desc = 
# 390 "asllib/Parser.mly"
                                                          T_Named      
# 7537 "asllib/Parser.ml"
       name in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_name_, _startpos_name_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7545 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_319 =
    fun _1 _2 _4 _5 _endpos_t_ _startpos__1_ e t ->
      let desc = 
# 391 "asllib/Parser.mly"
                                                        ( T_Array (ArrayLength_Expr e, t) )
# 7553 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_t_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7561 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_320 =
    fun _1 ->
      (_1 : (AST.ty))
  
  let _menhir_action_321 =
    fun _1 _endpos__3_ _startpos__1_ x ->
      let desc =
        let l = 
# 205 "<standard.mly>"
    ( x )
# 7574 "asllib/Parser.ml"
         in
        
# 396 "asllib/Parser.mly"
                                                          T_Enum       
# 7579 "asllib/Parser.ml"
         l
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7588 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_322 =
    fun _1 _endpos_l_ _startpos__1_ l ->
      let desc = 
# 397 "asllib/Parser.mly"
                                                          T_Record     
# 7596 "asllib/Parser.ml"
       l in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_l_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7604 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_323 =
    fun _1 _endpos_l_ _startpos__1_ l ->
      let desc = 
# 398 "asllib/Parser.mly"
                                                          T_Exception  
# 7612 "asllib/Parser.ml"
       l in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_l_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 143 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7620 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_324 =
    fun x y ->
      (
# 172 "<standard.mly>"
    ( (x, y) )
# 7628 "asllib/Parser.ml"
       : (AST.field))
  
  let _menhir_action_325 =
    fun i ->
      (
# 202 "asllib/Parser.mly"
                        L_Int          
# 7636 "asllib/Parser.ml"
       i : (AST.literal))
  
  let _menhir_action_326 =
    fun b ->
      (
# 203 "asllib/Parser.mly"
                        L_Bool         
# 7644 "asllib/Parser.ml"
       b : (AST.literal))
  
  let _menhir_action_327 =
    fun r ->
      (
# 204 "asllib/Parser.mly"
                        L_Real         
# 7652 "asllib/Parser.ml"
       r : (AST.literal))
  
  let _menhir_action_328 =
    fun b ->
      (
# 205 "asllib/Parser.mly"
                        L_BitVector    
# 7660 "asllib/Parser.ml"
       b : (AST.literal))
  
  let _menhir_action_329 =
    fun s ->
      (
# 206 "asllib/Parser.mly"
                        L_String       
# 7668 "asllib/Parser.ml"
       s : (AST.literal))
  
  let _menhir_print_token : token -> string =
    fun _tok ->
      match _tok with
      | Tokens.AND ->
          "AND"
      | Tokens.ARBITRARY ->
          "ARBITRARY"
      | Tokens.ARRAY ->
          "ARRAY"
      | Tokens.ARROBASE_LOOPLIMIT ->
          "ARROBASE_LOOPLIMIT"
      | Tokens.ARROW ->
          "ARROW"
      | Tokens.AS ->
          "AS"
      | Tokens.ASSERT ->
          "ASSERT"
      | Tokens.BAND ->
          "BAND"
      | Tokens.BEGIN ->
          "BEGIN"
      | Tokens.BEQ ->
          "BEQ"
      | Tokens.BIT ->
          "BIT"
      | Tokens.BITS ->
          "BITS"
      | Tokens.BITVECTOR_LIT _ ->
          "BITVECTOR_LIT"
      | Tokens.BNOT ->
          "BNOT"
      | Tokens.BOOLEAN ->
          "BOOLEAN"
      | Tokens.BOOL_LIT _ ->
          "BOOL_LIT"
      | Tokens.BOR ->
          "BOR"
      | Tokens.CASE ->
          "CASE"
      | Tokens.CATCH ->
          "CATCH"
      | Tokens.COLON ->
          "COLON"
      | Tokens.COLON_COLON ->
          "COLON_COLON"
      | Tokens.COMMA ->
          "COMMA"
      | Tokens.CONCAT ->
          "CONCAT"
      | Tokens.CONFIG ->
          "CONFIG"
      | Tokens.CONSTANT ->
          "CONSTANT"
      | Tokens.DEBUG ->
          "DEBUG"
      | Tokens.DIV ->
          "DIV"
      | Tokens.DIVRM ->
          "DIVRM"
      | Tokens.DO ->
          "DO"
      | Tokens.DOT ->
          "DOT"
      | Tokens.DOWNTO ->
          "DOWNTO"
      | Tokens.ELSE ->
          "ELSE"
      | Tokens.ELSIF ->
          "ELSIF"
      | Tokens.END ->
          "END"
      | Tokens.ENUMERATION ->
          "ENUMERATION"
      | Tokens.EOF ->
          "EOF"
      | Tokens.EOR ->
          "EOR"
      | Tokens.EQ ->
          "EQ"
      | Tokens.EQ_OP ->
          "EQ_OP"
      | Tokens.EXCEPTION ->
          "EXCEPTION"
      | Tokens.FOR ->
          "FOR"
      | Tokens.FUNC ->
          "FUNC"
      | Tokens.GEQ ->
          "GEQ"
      | Tokens.GETTER ->
          "GETTER"
      | Tokens.GT ->
          "GT"
      | Tokens.IDENTIFIER _ ->
          "IDENTIFIER"
      | Tokens.IF ->
          "IF"
      | Tokens.IMPL ->
          "IMPL"
      | Tokens.IN ->
          "IN"
      | Tokens.INTEGER ->
          "INTEGER"
      | Tokens.INT_LIT _ ->
          "INT_LIT"
      | Tokens.LBRACE ->
          "LBRACE"
      | Tokens.LBRACKET ->
          "LBRACKET"
      | Tokens.LEQ ->
          "LEQ"
      | Tokens.LET ->
          "LET"
      | Tokens.LLBRACKET ->
          "LLBRACKET"
      | Tokens.LOOPLIMIT ->
          "LOOPLIMIT"
      | Tokens.LPAR ->
          "LPAR"
      | Tokens.LT ->
          "LT"
      | Tokens.MASK_LIT _ ->
          "MASK_LIT"
      | Tokens.MINUS ->
          "MINUS"
      | Tokens.MOD ->
          "MOD"
      | Tokens.MUL ->
          "MUL"
      | Tokens.NEQ ->
          "NEQ"
      | Tokens.NOT ->
          "NOT"
      | Tokens.OF ->
          "OF"
      | Tokens.OR ->
          "OR"
      | Tokens.OTHERWISE ->
          "OTHERWISE"
      | Tokens.PASS ->
          "PASS"
      | Tokens.PLUS ->
          "PLUS"
      | Tokens.PLUS_COLON ->
          "PLUS_COLON"
      | Tokens.POW ->
          "POW"
      | Tokens.PRAGMA ->
          "PRAGMA"
      | Tokens.PRINT ->
          "PRINT"
      | Tokens.PRINTLN ->
          "PRINTLN"
      | Tokens.RBRACE ->
          "RBRACE"
      | Tokens.RBRACKET ->
          "RBRACKET"
      | Tokens.RDIV ->
          "RDIV"
      | Tokens.REAL ->
          "REAL"
      | Tokens.REAL_LIT _ ->
          "REAL_LIT"
      | Tokens.RECORD ->
          "RECORD"
      | Tokens.RECURSELIMIT ->
          "RECURSELIMIT"
      | Tokens.REPEAT ->
          "REPEAT"
      | Tokens.RETURN ->
          "RETURN"
      | Tokens.RPAR ->
          "RPAR"
      | Tokens.RRBRACKET ->
          "RRBRACKET"
      | Tokens.SEMI_COLON ->
          "SEMI_COLON"
      | Tokens.SETTER ->
          "SETTER"
      | Tokens.SHL ->
          "SHL"
      | Tokens.SHR ->
          "SHR"
      | Tokens.SLICING ->
          "SLICING"
      | Tokens.STAR_COLON ->
          "STAR_COLON"
      | Tokens.STRING ->
          "STRING"
      | Tokens.STRING_LIT _ ->
          "STRING_LIT"
      | Tokens.SUBTYPES ->
          "SUBTYPES"
      | Tokens.THEN ->
          "THEN"
      | Tokens.THROW ->
          "THROW"
      | Tokens.TO ->
          "TO"
      | Tokens.TRY ->
          "TRY"
      | Tokens.TYPE ->
          "TYPE"
      | Tokens.UNREACHABLE ->
          "UNREACHABLE"
      | Tokens.UNTIL ->
          "UNTIL"
      | Tokens.VAR ->
          "VAR"
      | Tokens.WHEN ->
          "WHEN"
      | Tokens.WHERE ->
          "WHERE"
      | Tokens.WHILE ->
          "WHILE"
      | Tokens.WITH ->
          "WITH"
  
  let _menhir_fail : unit -> 'a =
    fun () ->
      Printf.eprintf "Internal failure -- please contact the parser generator's developers.\n%!";
      assert false
  
  include struct
    
    [@@@ocaml.warning "-4-37"]
    
    let _menhir_run_660 : type  ttv_stack. ttv_stack -> _ -> _menhir_box_spec =
      fun _menhir_stack _v ->
        let x = _v in
        let _v = _menhir_action_262 x in
        MenhirBox_spec _v
    
    let rec _menhir_run_670 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_decl -> _ -> _menhir_box_spec =
      fun _menhir_stack _v ->
        let MenhirCell1_decl (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_170 x xs in
        _menhir_goto_list_decl_ _menhir_stack _v _menhir_s
    
    and _menhir_goto_list_decl_ : type  ttv_stack. ttv_stack -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _v _menhir_s ->
        match _menhir_s with
        | MenhirState669 ->
            _menhir_run_670 _menhir_stack _v
        | MenhirState560 ->
            _menhir_run_660 _menhir_stack _v
        | _ ->
            _menhir_fail ()
    
    let _menhir_run_557 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_opn =
      fun _menhir_stack _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.EOF ->
            let (_2, body) = ((), _v) in
            let _v = _menhir_action_218 _2 body in
            MenhirBox_opn _v
        | _ ->
            _eRR ()
    
    let _menhir_run_673 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_stmts =
      fun _menhir_stack _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.EOF ->
            let x = _v in
            let _v = _menhir_action_302 x in
            MenhirBox_stmts _v
        | _ ->
            _eRR ()
    
    let rec _menhir_run_001 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_WHILE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState001 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_002 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_329 s in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_goto_value : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState442 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState038 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState259 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState054 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState183 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState173 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState665 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState635 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState566 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState530 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState519 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState497 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState491 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState487 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState480 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState475 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState469 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState465 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState454 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState451 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState447 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState444 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState439 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState431 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState427 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState408 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState404 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState384 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState375 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState359 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState338 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState340 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState343 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState333 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState312 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState001 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState004 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState005 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState307 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState006 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState302 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState008 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState012 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState286 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState283 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState020 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState275 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState273 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState271 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState032 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState172 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState252 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState247 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState245 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState243 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState241 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState239 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState237 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState235 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState228 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState225 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState223 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState221 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState219 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState214 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState211 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState209 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState207 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState205 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState203 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState200 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState148 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState165 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState161 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState151 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState143 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState141 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState139 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState137 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState118 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState101 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState099 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState097 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState094 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState092 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState090 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState088 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState086 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState084 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState082 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState080 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState078 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState076 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState073 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState071 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState069 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState067 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState065 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState063 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState061 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState059 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState057 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState055 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState051 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState035 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState033 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState030 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState028 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState026 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState021 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_177 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_value_, _startpos_value_, value) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_105 _endpos_value_ _startpos_value_ value in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_value_ _startpos_value_ _v _menhir_s _tok
    
    and _menhir_goto_expr_pattern : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState054 ->
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState442 ->
            _menhir_run_185 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState038 ->
            _menhir_run_185 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState259 ->
            _menhir_run_185 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState173 ->
            _menhir_run_185 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState183 ->
            _menhir_run_185 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_262 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SLICING ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.RPAR ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos_0) in
            let _v = _menhir_action_147 _endpos__3_ _startpos__1_ x in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_203 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_205 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_207 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_209 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_211 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_214 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_216 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_219 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_221 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_223 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_225 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_235 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_237 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_239 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_241 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_243 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_245 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_247 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_249 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState262
        | Tokens.COMMA ->
            let (_endpos_expr_pattern_, _startpos_expr_pattern_, expr_pattern) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_231 _endpos_expr_pattern_ _startpos_expr_pattern_ expr_pattern in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_pattern_ _startpos_expr_pattern_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_186 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SLICING (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState186 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_003 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_r_, _startpos_r_, r) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_327 r in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_r_ _v _menhir_s _tok
    
    and _menhir_run_004 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState004 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_005 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState005 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_006 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState006 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_007 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_i_, _startpos_i_, i) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_325 i in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_i_ _startpos_i_ _v _menhir_s _tok
    
    and _menhir_run_008 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState008 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_009 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState009
        | Tokens.LBRACE ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_151 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState009
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_060 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_010 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.RPAR ->
            let _v = _menhir_action_021 () in
            _menhir_run_299 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_011 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_b_, _startpos_b_, b) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_326 b in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_b_ _v _menhir_s _tok
    
    and _menhir_run_012 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState012 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_013 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_b_, _startpos_b_, b) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_328 b in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_b_ _v _menhir_s _tok
    
    and _menhir_run_014 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            let _menhir_s = MenhirState015 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_016 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_314 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_ty : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState643 ->
            _menhir_run_644 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState586 ->
            _menhir_run_603 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_542 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState538 ->
            _menhir_run_539 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState015 ->
            _menhir_run_296 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState291 ->
            _menhir_run_290 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState018 ->
            _menhir_run_290 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState249 ->
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState175 ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState103 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState115 ->
            _menhir_run_124 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState121 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_644 : type  ttv_stack. (((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ARROW -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_ARROW (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, _2) = ((), _v) in
        let _v = _menhir_action_250 _1 _2 in
        _menhir_goto_return_type _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_return_type : type  ttv_stack. (((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState650 ->
            _menhir_run_651 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState642 ->
            _menhir_run_645 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_651 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_return_type (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.RECURSELIMIT ->
            let _menhir_s = MenhirState652 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.BEGIN ->
            let _v = _menhir_action_248 () in
            _menhir_goto_recurse_limit _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_goto_recurse_limit : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _menhir_stack = MenhirCell0_recurse_limit (_menhir_stack, _v) in
        _menhir_run_628 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState654
    
    and _menhir_run_628 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_BEGIN (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState628 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARROBASE_LOOPLIMIT ->
            _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.END ->
            let _endpos__0_ = _endpos in
            let _v = _menhir_action_176 _endpos__0_ in
            _menhir_goto_maybe_empty_stmt_list _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_316 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_317 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState316
        | Tokens.LPAR ->
            _menhir_run_318 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState316
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.COMMA ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, MenhirState316, _v, _startpos_0, _endpos) in
                _menhir_run_328 _menhir_stack _menhir_lexbuf _menhir_lexer
            | Tokens.COLON | Tokens.EQ | Tokens.SEMI_COLON ->
                let _v =
                  let _1 = _v in
                  _menhir_action_051 _1
                in
                _menhir_run_331 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState316 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_317 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_050 _1 in
        _menhir_goto_discard_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_discard_or_identifier : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState463 ->
            _menhir_run_331 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState316 ->
            _menhir_run_331 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState325 ->
            _menhir_run_324 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState321 ->
            _menhir_run_324 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState318 ->
            _menhir_run_320 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_331 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let discard_or_identifier = _v in
        let _v = _menhir_action_044 discard_or_identifier in
        _menhir_goto_decl_item _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_decl_item : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState463 ->
            _menhir_run_464 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState316 ->
            _menhir_run_332 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_464 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_decl_item (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_EQ (_menhir_stack, MenhirState464) in
            let _menhir_s = MenhirState465 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState464
        | _ ->
            _eRR ()
    
    and _menhir_run_115 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState115 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_017 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_312 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_018 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState018 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RPAR ->
            let _v = _menhir_action_023 () in
            _menhir_goto_clist_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_019 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_INTEGER (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState019 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RECURSELIMIT | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.SUBTYPES | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let _v = _menhir_action_033 () in
            _menhir_goto_constraint_kind_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_020 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.MINUS ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.REAL_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.RBRACE ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let (_startpos__1_, x, _endpos__3_) = (_startpos, (), _endpos_1) in
                let _v = _menhir_action_031 x in
                _menhir_goto_constraint_kind _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
            | Tokens.NOT ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.MINUS ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.INT_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.IF ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.IDENTIFIER _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.BOOL_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.BNOT ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.BITVECTOR_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.ARBITRARY ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | _ ->
                _eRR ())
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | _ ->
            _eRR ()
    
    and _menhir_goto_constraint_kind : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState019 ->
            _menhir_run_289 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState249 ->
            _menhir_run_251 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState103 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_289 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_INTEGER -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_032 _1 in
        _menhir_goto_constraint_kind_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _tok
    
    and _menhir_goto_constraint_kind_opt : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_INTEGER -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_INTEGER (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_c_, c) = ((), _endpos, _v) in
        let _v = _menhir_action_311 _1 _endpos_c_ _startpos__1_ c in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_c_ _v _menhir_s _tok
    
    and _menhir_run_251 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
        let (_2, _endpos_constraint_kind_, _startpos_constraint_kind_, constraint_kind) = ((), _endpos, _startpos, _v) in
        let _v = _menhir_action_142 _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_pattern_ constraint_kind expr_pattern in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_constraint_kind_ _startpos_expr_pattern_ _v _menhir_s _tok
    
    and _menhir_run_134 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
        let (_2, _endpos_constraint_kind_, _startpos_constraint_kind_, constraint_kind) = ((), _endpos, _startpos, _v) in
        let _v = _menhir_action_096 _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_ constraint_kind expr in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_constraint_kind_ _startpos_expr_ _v _menhir_s _tok
    
    and _menhir_goto_expr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState665 ->
            _menhir_run_666 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_653 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState566 ->
            _menhir_run_567 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState530 ->
            _menhir_run_531 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState519 ->
            _menhir_run_520 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState497 ->
            _menhir_run_498 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState491 ->
            _menhir_run_492 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState487 ->
            _menhir_run_488 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState480 ->
            _menhir_run_481 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState475 ->
            _menhir_run_476 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState469 ->
            _menhir_run_470 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState465 ->
            _menhir_run_466 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState454 ->
            _menhir_run_455 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState451 ->
            _menhir_run_452 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState447 ->
            _menhir_run_448 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState444 ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState439 ->
            _menhir_run_440 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState431 ->
            _menhir_run_432 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState427 ->
            _menhir_run_428 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState404 ->
            _menhir_run_405 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState384 ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_366 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState359 ->
            _menhir_run_361 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState338 ->
            _menhir_run_334 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState333 ->
            _menhir_run_334 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState312 ->
            _menhir_run_313 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState001 ->
            _menhir_run_311 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState004 ->
            _menhir_run_310 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState006 ->
            _menhir_run_305 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState302 ->
            _menhir_run_303 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState008 ->
            _menhir_run_301 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState012 ->
            _menhir_run_297 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState286 ->
            _menhir_run_287 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState020 ->
            _menhir_run_285 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState283 ->
            _menhir_run_285 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState275 ->
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState273 ->
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState271 ->
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState032 ->
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState172 ->
            _menhir_run_257 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState252 ->
            _menhir_run_253 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState247 ->
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState245 ->
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState243 ->
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState241 ->
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState239 ->
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState237 ->
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState235 ->
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState228 ->
            _menhir_run_229 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState225 ->
            _menhir_run_227 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState223 ->
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState221 ->
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState219 ->
            _menhir_run_220 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState214 ->
            _menhir_run_215 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState211 ->
            _menhir_run_212 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState209 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState207 ->
            _menhir_run_208 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState205 ->
            _menhir_run_206 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState203 ->
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState200 ->
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_199 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_197 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_195 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_193 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_191 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_189 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_187 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_171 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState635 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState408 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState375 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState340 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState343 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState307 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState148 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState165 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState151 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState161 ->
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState143 ->
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState141 ->
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState139 ->
            _menhir_run_140 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState137 ->
            _menhir_run_138 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_136 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState118 ->
            _menhir_run_119 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState101 ->
            _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState099 ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState097 ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState094 ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState092 ->
            _menhir_run_093 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState090 ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState088 ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState086 ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState084 ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState082 ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState080 ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState078 ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState076 ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState073 ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState071 ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState069 ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState067 ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState065 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState063 ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState061 ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState059 ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState057 ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState055 ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState051 ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState035 ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState033 ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState030 ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState028 ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState026 ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState005 ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState021 ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_666 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_option_as_ty_ (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_ignored_or_identifier (_menhir_stack, _, name) = _menhir_stack in
            let MenhirCell1_global_decl_keyword_non_var (_menhir_stack, _menhir_s, keyword, _startpos_keyword_) = _menhir_stack in
            let (_endpos__2_, x, _4) = (_endpos_0, _v, ()) in
            let _v = _menhir_action_040 _4 _endpos__2_ _startpos_keyword_ keyword name ty x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState666
        | _ ->
            _eRR ()
    
    and _menhir_run_028 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState028 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_035 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState035 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_decl : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_decl (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.VAR ->
            _menhir_run_561 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.TYPE ->
            _menhir_run_571 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.SETTER ->
            _menhir_run_606 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.PRAGMA ->
            _menhir_run_634 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.LET ->
            _menhir_run_638 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.GETTER ->
            _menhir_run_639 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.FUNC ->
            _menhir_run_647 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.CONSTANT ->
            _menhir_run_657 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.CONFIG ->
            _menhir_run_658 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState669
        | Tokens.EOF ->
            let _v_0 = _menhir_action_169 () in
            _menhir_run_670 _menhir_stack _v_0
        | _ ->
            _eRR ()
    
    and _menhir_run_561 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState561 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_562 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_563 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_562 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_160 _1 in
        _menhir_goto_ignored_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_ignored_or_identifier : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState662 ->
            _menhir_run_663 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState561 ->
            _menhir_run_564 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_663 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ignored_or_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState663
        | Tokens.EQ ->
            let _v_0 = _menhir_action_221 () in
            _menhir_run_664 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState663 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_664 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_option_as_ty_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState665 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_564 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ignored_or_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState564
        | Tokens.EQ ->
            let _v_0 = _menhir_action_221 () in
            _menhir_run_565 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState564 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_565 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_option_as_ty_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState566 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_563 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = _v in
        let _v = _menhir_action_161 _1 in
        _menhir_goto_ignored_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_571 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SUBTYPES ->
                _menhir_run_573 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState572
            | Tokens.OF ->
                let _menhir_stack = MenhirCell1_OF (_menhir_stack, MenhirState572) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING ->
                    _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | Tokens.RECORD ->
                    let _startpos_1 = _menhir_lexbuf.Lexing.lex_start_p in
                    let _endpos_2 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _menhir_stack = MenhirCell1_RECORD (_menhir_stack, MenhirState586, _startpos_1) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.LBRACE ->
                        _menhir_run_576 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState587
                    | Tokens.SEMI_COLON | Tokens.SUBTYPES ->
                        let _v_3 = _menhir_action_152 () in
                        _menhir_run_588 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_2 _v_3 _tok
                    | _ ->
                        _eRR ())
                | Tokens.REAL ->
                    _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | Tokens.LPAR ->
                    _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | Tokens.INTEGER ->
                    _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | Tokens.IDENTIFIER _v_4 ->
                    _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState586
                | Tokens.EXCEPTION ->
                    let _startpos_5 = _menhir_lexbuf.Lexing.lex_start_p in
                    let _endpos_6 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _menhir_stack = MenhirCell1_EXCEPTION (_menhir_stack, MenhirState586, _startpos_5) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.LBRACE ->
                        _menhir_run_576 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState590
                    | Tokens.SEMI_COLON | Tokens.SUBTYPES ->
                        let _v_7 = _menhir_action_152 () in
                        _menhir_run_591 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_6 _v_7 _tok
                    | _ ->
                        _eRR ())
                | Tokens.ENUMERATION ->
                    let _menhir_s = MenhirState586 in
                    let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                    let _menhir_stack = MenhirCell1_ENUMERATION (_menhir_stack, _menhir_s, _startpos) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.LBRACE ->
                        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                        let _menhir_stack = MenhirCell0_LBRACE (_menhir_stack, _startpos) in
                        let _menhir_s = MenhirState593 in
                        let _tok = _menhir_lexer _menhir_lexbuf in
                        (match (_tok : MenhirBasics.token) with
                        | Tokens.IDENTIFIER _v ->
                            _menhir_run_594 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                        | _ ->
                            _eRR ())
                    | _ ->
                        _eRR ())
                | Tokens.BOOLEAN ->
                    _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | Tokens.BITS ->
                    _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | Tokens.BIT ->
                    _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | Tokens.ARRAY ->
                    _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState586
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_573 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WITH ->
                let _menhir_stack = MenhirCell1_SUBTYPES (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
                let _menhir_s = MenhirState575 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LBRACE ->
                    _menhir_run_576 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.SEMI_COLON ->
                let (_startpos__1_, _1, _endpos__2_, _2) = (_startpos, (), _endpos, _v) in
                let _v = _menhir_action_303 _1 _2 in
                _menhir_goto_subtype _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _startpos__1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_576 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState576 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_577 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.RBRACE ->
            let _v = _menhir_action_309 () in
            _menhir_goto_tclist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_577 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState577 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_tclist_typed_identifier_ : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_151 x in
        _menhir_goto_fields _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
    
    and _menhir_goto_fields : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState590 ->
            _menhir_run_589 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState587 ->
            _menhir_run_589 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState575 ->
            _menhir_run_585 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_589 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_153 _1 in
        _menhir_goto_fields_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_fields_opt : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState590 ->
            _menhir_run_591 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState587 ->
            _menhir_run_588 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_591 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_EXCEPTION -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_EXCEPTION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_l_, l) = ((), _endpos, _v) in
        let _v = _menhir_action_323 _1 _endpos_l_ _startpos__1_ l in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_ty_decl : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty_decl (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SUBTYPES ->
            _menhir_run_573 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState599
        | Tokens.SEMI_COLON ->
            let _v = _menhir_action_305 () in
            _menhir_goto_subtype_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_subtype_opt : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ty_decl -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_ty_decl (_menhir_stack, _, t) = _menhir_stack in
            let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _3, subtype_opt) = ((), _endpos, (), _v) in
            let _v = _menhir_action_038 _1 _3 _endpos__2_ _startpos__1_ subtype_opt t x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_588 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_RECORD -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_RECORD (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_l_, l) = ((), _endpos, _v) in
        let _v = _menhir_action_322 _1 _endpos_l_ _startpos__1_ l in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_585 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_SUBTYPES _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, _2, _, _) = _menhir_stack in
        let MenhirCell1_SUBTYPES (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_fields_, fields, _1_inlined1) = ((), _endpos, _v, ()) in
        let _v = _menhir_action_304 _1 _1_inlined1 _2 fields in
        _menhir_goto_subtype _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_fields_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_subtype : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState572 ->
            _menhir_run_604 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState599 ->
            _menhir_run_602 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_604 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _endpos_desc_, _startpos_desc_, desc) = ((), _endpos_0, _endpos, _startpos, _v) in
            let _v = _menhir_action_039 _1 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_602 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ty_decl -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let x = _v in
        let _v = _menhir_action_306 x in
        _menhir_goto_subtype_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
    
    and _menhir_run_104 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_name_, _startpos_name_, name) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_318 _endpos_name_ _startpos_name_ name in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_name_ _v _menhir_s _tok
    
    and _menhir_run_594 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v_0 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                _menhir_run_594 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState595
            | Tokens.RBRACE ->
                let (x, x_inlined1) = (_v, ()) in
                let _v = _menhir_action_210 x x_inlined1 in
                _menhir_goto_ntclist_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_209 x in
            _menhir_goto_ntclist_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_ntclist_IDENTIFIER_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState593 ->
            _menhir_run_597 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState595 ->
            _menhir_run_596 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_597 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ENUMERATION _menhir_cell0_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell0_LBRACE (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_ENUMERATION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, x, _endpos__3_) = ((), _v, _endpos) in
        let _v = _menhir_action_321 _1 _endpos__3_ _startpos__1_ x in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_596 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, h, _, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_211 _2 h t in
        _menhir_goto_ntclist_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_105 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_313 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_106 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BITS (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState107 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_116 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_315 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_117 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ARRAY (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_s = MenhirState118 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_606 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_SETTER (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                _menhir_run_608 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState607
            | Tokens.LPAR ->
                let _v_1 = _menhir_action_229 () in
                _menhir_run_616 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState607 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_608 : type  ttv_stack. (ttv_stack _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState608 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_609 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_609 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState609
        | Tokens.COMMA | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_219 x in
            _menhir_goto_opt_typed_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_opt_typed_identifier : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_opt_typed_identifier (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState612 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_609 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_191 x in
            _menhir_goto_nclist_opt_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_opt_typed_identifier_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState608 ->
            _menhir_run_614 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState612 ->
            _menhir_run_613 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_614 : type  ttv_stack. (ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let x = _v in
        let _v = _menhir_action_230 x in
        _menhir_goto_params_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_params_opt : type  ttv_stack. (ttv_stack _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState648 ->
            _menhir_run_649 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState640 ->
            _menhir_run_641 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState607 ->
            _menhir_run_616 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_649 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_617 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState649
        | _ ->
            _eRR ()
    
    and _menhir_run_617 : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState617 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_577 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.RPAR ->
            let _v = _menhir_action_025 () in
            _menhir_goto_clist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist_typed_identifier_ : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_LPAR -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let x = _v in
        let _v = _menhir_action_244 x in
        let _1 = _v in
        let _v = _menhir_action_154 _1 in
        _menhir_goto_func_args _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_func_args : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState649 ->
            _menhir_run_650 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState641 ->
            _menhir_run_642 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState616 ->
            _menhir_run_625 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_650 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            _menhir_run_628 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState650
        | Tokens.ARROW ->
            _menhir_run_643 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState650
        | _ ->
            _eRR ()
    
    and _menhir_run_643 : type  ttv_stack. (((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_ARROW (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState643 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_642 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ARROW ->
            _menhir_run_643 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState642
        | _ ->
            _eRR ()
    
    and _menhir_run_625 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState626 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_577 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_641 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_617 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState641
        | _ ->
            _eRR ()
    
    and _menhir_run_616 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_617 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState616
        | _ ->
            _eRR ()
    
    and _menhir_run_613 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_opt_typed_identifier -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_opt_typed_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_192 _2 h t in
        _menhir_goto_nclist_opt_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_634 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState635
            | Tokens.REAL_LIT _v_2 ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState635
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState635
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState635
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState635
            | Tokens.INT_LIT _v_3 ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState635
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState635
            | Tokens.IDENTIFIER _v_4 ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState635
            | Tokens.BOOL_LIT _v_5 ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState635
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState635
            | Tokens.BITVECTOR_LIT _v_6 ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState635
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState635
            | Tokens.SEMI_COLON ->
                let _v_7 = _menhir_action_021 () in
                _menhir_run_636 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_636 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_043 _1 _endpos__2_ _startpos__1_ e x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_638 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_157 _1 in
        _menhir_goto_global_decl_keyword_non_var _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_global_decl_keyword_non_var : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_global_decl_keyword_non_var (_menhir_stack, _menhir_s, _v, _startpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_562 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
        | Tokens.IDENTIFIER _v_0 ->
            _menhir_run_563 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState662
        | _ ->
            _eRR ()
    
    and _menhir_run_639 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_GETTER (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                _menhir_run_608 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState640
            | Tokens.LPAR ->
                let _v_1 = _menhir_action_229 () in
                _menhir_run_641 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState640 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_647 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                _menhir_run_608 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState648
            | Tokens.LPAR ->
                let _v_1 = _menhir_action_229 () in
                _menhir_run_649 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState648 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_657 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_158 _1 in
        _menhir_goto_global_decl_keyword_non_var _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_658 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_159 _1 in
        _menhir_goto_global_decl_keyword_non_var _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_057 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_RDIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState057 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_030 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_POW (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState030 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_061 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_PLUS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState061 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_071 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_OR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState071 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_073 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState073
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState073
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState073
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState073
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_099 _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_e_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState073
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState073
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState073
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState073
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState073
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState073
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState073
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState073
        | _ ->
            _eRR ()
    
    and _menhir_run_063 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MUL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState063 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_065 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MOD (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState065 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_076 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState076 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_078 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState078 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_026 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LLBRACKET (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState026 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_086 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState086 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_032 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACKET (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState032 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.COLON ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_033 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState033 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_037 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IN (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState037 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BNOT ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_038 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState038 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_039 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState039 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_051 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState051
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState051
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState051
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState051
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState051
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState051
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState051
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState051
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState051
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState051
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState051
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState051
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
            let _v = _menhir_action_233 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_pattern : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState054 ->
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState442 ->
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState038 ->
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState259 ->
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState183 ->
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState173 ->
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_258 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState259 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MASK_LIT _v ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.LPAR ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LEQ ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LBRACE ->
                _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.GEQ ->
                _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_053 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_237 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_054 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState054 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_055 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState055 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_135 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState135 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_147 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState147
        | Tokens.LBRACE ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState148
            | Tokens.REAL_LIT _v_2 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState148
            | Tokens.RBRACE ->
                let _endpos_3 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let (_2, _endpos__3_, _3, _endpos_desc_, _startpos_desc_, desc) = ((), _endpos_3, (), _endpos, _startpos, _v) in
                let _v = _menhir_action_148 _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc in
                _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
            | Tokens.NOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState148
            | Tokens.MINUS ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState148
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState148
            | Tokens.INT_LIT _v_4 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState148
            | Tokens.IF ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState148
            | Tokens.IDENTIFIER _v_5 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState148
            | Tokens.BOOL_LIT _v_6 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState148
            | Tokens.BNOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState148
            | Tokens.BITVECTOR_LIT _v_7 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7 MenhirState148
            | Tokens.ARBITRARY ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState147, _startpos_0) in
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState148
            | _ ->
                _eRR ())
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_106 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_150 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState150
        | Tokens.LBRACE ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_151 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState150
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_161 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState150
        | Tokens.AND | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.SHL | Tokens.SHR ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_060 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_151 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState151
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState151
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_2, _endpos__3_, _3) = ((), _endpos, ()) in
            let _v = _menhir_action_102 _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState151
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState151
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState151
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState151
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState151
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState151
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState151
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState151
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState151
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState151
        | _ ->
            _eRR ()
    
    and _menhir_run_161 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_EQ (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState161 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_170 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState170 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_172 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState172 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_173 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_173 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState173 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_174 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            let _menhir_s = MenhirState175 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_182 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState183 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MASK_LIT _v ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.LPAR ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LEQ ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LBRACE ->
                _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.GEQ ->
                _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.ARROW | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let (_endpos_x_, _startpos_x_, x) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_193 x in
            _menhir_goto_nclist_pattern_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_pattern_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState259 ->
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState442 ->
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState038 ->
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState173 ->
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState183 ->
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_260 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_pattern -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_pattern (_menhir_stack, _, x, _, _) = _menhir_stack in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_2, _endpos__3_, li) = ((), _endpos_0, _v) in
            let _v = _menhir_action_247 _2 li x in
            let (_endpos, _startpos) = (_endpos__3_, _startpos__1_) in
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_236 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_256 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_239 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_pattern_list _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_pattern_list : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState442 ->
            _menhir_run_443 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState038 ->
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState173 ->
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_443 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_pattern_list (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHERE ->
            let _menhir_s = MenhirState444 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.ARROW ->
            let _menhir_s = MenhirState501 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_354 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.RPAR ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.SEMI_COLON ->
                    let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    let (_startpos__1_, _1, _endpos__2_, _3, _2_inlined1) = (_startpos, (), _endpos_1, (), ()) in
                    let _v = _menhir_action_295 _1 _2_inlined1 _3 _endpos__2_ _startpos__1_ in
                    _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_stmt : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState000 ->
            _menhir_run_557 _menhir_stack _v _tok
        | MenhirState671 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState628 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState315 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState547 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState543 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState406 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState523 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState434 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState506 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState501 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState446 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState495 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState460 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState456 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_460 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.VAR ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.UNREACHABLE ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.TRY ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.THROW ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.RETURN ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.REPEAT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.PRINTLN ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.PRINT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.PRAGMA ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.PASS ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.LET ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.IDENTIFIER _v_0 ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState460
        | Tokens.FOR ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.DEBUG ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.CONSTANT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.CASE ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.ASSERT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.ARROBASE_LOOPLIMIT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | Tokens.CATCH | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.OTHERWISE | Tokens.UNTIL | Tokens.WHEN ->
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_207 x in
            _menhir_goto_nlist_stmt_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_358 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_TRY (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState358 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARROBASE_LOOPLIMIT ->
            _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_359 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_startpos__1_, _1, _endpos__2_) = (_startpos, (), _endpos) in
            let _v = _menhir_action_299 _1 _endpos__2_ _startpos__1_ in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | _ ->
            _eRR ()
    
    and _menhir_run_363 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_RETURN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState363 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.SEMI_COLON ->
            let _v = _menhir_action_223 () in
            _menhir_goto_option_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_goto_option_expr_ : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_RETURN -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_RETURN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos__2_, _2_inlined1) = ((), _endpos, _v) in
        let _v = _menhir_action_278 _1 _2_inlined1 _endpos__2_ _startpos__1_ in
        _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_run_367 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_REPEAT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState367 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARROBASE_LOOPLIMIT ->
            _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_368 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRINTLN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState368 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_371 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRINT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState371 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_374 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState375
            | Tokens.REAL_LIT _v_2 ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState375
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState375
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState375
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState375
            | Tokens.INT_LIT _v_3 ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState375
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState375
            | Tokens.IDENTIFIER _v_4 ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState375
            | Tokens.BOOL_LIT _v_5 ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState375
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState375
            | Tokens.BITVECTOR_LIT _v_6 ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState375
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState375
            | Tokens.SEMI_COLON ->
                let _v_7 = _menhir_action_021 () in
                _menhir_run_376 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_376 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_300 _1 _endpos__2_ _startpos__1_ e x in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_378 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_startpos__1_, _1, _endpos__2_) = (_startpos, (), _endpos) in
            let _v = _menhir_action_277 _1 _endpos__2_ _startpos__1_ in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_380 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_166 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_lexpr (_menhir_stack, _menhir_s, _v, _startpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState475 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_381 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState381 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_383 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_382 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_052 _1 in
        _menhir_goto_discard_or_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_discard_or_sliced_basic_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState397 ->
            _menhir_run_396 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState394 ->
            _menhir_run_396 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState381 ->
            _menhir_run_393 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_396 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState397 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_383 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_183 x in
            _menhir_goto_nclist_discard_or_sliced_basic_lexpr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_383 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState383
        | Tokens.DOT ->
            _menhir_run_387 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState383
        | Tokens.COMMA | Tokens.LBRACKET | Tokens.RPAR ->
            let _v_0 = _menhir_action_201 () in
            _menhir_run_391 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_384 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LLBRACKET (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState384 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_387 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DOT (_menhir_stack, _menhir_s) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_388 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_388 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_DOT -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.DOT ->
            _menhir_run_387 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState388
        | Tokens.COMMA | Tokens.EQ | Tokens.LBRACKET | Tokens.RPAR ->
            let _v_0 = _menhir_action_201 () in
            _menhir_run_389 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_389 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_DOT _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let MenhirCell1_DOT (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, t) = ((), _v) in
        let _v = _menhir_action_202 _1 _endpos_desc_ _startpos_desc_ desc t in
        _menhir_goto_nested_fields _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_nested_fields : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState407 ->
            _menhir_run_391 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState383 ->
            _menhir_run_391 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState386 ->
            _menhir_run_390 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState388 ->
            _menhir_run_389 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_391 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let fields = _v in
        let _v = _menhir_action_004 _endpos_desc_ _startpos_desc_ desc fields in
        _menhir_goto_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_goto_basic_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_basic_lexpr (_menhir_stack, _menhir_s, _v, _startpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState399
        | Tokens.COMMA | Tokens.EQ | Tokens.RPAR ->
            let (_startpos_basic_lexpr_, basic_lexpr) = (_startpos, _v) in
            let _v = _menhir_action_259 basic_lexpr in
            _menhir_goto_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_basic_lexpr_ _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_goto_sliced_basic_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState671 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState628 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState315 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState547 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState543 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState406 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState523 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState434 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState506 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState501 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState446 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState495 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState456 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState460 ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState397 ->
            _menhir_run_392 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState394 ->
            _menhir_run_392 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState381 ->
            _menhir_run_392 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_461 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let (_startpos_sliced_basic_lexpr_, sliced_basic_lexpr) = (_startpos, _v) in
        let _v = _menhir_action_164 sliced_basic_lexpr in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_sliced_basic_lexpr_ _v _menhir_s _tok
    
    and _menhir_run_392 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let sliced_basic_lexpr = _v in
        let _v = _menhir_action_053 sliced_basic_lexpr in
        _menhir_goto_discard_or_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_390 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LLBRACKET, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RRBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_RRBRACKET (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, idx, _, _) = _menhir_stack in
        let MenhirCell1_LLBRACKET (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (_2, fields, _4) = ((), _v, ()) in
        let _v = _menhir_action_005 _2 _4 _endpos_desc_ _startpos_desc_ desc fields idx in
        _menhir_goto_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_goto_nclist_discard_or_sliced_basic_lexpr_ : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState397 ->
            _menhir_run_398 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState394 ->
            _menhir_run_395 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_398 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_184 _2 h t in
        _menhir_goto_nclist_discard_or_sliced_basic_lexpr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_395 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _, x) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_029 _2 li x in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_165 _endpos__3_ _startpos__1_ x in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_393 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState394 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_383 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_403 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_171 _1 in
        _menhir_goto_local_decl_keyword_non_var _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_local_decl_keyword_non_var : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, _v, _startpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_317 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState463
        | Tokens.LPAR ->
            _menhir_run_318 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState463
        | Tokens.IDENTIFIER _v_0 ->
            _menhir_run_319 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState463
        | _ ->
            _eRR ()
    
    and _menhir_run_318 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState318 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_317 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_319 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_319 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = _v in
        let _v = _menhir_action_051 _1 in
        _menhir_goto_discard_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_404 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState404 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_407 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState407
        | Tokens.LLBRACKET ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState407
        | Tokens.LBRACE ->
            let _menhir_s = MenhirState407 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_s = MenhirState408 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_DOT (_menhir_stack, MenhirState407) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LPAR ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState410 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.MINUS ->
                    _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_412 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.LBRACKET ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState421 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_422 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.IDENTIFIER _v ->
                _menhir_run_388 _menhir_stack _menhir_lexbuf _menhir_lexer _v
            | _ ->
                _eRR ())
        | Tokens.EQ | Tokens.LBRACKET ->
            let _v_12 = _menhir_action_201 () in
            _menhir_run_391 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_411 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_048 _1 in
        _menhir_goto_discard_or_field _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_discard_or_field : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState417 ->
            _menhir_run_416 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState414 ->
            _menhir_run_416 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState410 ->
            _menhir_run_413 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_416 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_field as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_discard_or_field (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState417 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_412 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_179 x in
            _menhir_goto_nclist_discard_or_field_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_412 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_desc_, _startpos_desc_, desc) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_049 _endpos_desc_ _startpos_desc_ desc in
        _menhir_goto_discard_or_field _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_nclist_discard_or_field_ : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_field as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState417 ->
            _menhir_run_418 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState414 ->
            _menhir_run_415 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_418 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_field, ttv_result) _menhir_cell1_discard_or_field -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_field (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_180 _2 h t in
        _menhir_goto_nclist_discard_or_field_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_415 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_DOT _menhir_cell0_LPAR, ttv_result) _menhir_cell1_discard_or_field -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_field (_menhir_stack, _, x) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_028 _2 li x in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (_2, x, _endpos__3_) = ((), _v, _endpos) in
        let _v = _menhir_action_168 _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_run_413 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_DOT _menhir_cell0_LPAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_discard_or_field (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState414 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_412 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_422 : type  ttv_stack ttv_result. (ttv_stack _menhir_cell0_LBRACKET as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            _menhir_run_328 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR ()
    
    and _menhir_run_328 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState328 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_043 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState044 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON | Tokens.RBRACKET ->
            let x = _v in
            let _v = _menhir_action_177 x in
            _menhir_goto_nclist_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_IDENTIFIER_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState328 ->
            _menhir_run_329 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState231 ->
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState042 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState044 ->
            _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_329 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_027 _2 li x in
        _menhir_goto_clist2_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_clist2_IDENTIFIER_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState484 ->
            _menhir_run_485 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState421 ->
            _menhir_run_423 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState316 ->
            _menhir_run_351 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_485 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_call _menhir_cell0_LBRACKET as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_RBRACKET (_menhir_stack, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.EQ ->
                let _menhir_s = MenhirState487 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.REAL_LIT _v ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INT_LIT _v ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOL_LIT _v ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITVECTOR_LIT _v ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_423 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_DOT _menhir_cell0_LBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_2, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_167 _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
            _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_351 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState351
        | _ ->
            _eRR ()
    
    and _menhir_run_232 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_DOT _menhir_cell0_LBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_140 _2 _endpos__3_ _startpos_e_ e x in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_046 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_DOT _menhir_cell0_LBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_094 _2 _endpos__3_ _startpos_e_ e x in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_045 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, h, _, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_178 _2 h t in
        _menhir_goto_nclist_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_425 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_FOR (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.EQ ->
                let _menhir_s = MenhirState427 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.REAL_LIT _v ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INT_LIT _v ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOL_LIT _v ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITVECTOR_LIT _v ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_435 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_DEBUG (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState435 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_438 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_172 _1 in
        _menhir_goto_local_decl_keyword_non_var _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_439 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState439 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_447 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ASSERT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState447 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_450 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ARROBASE_LOOPLIMIT (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState451 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_nlist_stmt_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState671 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState628 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState315 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState547 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState543 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState406 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState523 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState434 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState506 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState501 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState446 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState495 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState456 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState460 ->
            _menhir_run_462 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_494 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_301 _1 in
        _menhir_goto_stmt_list _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_stmt_list : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState671 ->
            _menhir_run_673 _menhir_stack _v _tok
        | MenhirState628 ->
            _menhir_run_629 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState315 ->
            _menhir_run_554 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState547 ->
            _menhir_run_548 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState543 ->
            _menhir_run_544 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState540 ->
            _menhir_run_541 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState358 ->
            _menhir_run_534 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_529 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState523 ->
            _menhir_run_524 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState521 ->
            _menhir_run_522 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState406 ->
            _menhir_run_518 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState434 ->
            _menhir_run_515 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState506 ->
            _menhir_run_507 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState501 ->
            _menhir_run_502 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState446 ->
            _menhir_run_500 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState495 ->
            _menhir_run_496 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState456 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_629 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_BEGIN -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let _1 = _v in
        let _v = _menhir_action_175 _1 in
        _menhir_goto_maybe_empty_stmt_list _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
    
    and _menhir_goto_maybe_empty_stmt_list : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_BEGIN -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_BEGIN (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_endpos__2_, _2, x, _1_inlined1) = (_endpos_0, (), _v, ()) in
                let _v = _menhir_action_155 _1_inlined1 _2 x in
                _menhir_goto_func_body _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.GETTER | Tokens.LET | Tokens.PRAGMA | Tokens.SETTER | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_BEGIN (_menhir_stack, _menhir_s) = _menhir_stack in
                let (x, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_v, _endpos, _startpos, ()) in
                let _v = _menhir_action_156 _1_inlined1 _endpos__1_inlined1_ _startpos__1_inlined1_ x in
                _menhir_goto_func_body _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_func_body : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState650 ->
            _menhir_run_656 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState654 ->
            _menhir_run_655 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState645 ->
            _menhir_run_646 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState627 ->
            _menhir_run_633 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_656 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
        let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
        let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
        let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_body_, body) = ((), _endpos, _v) in
        let _v = _menhir_action_035 _1 _endpos_body_ _startpos__1_ body func_args name params_opt in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_655 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type _menhir_cell0_recurse_limit -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_recurse_limit (_menhir_stack, recurse_limit) = _menhir_stack in
        let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
        let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
        let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
        let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
        let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_body_, body) = ((), _endpos, _v) in
        let _v = _menhir_action_034 _1 _endpos_body_ _startpos__1_ body func_args name params_opt recurse_limit return_type in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_646 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
        let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
        let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
        let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
        let MenhirCell1_GETTER (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_func_body_, func_body) = ((), _endpos, _v) in
        let _v = _menhir_action_036 _1 _endpos_func_body_ _startpos__1_ func_args func_body name params_opt return_type in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_633 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_typed_identifier -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_typed_identifier (_menhir_stack, _, v) = _menhir_stack in
        let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
        let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
        let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
        let MenhirCell1_SETTER (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _5, _endpos_func_body_, func_body) = ((), (), _endpos, _v) in
        let _v = _menhir_action_037 _1 _5 _endpos_func_body_ _startpos__1_ func_args func_body name params_opt v in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_554 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHILE, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_loop_limit -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_loop_limit (_menhir_stack, _, loop_limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_WHILE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _4, _1_inlined1, stmt_list) = ((), _endpos_1, (), (), (), _v) in
                let _v = _menhir_action_269 _1 _1_inlined1 _2 _4 _endpos__2_ _startpos__1_ expr loop_limit stmt_list in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ARROBASE_LOOPLIMIT | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_loop_limit (_menhir_stack, _, loop_limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_WHILE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _4, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, stmt_list) = ((), (), _endpos_0, _startpos, (), _v) in
                let _v = _menhir_action_270 _1 _1_inlined1 _4 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ expr loop_limit stmt_list in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_548 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_TRY, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_nlist_catcher_ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let (_1, _2, _3) = ((), (), _v) in
        let _v = _menhir_action_228 _1 _2 _3 in
        _menhir_goto_otherwise_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
    
    and _menhir_goto_otherwise_opt : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_TRY, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_nlist_catcher_ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_nlist_catcher_ (_menhir_stack, _, c) = _menhir_stack in
                let MenhirCell1_stmt_list (_menhir_stack, _, s, _) = _menhir_stack in
                let MenhirCell1_TRY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _1_inlined1, o) = ((), _endpos_0, (), (), (), _v) in
                let _v = _menhir_action_273 _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ c o s in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ARROBASE_LOOPLIMIT | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_nlist_catcher_ (_menhir_stack, _, c) = _menhir_stack in
                let MenhirCell1_stmt_list (_menhir_stack, _, s, _) = _menhir_stack in
                let MenhirCell1_TRY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, o) = ((), (), _endpos, _startpos, (), _v) in
                let _v = _menhir_action_274 _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ c o s in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_544 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_ty -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (_1, _4, stmt_list) = ((), (), _v) in
        let _v = _menhir_action_019 _1 _4 stmt_list ty in
        _menhir_goto_catcher _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_catcher : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.WHEN ->
            let _menhir_stack = MenhirCell1_catcher (_menhir_stack, _menhir_s, _v) in
            _menhir_run_536 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState552
        | Tokens.END | Tokens.OTHERWISE ->
            let x = _v in
            let _v = _menhir_action_205 x in
            _menhir_goto_nlist_catcher_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_536 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.COLON ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, MenhirState536, _v, _startpos_0, _endpos) in
                let _menhir_s = MenhirState538 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING ->
                    _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REAL ->
                    _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INTEGER ->
                    _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOLEAN ->
                    _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITS ->
                    _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BIT ->
                    _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ARRAY ->
                    _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.ARROW ->
                let _v =
                  let (_endpos_name_, _startpos_name_, name) = (_endpos, _startpos_0, _v) in
                  _menhir_action_318 _endpos_name_ _startpos_name_ name
                in
                _menhir_run_542 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v MenhirState536 _tok
            | _ ->
                _eRR ())
        | Tokens.BOOLEAN ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | Tokens.BITS ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | Tokens.BIT ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | Tokens.ARRAY ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState536
        | _ ->
            _eRR ()
    
    and _menhir_run_542 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ARROW ->
            let _menhir_s = MenhirState543 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_nlist_catcher_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState552 ->
            _menhir_run_553 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState535 ->
            _menhir_run_545 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_553 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_catcher -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_catcher (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let l = _v in
        let _v = _menhir_action_206 l x in
        _menhir_goto_nlist_catcher_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_545 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_TRY, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nlist_catcher_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.OTHERWISE ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.ARROW ->
                let _menhir_s = MenhirState547 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.WHILE ->
                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.VAR ->
                    _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.UNREACHABLE ->
                    _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.TRY ->
                    _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.THROW ->
                    _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.RETURN ->
                    _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REPEAT ->
                    _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINTLN ->
                    _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINT ->
                    _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRAGMA ->
                    _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PASS ->
                    _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LET ->
                    _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IF ->
                    _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.FOR ->
                    _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.DEBUG ->
                    _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CONSTANT ->
                    _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CASE ->
                    _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ASSERT ->
                    _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ARROBASE_LOOPLIMIT ->
                    _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.END ->
            let _v = _menhir_action_227 () in
            _menhir_goto_otherwise_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_541 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_ty -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _, x, _, _) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (_1, _4, stmt_list) = ((), (), _v) in
        let _v = _menhir_action_020 _1 _4 stmt_list ty x in
        _menhir_goto_catcher _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_534 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_TRY as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.CATCH ->
            let _menhir_s = MenhirState535 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHEN ->
                _menhir_run_536 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_529 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_REPEAT as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.UNTIL ->
            let _menhir_s = MenhirState530 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_524 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSE -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ELSE (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, _endpos__2_, _2) = ((), _endpos, _v) in
        let _v = _menhir_action_253 _1 _2 in
        _menhir_goto_s_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_goto_s_else : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState518 ->
            _menhir_run_526 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState522 ->
            _menhir_run_525 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_526 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_stmt_list (_menhir_stack, _, s1, _) = _menhir_stack in
                let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
                let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _1_inlined1, s2) = ((), _endpos_1, (), (), (), _v) in
                let _v = _menhir_action_263 _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ e s1 s2 in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ARROBASE_LOOPLIMIT | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_stmt_list (_menhir_stack, _, s1, _) = _menhir_stack in
                let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
                let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, s2) = ((), (), _endpos_0, _startpos, (), _v) in
                let _v = _menhir_action_264 _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ e s1 s2 in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_525 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_stmt_list (_menhir_stack, _, s1, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
        let MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_s2_, s2) = ((), (), _endpos, _v) in
        let _v = _menhir_action_251 _1 _3 _endpos_s2_ _startpos__1_ e s1 s2 in
        _menhir_goto_s_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s2_ _v _menhir_s _tok
    
    and _menhir_run_522 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ELSIF ->
            _menhir_run_519 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState522
        | Tokens.ELSE ->
            _menhir_run_523 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState522
        | Tokens.END ->
            let _v_0 =
              let _endpos__0_ = _endpos in
              _menhir_action_252 _endpos__0_
            in
            _menhir_run_525 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_519 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState519 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_523 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_ELSE (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState523 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARROBASE_LOOPLIMIT ->
            _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_518 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ELSIF ->
            _menhir_run_519 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState518
        | Tokens.ELSE ->
            _menhir_run_523 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState518
        | Tokens.END ->
            let _v_0 =
              let _endpos__0_ = _endpos in
              _menhir_action_252 _endpos__0_
            in
            _menhir_run_526 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_515 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_direction, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_loop_limit -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_loop_limit (_menhir_stack, _, limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, end_e, _, _) = _menhir_stack in
                let MenhirCell1_direction (_menhir_stack, _, dir) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, start_e, _, _) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, index_name, _, _) = _menhir_stack in
                let MenhirCell1_FOR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, body, _1_inlined1, _8) = ((), _endpos_1, (), (), _v, (), ()) in
                let _v = _menhir_action_271 _1 _1_inlined1 _2 _3 _8 _endpos__2_ _startpos__1_ body dir end_e index_name limit start_e in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ARROBASE_LOOPLIMIT | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_loop_limit (_menhir_stack, _, limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, end_e, _, _) = _menhir_stack in
                let MenhirCell1_direction (_menhir_stack, _, dir) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, start_e, _, _) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, index_name, _, _) = _menhir_stack in
                let MenhirCell1_FOR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, body, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, _8) = ((), (), _v, _endpos_0, _startpos, (), ()) in
                let _v = _menhir_action_272 _1 _1_inlined1 _3 _8 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ body dir end_e index_name limit start_e in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_507 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_CASE, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_OF, ttv_result) _menhir_cell1_case_alt_list -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let (_1, _2, otherwise_stmt) = ((), (), _v) in
        let _v = _menhir_action_018 _1 _2 otherwise_stmt in
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_case_alt_list (_menhir_stack, _, alt) = _menhir_stack in
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _1_inlined1, case_otherwise) = ((), _endpos_0, (), (), (), _v) in
                let _v = _menhir_action_267 _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ alt case_otherwise expr in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ARROBASE_LOOPLIMIT | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_case_alt_list (_menhir_stack, _, alt) = _menhir_stack in
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, case_otherwise) = ((), (), _endpos, _startpos, (), _v) in
                let _v = _menhir_action_268 _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ alt case_otherwise expr in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_502 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_pattern_list -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_pattern_list (_menhir_stack, _, pattern) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _4, _endpos_stmt_, stmt) = ((), (), _endpos, _v) in
        let _v = _menhir_action_015 _1 _4 _endpos_stmt_ _startpos__1_ pattern stmt in
        _menhir_goto_case_alt _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_case_alt : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.WHEN ->
            let _menhir_stack = MenhirCell1_case_alt (_menhir_stack, _menhir_s, _v) in
            _menhir_run_442 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState513
        | Tokens.END | Tokens.OTHERWISE ->
            let x = _v in
            let _v = _menhir_action_203 x in
            _menhir_goto_nlist_case_alt_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_442 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState442 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_nlist_case_alt_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState513 ->
            _menhir_run_514 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState441 ->
            _menhir_run_503 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_514 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_case_alt -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_case_alt (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let l = _v in
        let _v = _menhir_action_204 l x in
        _menhir_goto_nlist_case_alt_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_503 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_CASE, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_OF as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _1 = _v in
        let _v = _menhir_action_017 _1 in
        match (_tok : MenhirBasics.token) with
        | Tokens.OTHERWISE ->
            let _menhir_stack = MenhirCell1_case_alt_list (_menhir_stack, _menhir_s, _v) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.ARROW ->
                let _menhir_s = MenhirState506 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.WHILE ->
                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.VAR ->
                    _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.UNREACHABLE ->
                    _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.TRY ->
                    _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.THROW ->
                    _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.RETURN ->
                    _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REPEAT ->
                    _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINTLN ->
                    _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINT ->
                    _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRAGMA ->
                    _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PASS ->
                    _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LET ->
                    _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IF ->
                    _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.FOR ->
                    _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.DEBUG ->
                    _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CONSTANT ->
                    _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CASE ->
                    _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ASSERT ->
                    _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ARROBASE_LOOPLIMIT ->
                    _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _1_inlined1, alt) = ((), _endpos_1, (), (), (), _v) in
                let _v = _menhir_action_265 _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ alt expr in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ARROBASE_LOOPLIMIT | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, alt) = ((), (), _endpos, _startpos, (), _v) in
                let _v = _menhir_action_266 _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ alt expr in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_500 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_pattern_list, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ARROW -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ARROW (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, _2, _, _) = _menhir_stack in
        let MenhirCell1_pattern_list (_menhir_stack, _, pattern) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _4, _1_inlined1, _endpos_stmt_, stmt) = ((), (), (), _endpos, _v) in
        let _v = _menhir_action_016 _1 _1_inlined1 _2 _4 _endpos_stmt_ _startpos__1_ pattern stmt in
        _menhir_goto_case_alt _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_496 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR _menhir_cell0_REPEAT as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.UNTIL ->
            let _menhir_s = MenhirState497 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_457 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR _menhir_cell0_WHILE, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_DO -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_DO (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, cond, _, _) = _menhir_stack in
                let MenhirCell0_WHILE (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_RPAR (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, x, _, _) = _menhir_stack in
                let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_ARROBASE_LOOPLIMIT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _5, body, _1_inlined2) = ((), _endpos_1, (), (), (), _v, ()) in
                let _v = _menhir_action_275 _1 _1_inlined2 _2 _3 _5 _endpos__2_ _startpos__1_ body cond x in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ARROBASE_LOOPLIMIT | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_DO (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, cond, _, _) = _menhir_stack in
                let MenhirCell0_WHILE (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_RPAR (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, x, _, _) = _menhir_stack in
                let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_ARROBASE_LOOPLIMIT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _5, body, _endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) = ((), (), (), _v, _endpos_0, _startpos, ()) in
                let _v = _menhir_action_276 _1 _1_inlined2 _3 _5 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ body cond x in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined2_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_462 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_stmt -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_stmt (_menhir_stack, _menhir_s, x, _) = _menhir_stack in
        let (_endpos_l_, l) = (_endpos, _v) in
        let _v = _menhir_action_208 l x in
        _menhir_goto_nlist_stmt_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_l_ _v _menhir_s _tok
    
    and _menhir_run_264 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos) in
            let _v = _menhir_action_240 x in
            _menhir_goto_pattern_set _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_pattern_set : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState037 ->
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState216 ->
            _menhir_run_218 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState442 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState038 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState259 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState054 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState183 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState173 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_266 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_IN -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_IN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
        let (_2, _endpos_pattern_set_, pattern_set) = ((), _endpos, _v) in
        let _v = _menhir_action_097 _2 _endpos_pattern_set_ _startpos_expr_ expr pattern_set in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_pattern_set_ _startpos_expr_ _v _menhir_s _tok
    
    and _menhir_run_218 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_IN -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_IN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
        let (_2, _endpos_pattern_set_, pattern_set) = ((), _endpos, _v) in
        let _v = _menhir_action_143 _2 _endpos_pattern_set_ _startpos_expr_pattern_ expr_pattern pattern_set in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_pattern_set_ _startpos_expr_pattern_ _v _menhir_s _tok
    
    and _menhir_run_179 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_238 _1 in
        _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_180 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_241 _1 _endpos__3_ _startpos__1_ x in
            _menhir_goto_pattern_set _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_184 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_pattern -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_pattern (_menhir_stack, _menhir_s, h, _startpos_h_, _) = _menhir_stack in
        let (_2, _endpos_t_, t) = ((), _endpos, _v) in
        let _v = _menhir_action_194 _2 h t in
        _menhir_goto_nclist_pattern_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_h_ _v _menhir_s _tok
    
    and _menhir_run_217 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IN as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState217 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_173 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_092 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IMPL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState092 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_088 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_GT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState088 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_090 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState090 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_094 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState094
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState094
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState094
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState094
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_098 _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_ expr in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_expr_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState094
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState094
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState094
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState094
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState094
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState094
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState094
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState094
        | _ ->
            _eRR ()
    
    and _menhir_run_080 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_EOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState080 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_041 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_DOT (_menhir_stack, _menhir_s) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState042 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.IDENTIFIER _v ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, _endpos_x_, x) = ((), _endpos, _v) in
            let _v = _menhir_action_093 _2 _endpos_x_ _startpos_e_ e x in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_067 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIVRM (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState067 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_069 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState069 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_059 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_CONCAT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState059 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_082 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState082 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_097 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState097 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_099 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BEQ (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState099 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_101 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BAND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState101 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_103 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState103 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_084 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState084 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_653 : type  ttv_stack. (((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState653
        | Tokens.BEGIN ->
            let (_1, _2) = ((), _v) in
            let _v = _menhir_action_249 _1 _2 in
            _menhir_goto_recurse_limit _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_567 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_option_as_ty_ (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_ignored_or_identifier (_menhir_stack, _, name) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, x, _4) = ((), _endpos_0, _v, ()) in
            let _v = _menhir_action_041 _1 _4 _endpos__2_ _startpos__1_ name ty x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState567
        | _ ->
            _eRR ()
    
    and _menhir_run_531 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_REPEAT, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.LOOPLIMIT ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Tokens.SEMI_COLON ->
            let _v_0 = _menhir_action_173 () in
            _menhir_run_532 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_312 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LOOPLIMIT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState312 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_532 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_REPEAT, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
            let MenhirCell1_stmt_list (_menhir_stack, _, stmt_list, _) = _menhir_stack in
            let MenhirCell1_REPEAT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _3, loop_limit) = ((), _endpos, (), _v) in
            let _v = _menhir_action_296 _1 _3 _endpos__2_ _startpos__1_ expr loop_limit stmt_list in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_520 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSIF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState520) in
            let _menhir_s = MenhirState521 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | _ ->
            _eRR ()
    
    and _menhir_run_498 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR _menhir_cell0_REPEAT, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_stmt_list (_menhir_stack, _, body, _) = _menhir_stack in
            let MenhirCell0_REPEAT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_RPAR (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _, x, _, _) = _menhir_stack in
            let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_ARROBASE_LOOPLIMIT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _3, cond, _5) = ((), _endpos_0, (), _v, ()) in
            let _v = _menhir_action_297 _1 _3 _5 _endpos__2_ _startpos__1_ body cond x in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState498
        | _ ->
            _eRR ()
    
    and _menhir_run_492 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_call _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, fld, _, _) = _menhir_stack in
            let MenhirCell1_call (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, _4, rhs) = (_endpos_0, (), (), _v) in
            let _v = _menhir_action_285 _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc fld rhs in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState492
        | _ ->
            _eRR ()
    
    and _menhir_run_488 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_call _menhir_cell0_LBRACKET, ttv_result) _menhir_cell1_clist2_IDENTIFIER_ _menhir_cell0_RBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_RBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _, x) = _menhir_stack in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_call (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, _4, rhs) = (_endpos_0, (), (), _v) in
            let _v = _menhir_action_286 _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs x in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | _ ->
            _eRR ()
    
    and _menhir_run_481 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_call as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_call (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, rhs) = (_endpos_0, (), _v) in
            let _v = _menhir_action_284 _2_inlined1 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState481
        | _ ->
            _eRR ()
    
    and _menhir_run_476 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_lexpr as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_lexpr (_menhir_stack, _menhir_s, le, _startpos_le_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, e) = (_endpos_0, (), _v) in
            let _v = _menhir_action_283 _2_inlined1 _endpos__2_ _startpos_le_ e le in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState476
        | _ ->
            _eRR ()
    
    and _menhir_run_470 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_as_ty (_menhir_stack, _, x) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, decl_item) = _menhir_stack in
            let MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, local_decl_keyword_non_var, _startpos_local_decl_keyword_non_var_) = _menhir_stack in
            let (_endpos__2_, _4, x_inlined1) = (_endpos_0, (), _v) in
            let _v = _menhir_action_282 _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x x_inlined1 in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState470
        | _ ->
            _eRR ()
    
    and _menhir_run_466 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_EQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, decl_item) = _menhir_stack in
            let MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, local_decl_keyword_non_var, _startpos_local_decl_keyword_non_var_) = _menhir_stack in
            let (_endpos__2_, x, _4) = (_endpos_0, _v, ()) in
            let _v = _menhir_action_281 _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | _ ->
            _eRR ()
    
    and _menhir_run_455 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR _menhir_cell0_WHILE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.DO ->
            let _menhir_stack = MenhirCell1_DO (_menhir_stack, MenhirState455) in
            let _menhir_s = MenhirState456 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState455
        | _ ->
            _eRR ()
    
    and _menhir_run_452 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_ARROBASE_LOOPLIMIT _menhir_cell0_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.RPAR ->
            let _menhir_s = MenhirState452 in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RPAR (_menhir_stack, _menhir_s, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_WHILE (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState454 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.REAL_LIT _v ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INT_LIT _v ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOL_LIT _v ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITVECTOR_LIT _v ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.REPEAT ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_REPEAT (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState495 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.WHILE ->
                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.VAR ->
                    _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.UNREACHABLE ->
                    _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.TRY ->
                    _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.THROW ->
                    _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.RETURN ->
                    _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REPEAT ->
                    _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINTLN ->
                    _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINT ->
                    _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRAGMA ->
                    _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PASS ->
                    _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LET ->
                    _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IF ->
                    _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.FOR ->
                    _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.DEBUG ->
                    _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CONSTANT ->
                    _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CASE ->
                    _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ASSERT ->
                    _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ARROBASE_LOOPLIMIT ->
                    _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState452
        | _ ->
            _eRR ()
    
    and _menhir_run_448 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_ASSERT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_ASSERT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos_0, _v) in
            let _v = _menhir_action_280 _1 _endpos__2_ _startpos__1_ e in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState448
        | _ ->
            _eRR ()
    
    and _menhir_run_445 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_pattern_list as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_ARROW (_menhir_stack, MenhirState445) in
            let _menhir_s = MenhirState446 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState445
        | _ ->
            _eRR ()
    
    and _menhir_run_440 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_CASE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.OF ->
            let _menhir_stack = MenhirCell1_OF (_menhir_stack, MenhirState440) in
            let _menhir_s = MenhirState441 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHEN ->
                _menhir_run_442 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
        | _ ->
            _eRR ()
    
    and _menhir_run_432 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_direction as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LOOPLIMIT ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.DO ->
            let _v_0 = _menhir_action_173 () in
            _menhir_run_433 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState432 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_433 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_direction, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_loop_limit (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.DO ->
            let _menhir_s = MenhirState434 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_428 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.TO ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_menhir_s, _1) = (MenhirState428, ()) in
            let _v = _menhir_action_046 _1 in
            _menhir_goto_direction _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.DOWNTO ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_menhir_s, _1) = (MenhirState428, ()) in
            let _v = _menhir_action_047 _1 in
            _menhir_goto_direction _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | _ ->
            _eRR ()
    
    and _menhir_goto_direction : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_direction (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v_0 ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState431
        | Tokens.REAL_LIT _v_1 ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState431
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState431
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState431
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState431
        | Tokens.INT_LIT _v_2 ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState431
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState431
        | Tokens.IDENTIFIER _v_3 ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState431
        | Tokens.BOOL_LIT _v_4 ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState431
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState431
        | Tokens.BITVECTOR_LIT _v_5 ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState431
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState431
        | _ ->
            _eRR ()
    
    and _menhir_run_405 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState405) in
            let _menhir_s = MenhirState406 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState405
        | _ ->
            _eRR ()
    
    and _menhir_run_385 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LLBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.RRBRACKET ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RRBRACKET (_menhir_stack, MenhirState385, _endpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.DOT ->
                _menhir_run_387 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState386
            | Tokens.COMMA | Tokens.EQ | Tokens.LBRACKET | Tokens.RPAR ->
                let _v_1 = _menhir_action_201 () in
                _menhir_run_390 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
            | _ ->
                _eRR ())
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState385
        | _ ->
            _eRR ()
    
    and _menhir_run_366 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_RETURN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.SEMI_COLON ->
            let x = _v in
            let _v = _menhir_action_224 x in
            _menhir_goto_option_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_361 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_THROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos_0, _v) in
            let _v = _menhir_action_298 _1 _endpos__2_ _startpos__1_ e in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | _ ->
            _eRR ()
    
    and _menhir_run_334 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState334
        | Tokens.SEMI_COLON ->
            let MenhirCell1_EQ (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, _2) = ((), _v) in
            let _v = _menhir_action_226 _1 _2 in
            _menhir_goto_option_eq_expr _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_option_eq_expr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState337 ->
            _menhir_run_349 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState332 ->
            _menhir_run_335 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_349 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_as_ty (_menhir_stack, _, x) = _menhir_stack in
        let MenhirCell1_decl_item (_menhir_stack, _, ldi) = _menhir_stack in
        let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos__2_, e) = ((), _endpos, _v) in
        let _v = _menhir_action_289 _1 _endpos__2_ _startpos__1_ e ldi x in
        _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_run_335 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_decl_item (_menhir_stack, _, ldi) = _menhir_stack in
        let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos__2_, e) = ((), _endpos, _v) in
        let _v = _menhir_action_288 _1 _endpos__2_ _startpos__1_ e ldi in
        _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_run_313 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LOOPLIMIT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState313
        | Tokens.DO | Tokens.SEMI_COLON ->
            let MenhirCell1_LOOPLIMIT (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, _2) = ((), _v) in
            let _v = _menhir_action_174 _1 _2 in
            _menhir_goto_loop_limit _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_loop_limit : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState531 ->
            _menhir_run_532 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState432 ->
            _menhir_run_433 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState311 ->
            _menhir_run_314 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_314 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHILE, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_loop_limit (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.DO ->
            let _menhir_s = MenhirState315 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_311 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHILE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.LOOPLIMIT ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState311
        | Tokens.DO ->
            let _v_0 = _menhir_action_173 () in
            _menhir_run_314 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState311 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_310 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_NOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState310
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState310
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState310
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState310
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_088 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_305 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.RPAR ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos_0) in
            let _v = _menhir_action_101 _endpos__3_ _startpos__1_ x in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_COMMA (_menhir_stack, MenhirState305) in
            let _menhir_s = MenhirState307 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState305
        | _ ->
            _eRR ()
    
    and _menhir_run_303 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.ELSIF ->
            _menhir_run_139 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.ELSE ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState303
        | _ ->
            _eRR ()
    
    and _menhir_run_139 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState139 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_143 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_ELSE (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState143 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_301 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState301) in
            let _menhir_s = MenhirState302 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState301
        | _ ->
            _eRR ()
    
    and _menhir_run_297 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState297
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState297
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState297
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState297
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_086 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_287 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_SLICING as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState287
        | Tokens.COMMA | Tokens.RBRACE ->
            let MenhirCell1_SLICING (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_163 _2 e1 e2 in
            _menhir_goto_int_constraint _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_int_constraint : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_int_constraint (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState283 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_189 x in
            _menhir_goto_nclist_int_constraint_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _menhir_fail ()
    
    and _menhir_goto_nclist_int_constraint_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState283 ->
            _menhir_run_284 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState020 ->
            _menhir_run_280 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_284 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_int_constraint -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_int_constraint (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_190 _2 h t in
        _menhir_goto_nclist_int_constraint_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_280 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_030 x in
        _menhir_goto_constraint_kind _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_285 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SLICING ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_SLICING (_menhir_stack, MenhirState285) in
            let _menhir_s = MenhirState286 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState285
        | Tokens.COMMA | Tokens.RBRACE ->
            let expr = _v in
            let _v = _menhir_action_162 expr in
            _menhir_goto_int_constraint _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_276 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState276
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_255 _2 e1 e2 in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_slice : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_slice (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState268 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.COLON ->
                _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACKET ->
            let x = _v in
            let _v = _menhir_action_195 x in
            _menhir_goto_nclist_slice_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _menhir_fail ()
    
    and _menhir_goto_nclist_slice_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState032 ->
            _menhir_run_277 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState268 ->
            _menhir_run_269 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_277 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACKET -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACKET (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_261 x in
        _menhir_goto_slices _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_slices : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState399 ->
            _menhir_run_400 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState262 ->
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState185 ->
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState129 ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState110 ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState666 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState653 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState567 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState531 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState405 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState520 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState440 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState445 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState452 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState498 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState455 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState492 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState488 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState481 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState476 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState470 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState466 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState448 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState432 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState428 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState385 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState366 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState361 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState334 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState311 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState313 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState310 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState305 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState301 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState303 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState297 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState285 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState287 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState025 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState027 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState029 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState031 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState270 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState276 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState274 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState272 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState034 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState036 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState257 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState253 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState248 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState246 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState244 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState242 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState240 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState238 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState236 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState229 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState227 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState224 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState222 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState220 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState215 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState212 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState210 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState208 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState206 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState204 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState202 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState199 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState197 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState195 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState193 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState191 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState189 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState187 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState171 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState164 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState162 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState136 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState138 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState140 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState142 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState144 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState056 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState108 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState119 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState102 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState100 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState098 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState093 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState096 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState075 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState091 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState089 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState087 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState079 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState085 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState083 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState081 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState077 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState072 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState062 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState070 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState068 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState066 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState064 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState058 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState060 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState052 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState040 ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_400 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_basic_lexpr -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        let MenhirCell1_basic_lexpr (_menhir_stack, _menhir_s, basic, _startpos_basic_) = _menhir_stack in
        let (_endpos_desc_, _startpos_desc_, desc) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_260 _endpos_desc_ _startpos_desc_ basic desc in
        _menhir_goto_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_basic_ _v _menhir_s _tok
    
    and _menhir_run_254 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_expr_pattern -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
        let (_endpos_slices_, slices) = (_endpos, _v) in
        let _v = _menhir_action_137 _endpos_slices_ _startpos_e_ e slices in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_slices_ _startpos_e_ _v _menhir_s _tok
    
    and _menhir_run_113 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v_0 ->
            let _startpos_1 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_2 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                let _menhir_stack = MenhirCell1_slices (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v_0, _startpos_1, _endpos_2) in
                _menhir_run_110 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState114
            | Tokens.COLON ->
                let _menhir_stack = MenhirCell1_slices (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v_0, _startpos_1, _endpos_2) in
                _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState114
            | Tokens.COMMA | Tokens.RBRACE ->
                let (s, x) = (_v, _v_0) in
                let _v = _menhir_action_006 s x in
                _menhir_goto_bitfield _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_110 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState110 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RBRACE ->
            let _v = _menhir_action_307 () in
            _menhir_goto_tclist_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_goto_tclist_bitfield_ : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_009 x in
        _menhir_goto_bitfields _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
    
    and _menhir_goto_bitfields : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState109 ->
            _menhir_run_132 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState114 ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_132 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_BITS _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_011 _1 in
        _menhir_goto_bitfields_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _tok
    
    and _menhir_goto_bitfields_opt : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_BITS _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_RPAR (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, x, _, _) = _menhir_stack in
        let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_BITS (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_bitfields_opt_, bitfields_opt) = ((), _endpos, _v) in
        let _v = _menhir_action_316 _1 _endpos_bitfields_opt_ _startpos__1_ bitfields_opt x in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_bitfields_opt_ _v _menhir_s _tok
    
    and _menhir_run_125 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_slices _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
        let MenhirCell1_slices (_menhir_stack, _menhir_s, s, _, _) = _menhir_stack in
        let bf = _v in
        let _v = _menhir_action_007 bf s x in
        _menhir_goto_bitfield _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_bitfield : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACKET ->
                let _menhir_stack = MenhirCell1_bitfield (_menhir_stack, _menhir_s, _v) in
                _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
            | Tokens.RBRACE ->
                let (x, x_inlined1) = (_v, ()) in
                let _v = _menhir_action_213 x x_inlined1 in
                _menhir_goto_ntclist_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_212 x in
            _menhir_goto_ntclist_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_ntclist_bitfield_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState129 ->
            _menhir_run_130 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState110 ->
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_130 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_bitfield -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_bitfield (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_214 _2 h t in
        _menhir_goto_ntclist_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_127 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_308 _1 in
        _menhir_goto_tclist_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_049 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
        let (_endpos_slices_, slices) = (_endpos, _v) in
        let _v = _menhir_action_091 _endpos_slices_ _startpos_e_ e slices in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_slices_ _startpos_e_ _v _menhir_s _tok
    
    and _menhir_run_269 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_slice -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_slice (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_196 _2 h t in
        _menhir_goto_nclist_slice_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_274 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_PLUS_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState274
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_PLUS_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_256 _2 e1 e2 in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_272 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_STAR_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState272
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_STAR_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_258 _2 e1 e2 in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_270 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.STAR_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_STAR_COLON (_menhir_stack, MenhirState270) in
            let _menhir_s = MenhirState271 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.PLUS_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_PLUS_COLON (_menhir_stack, MenhirState270) in
            let _menhir_s = MenhirState273 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_COLON (_menhir_stack, MenhirState270) in
            let _menhir_s = MenhirState275 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState270
        | Tokens.COMMA | Tokens.RBRACKET ->
            let expr = _v in
            let _v = _menhir_action_254 expr in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_257 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState257
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState257
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState257
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState257
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_132 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_253 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_AND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState253
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_AND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_107 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_248 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_BAND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState248
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_BAND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_108 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_246 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_BEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState246
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_BEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_110 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_244 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_BOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState244
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_BOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_109 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_242 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_COLON_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_COLON_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_130 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_240 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_CONCAT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_CONCAT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_131 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_238 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_DIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_DIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_111 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_236 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_DIVRM as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_DIVRM (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_112 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_229 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_EOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState229
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_EOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_113 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_227 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_EQ_OP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState227
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_EQ_OP (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_114 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_224 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_GEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState224
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_GEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_117 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_222 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_GT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_GT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_116 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_220 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_IMPL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_IMPL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_118 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_215 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_LEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState215
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_LEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_120 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_212 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_LLBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.RRBRACKET ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LLBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_2, _endpos__4_, _4, e2) = ((), _endpos_0, (), _v) in
            let _v = _menhir_action_138 _2 _4 _endpos__4_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__4_ _startpos_e1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState212
        | _ ->
            _eRR ()
    
    and _menhir_run_210 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_LT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_LT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_119 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_208 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_MINUS (_menhir_stack, _, _, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_122 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_206 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_MOD as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_MOD (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_123 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_204 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_MUL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_MUL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_124 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_202 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_NEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_NEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_115 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_199 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState199
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_OR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_125 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_197 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_PLUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_PLUS (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_121 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_195 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_POW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_POW (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_129 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_193 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_RDIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_RDIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_126 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_191 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_SHL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_SHL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_127 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_189 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_SHR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_SHR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_128 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_187 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_SLICING as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let MenhirCell1_SLICING (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_2, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_232 _2 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_171 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_GEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState171
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_expr_, expr) = ((), _endpos, _v) in
            let _v = _menhir_action_235 _1 _endpos_expr_ _startpos__1_ expr in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_164 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_COMMA (_menhir_stack, MenhirState164) in
            let _menhir_s = MenhirState165 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState164
        | Tokens.RBRACE | Tokens.RPAR | Tokens.SEMI_COLON ->
            let x = _v in
            let _v = _menhir_action_185 x in
            _menhir_goto_nclist_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_expr_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState343 ->
            _menhir_run_344 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState307 ->
            _menhir_run_308 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState635 ->
            _menhir_run_298 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState375 ->
            _menhir_run_298 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_298 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState165 ->
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState408 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState340 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState148 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState151 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_344 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_COMMA as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_nclist_expr_ (_menhir_stack, _menhir_s, _v) in
                let _menhir_stack = MenhirCell0_RBRACE (_menhir_stack, _endpos) in
                _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState345
            | Tokens.SEMI_COLON ->
                let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
                let (_2, _3, _5, _endpos__5_, params) = ((), (), (), _endpos, _v) in
                let _v = _menhir_action_057 _2 _3 _5 name params in
                _menhir_goto_elided_param_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__5_ _startpos_name_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_elided_param_call : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState469 ->
            _menhir_run_472 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState338 ->
            _menhir_run_347 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_472 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_as_ty (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, lhs) = _menhir_stack in
            let MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, ldk, _startpos_ldk_) = _menhir_stack in
            let (_endpos__2_, _endpos_desc_, _startpos_desc_, desc, _4) = (_endpos_0, _endpos, _startpos, _v, ()) in
            let _v = _menhir_action_287 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ _startpos_ldk_ desc ldk lhs ty in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_347 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty, ttv_result) _menhir_cell1_EQ -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_EQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_as_ty (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, lhs) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _endpos_desc_, _startpos_desc_, desc, _4) = ((), _endpos_0, _endpos, _startpos, _v, ()) in
            let _v = _menhir_action_291 _1 _4 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc lhs ty in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_308 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COMMA -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _, x, _, _) = _menhir_stack in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_2, _endpos__3_, li) = ((), _endpos, _v) in
            let _v = _menhir_action_246 _2 li x in
            let (_endpos, _startpos) = (_endpos__3_, _startpos__1_) in
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_104 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_298 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _1 = _v in
        let _v = _menhir_action_022 _1 in
        _menhir_goto_clist_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_clist_expr_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState635 ->
            _menhir_run_636 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState375 ->
            _menhir_run_376 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState010 ->
            _menhir_run_299 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_299 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos) in
            let _v = _menhir_action_242 x in
            _menhir_goto_plist_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_plist_expr_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState435 ->
            _menhir_run_436 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState371 ->
            _menhir_run_372 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState368 ->
            _menhir_run_369 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState345 ->
            _menhir_run_346 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState341 ->
            _menhir_run_342 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState407 ->
            _menhir_run_167 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState339 ->
            _menhir_run_167 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState009 ->
            _menhir_run_167 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState147 ->
            _menhir_run_167 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState150 ->
            _menhir_run_167 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState156 ->
            _menhir_run_157 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_436 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_DEBUG -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_DEBUG (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, args) = ((), _endpos_0, _v) in
            let _v = _menhir_action_294 _1 _endpos__2_ _startpos__1_ args in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_372 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_PRINT -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_PRINT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, args) = ((), _endpos_0, _v) in
            let _v = _menhir_action_293 _1 _endpos__2_ _startpos__1_ args in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_369 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_PRINTLN -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_PRINTLN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, args) = ((), _endpos_0, _v) in
            let _v = _menhir_action_292 _1 _endpos__2_ _startpos__1_ args in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_346 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_COMMA, ttv_result) _menhir_cell1_nclist_expr_ _menhir_cell0_RBRACE -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_RBRACE (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_nclist_expr_ (_menhir_stack, _, params) = _menhir_stack in
        let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
        let (_endpos__1_, _1, _2, _3, _5) = (_endpos, _v, (), (), ()) in
        let _v = _menhir_action_058 _1 _2 _3 _5 name params in
        _menhir_goto_elided_param_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_name_ _v _menhir_s _tok
    
    and _menhir_run_342 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_RBRACE -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_RBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
        let (_2, _3, _endpos_args_, args) = ((), (), _endpos, _v) in
        let _v = _menhir_action_056 _2 _3 args name in
        _menhir_goto_elided_param_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_args_ _startpos_name_ _v _menhir_s _tok
    
    and _menhir_run_167 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
        let (_endpos_args_, args) = (_endpos, _v) in
        let _v = _menhir_action_012 args name in
        _menhir_goto_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_args_ _startpos_name_ _v _menhir_s _tok
    
    and _menhir_goto_call : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState671 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState628 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState315 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState547 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState543 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState406 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState523 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState434 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState506 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState501 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState446 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState495 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState456 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState460 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState442 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState038 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState054 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState259 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState173 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState183 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState665 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState635 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState566 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState001 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState530 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState404 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState519 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState427 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState431 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState439 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState444 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState451 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState497 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState454 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState491 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState487 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState480 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState475 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState469 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState465 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState447 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState408 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState384 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState375 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState359 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState338 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState340 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState343 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState333 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState312 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState004 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState005 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState006 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState307 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState008 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState302 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState012 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState020 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState283 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState286 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState021 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState026 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState028 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState030 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState032 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState275 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState273 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState271 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState033 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState035 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState172 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState252 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState247 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState245 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState243 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState241 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState239 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState237 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState235 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState228 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState225 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState223 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState221 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState219 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState214 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState211 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState209 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState207 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState205 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState203 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState200 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState148 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState151 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState165 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState161 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState137 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState139 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState141 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState143 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState055 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState118 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState101 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState099 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState097 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState092 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState094 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState073 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState090 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState088 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState086 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState078 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState084 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState082 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState080 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState076 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState071 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState061 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState069 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState067 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState065 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState063 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState057 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState059 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState051 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_478 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos__2_, _startpos_call_, call) = (_endpos_0, _startpos, _v) in
            let _v = _menhir_action_279 _endpos__2_ _startpos_call_ call in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_call (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState480 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_call (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACKET ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState484 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_422 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.IDENTIFIER _v ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.EQ ->
                    let _menhir_s = MenhirState491 in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.STRING_LIT _v ->
                        _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.REAL_LIT _v ->
                        _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.NOT ->
                        _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.MINUS ->
                        _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.LPAR ->
                        _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.INT_LIT _v ->
                        _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.IF ->
                        _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.IDENTIFIER _v ->
                        _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.BOOL_LIT _v ->
                        _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.BNOT ->
                        _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.BITVECTOR_LIT _v ->
                        _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.ARBITRARY ->
                        _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_255 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_call_, _startpos_call_, call) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_136 _endpos_call_ _startpos_call_ call in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_call_ _startpos_call_ _v _menhir_s _tok
    
    and _menhir_run_050 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_call_, _startpos_call_, call) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_090 _endpos_call_ _startpos_call_ call in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_call_ _startpos_call_ _v _menhir_s _tok
    
    and _menhir_run_157 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_nclist_expr_ _menhir_cell0_RBRACE -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_RBRACE (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_nclist_expr_ (_menhir_stack, _, x) = _menhir_stack in
        let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
        let (_endpos__1_inlined1_, _1_inlined1) = (_endpos, _v) in
        let _v = _menhir_action_014 _1_inlined1 name x in
        _menhir_goto_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _startpos_name_ _v _menhir_s _tok
    
    and _menhir_run_166 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COMMA -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, h, _, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_186 _2 h t in
        _menhir_goto_nclist_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_155 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_nclist_expr_ (_menhir_stack, _menhir_s, _v) in
                let _menhir_stack = MenhirCell0_RBRACE (_menhir_stack, _endpos) in
                _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState156
            | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
                let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
                let (x, _endpos__3_) = (_v, _endpos) in
                let _v = _menhir_action_013 name x in
                _menhir_goto_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_name_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_162 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
        | Tokens.COMMA | Tokens.RBRACE ->
            let MenhirCell1_EQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
            let y = _v in
            let _v = _menhir_action_150 x y in
            (match (_tok : MenhirBasics.token) with
            | Tokens.COMMA ->
                let _menhir_stack = MenhirCell1_field_assign (_menhir_stack, _menhir_s, _v) in
                let _menhir_s = MenhirState159 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.IDENTIFIER _v ->
                    let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                    let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                    let _menhir_s = MenhirState160 in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.EQ ->
                        _menhir_run_161 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | Tokens.RBRACE ->
                let x = _v in
                let _v = _menhir_action_187 x in
                _menhir_goto_nclist_field_assign_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _menhir_fail ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_field_assign_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState148 ->
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState159 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState340 ->
            _menhir_run_153 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState151 ->
            _menhir_run_153 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_168 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_149 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_run_163 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_field_assign -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_field_assign (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_188 _2 h t in
        _menhir_goto_nclist_field_assign_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_153 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_103 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_run_144 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
        | Tokens.ARROW | Tokens.BEGIN | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_ELSE (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, _endpos__2_, _2) = ((), _endpos, _v) in
            let _v = _menhir_action_054 _1 _2 in
            _menhir_goto_e_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_e_else : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState303 ->
            _menhir_run_304 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState138 ->
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState142 ->
            _menhir_run_145 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_304 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _, e2, _, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e1, _, _) = _menhir_stack in
        let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_e_else_, e_else) = ((), (), _endpos, _v) in
        let _v = _menhir_action_089 _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_else_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_146 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _, e2, _, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e1, _, _) = _menhir_stack in
        let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_e_else_, e_else) = ((), (), _endpos, _v) in
        let _v = _menhir_action_135 _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_else_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_145 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, c, _, _) = _menhir_stack in
        let MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_e_else_, e_else) = ((), (), _endpos, _v) in
        let _v = _menhir_action_055 _1 _3 _endpos_e_else_ _startpos__1_ c e e_else in
        _menhir_goto_e_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_else_ _v _menhir_s _tok
    
    and _menhir_run_142 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.ELSIF ->
            _menhir_run_139 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.ELSE ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState142
        | _ ->
            _eRR ()
    
    and _menhir_run_140 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSIF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState140) in
            let _menhir_s = MenhirState141 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | _ ->
            _eRR ()
    
    and _menhir_run_138 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.ELSIF ->
            _menhir_run_139 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.ELSE ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | _ ->
            _eRR ()
    
    and _menhir_run_136 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState136) in
            let _menhir_s = MenhirState137 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | _ ->
            _eRR ()
    
    and _menhir_run_119 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_ARRAY as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.RRBRACKET ->
            let _menhir_s = MenhirState119 in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RRBRACKET (_menhir_stack, _menhir_s, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.OF ->
                let _menhir_s = MenhirState121 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING ->
                    _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REAL ->
                    _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INTEGER ->
                    _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOLEAN ->
                    _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITS ->
                    _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BIT ->
                    _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ARRAY ->
                    _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState119
        | _ ->
            _eRR ()
    
    and _menhir_run_108 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BITS _menhir_cell0_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.SHL ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.RPAR ->
            let _menhir_s = MenhirState108 in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RPAR (_menhir_stack, _menhir_s, _endpos) in
            let _menhir_s = MenhirState109 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                _menhir_run_110 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RECURSELIMIT | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.SUBTYPES | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
                let _v = _menhir_action_010 () in
                _menhir_goto_bitfields_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
            | _ ->
                _eRR ())
        | Tokens.RDIV ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.POW ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.PLUS ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.OR ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.NEQ ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.MUL ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.MOD ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.MINUS ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.LT ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.LLBRACKET ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.LEQ ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.LBRACKET ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.IN ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.IMPL ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.GT ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.GEQ ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.EQ_OP ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.EOR ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.DOT ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.DIVRM ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.DIV ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.CONCAT ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.COLON_COLON ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.BOR ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.BEQ ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.BAND ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.AS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | Tokens.AND ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState108
        | _ ->
            _eRR ()
    
    and _menhir_run_102 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_BAND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_BAND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_062 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_100 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_BEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState100
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_BEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_064 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_098 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_BOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_BOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_063 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_096 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_EQ_OP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState096
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_EQ_OP (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_068 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_093 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_IMPL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_IMPL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_072 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_091 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_GEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_GEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_071 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_089 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_GT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState089
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_GT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_070 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_087 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState087
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_LEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_074 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_085 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_AND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState085
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_AND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_061 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_083 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COLON_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState083
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_COLON_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_084 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_081 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_EOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState081
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_EOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_067 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_079 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState079
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_LT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_073 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_077 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState077
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_MINUS (_menhir_stack, _, _, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_076 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_075 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_NEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState075
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_NEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_069 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_072 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_OR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_079 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_070 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_DIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState070
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState070
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState070
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState070
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState070
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState070
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_DIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_065 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_068 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_DIVRM as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_DIVRM (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_066 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_066 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_MOD as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_MOD (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_077 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_064 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_MUL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState064
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState064
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState064
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState064
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState064
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState064
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_MUL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_078 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_062 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_PLUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState062
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_PLUS (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_075 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_060 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_CONCAT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState060
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState060
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState060
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState060
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_CONCAT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_085 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_058 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RDIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState058
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState058
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState058
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState058
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState058
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState058
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_RDIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_080 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_056 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState056
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_expr_, expr) = ((), _endpos, _v) in
            let _v = _menhir_action_234 _1 _endpos_expr_ _startpos__1_ expr in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_052 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos__1_, _) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_133 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_040 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_NOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState040
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState040
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState040
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState040
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_134 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_036 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_SHL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState036
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState036
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState036
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState036
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState036
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState036
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_SHL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_081 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_034 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState034
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_COLON (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, e) = ((), _v) in
            let _v = _menhir_action_257 _1 e in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_031 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_POW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState031
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState031
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState031
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState031
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_POW (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_083 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_029 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_SHR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState029
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState029
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState029
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState029
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState029
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState029
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_SHR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_082 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_027 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LLBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.RRBRACKET ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LLBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_2, _endpos__4_, _4, e2) = ((), _endpos_0, (), _v) in
            let _v = _menhir_action_092 _2 _4 _endpos__4_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__4_ _startpos_e1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState027
        | _ ->
            _eRR ()
    
    and _menhir_run_025 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState025
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState025
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState025
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState025
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos__1_, _) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_087 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist_ty_ : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAR -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_243 x in
        let (_endpos, _startpos) = (_endpos__3_, _startpos__1_) in
        let (_endpos_l_, _startpos_l_, l) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_317 _endpos_l_ _startpos_l_ l in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_l_ _v _menhir_s _tok
    
    and _menhir_run_332 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_decl_item (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_EQ (_menhir_stack, MenhirState332) in
            let _menhir_s = MenhirState333 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState332
        | Tokens.SEMI_COLON ->
            let _v_6 = _menhir_action_225 () in
            _menhir_run_335 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | _ ->
            _eRR ()
    
    and _menhir_run_324 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_discard_or_identifier (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState325 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_317 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_319 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_181 x in
            _menhir_goto_nclist_discard_or_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_discard_or_identifier_ : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState325 ->
            _menhir_run_326 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState321 ->
            _menhir_run_322 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_326 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_identifier, ttv_result) _menhir_cell1_discard_or_identifier -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_182 _2 h t in
        _menhir_goto_nclist_discard_or_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_322 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_discard_or_identifier -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_discard_or_identifier (_menhir_stack, _, x) = _menhir_stack in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_245 _2 li x in
        let _1 = _v in
        let _v = _menhir_action_045 _1 in
        _menhir_goto_decl_item _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_320 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_discard_or_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState321 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_317 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_319 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_645 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_GETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_return_type (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            _menhir_run_628 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState645
        | _ ->
            _eRR ()
    
    and _menhir_run_603 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _1 = _v in
        let _v = _menhir_action_320 _1 in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_539 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ARROW ->
            let _menhir_s = MenhirState540 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARROBASE_LOOPLIMIT ->
                _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_296 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_ARBITRARY -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _2, _endpos_ty_, ty) = ((), (), _endpos, _v) in
        let _v = _menhir_action_100 _1 _2 _endpos_ty_ _startpos__1_ ty in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_290 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
            let _menhir_s = MenhirState291 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_197 x in
            _menhir_goto_nclist_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_ty_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState018 ->
            _menhir_run_293 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState291 ->
            _menhir_run_292 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_293 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAR -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_024 _1 in
        _menhir_goto_clist_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_292 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_ty -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_ty (_menhir_stack, _menhir_s, h, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_198 _2 h t in
        _menhir_goto_nclist_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_250 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
        let (_2, _endpos_ty_, ty) = ((), _endpos, _v) in
        let _v = _menhir_action_141 _2 _endpos_ty_ _startpos_expr_pattern_ expr_pattern ty in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos_expr_pattern_ _v _menhir_s _tok
    
    and _menhir_run_176 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_ARBITRARY -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _2, _endpos_ty_, ty) = ((), (), _endpos, _v) in
        let _v = _menhir_action_146 _1 _2 _endpos_ty_ _startpos__1_ ty in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_133 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
        let (_2, _endpos_ty_, ty) = ((), _endpos, _v) in
        let _v = _menhir_action_095 _2 _endpos_ty_ _startpos_expr_ expr ty in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos_expr_ _v _menhir_s _tok
    
    and _menhir_run_124 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_COLON -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_COLON (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, _2) = ((), _v) in
        let _v = _menhir_action_003 _1 _2 in
        _menhir_goto_as_ty _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_as_ty : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState663 ->
            _menhir_run_668 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState609 ->
            _menhir_run_610 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState577 ->
            _menhir_run_578 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState564 ->
            _menhir_run_569 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState464 ->
            _menhir_run_468 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState351 ->
            _menhir_run_352 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState332 ->
            _menhir_run_337 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState114 ->
            _menhir_run_126 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_668 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_global_decl_keyword_non_var, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let x = _v in
        let _v = _menhir_action_222 x in
        _menhir_goto_option_as_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_option_as_ty_ : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState663 ->
            _menhir_run_664 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState564 ->
            _menhir_run_565 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_610 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let x_inlined1 = _v in
        let _v = _menhir_action_220 x x_inlined1 in
        _menhir_goto_opt_typed_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_578 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let y = _v in
        let _v = _menhir_action_324 x y in
        _menhir_goto_typed_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_typed_identifier : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState626 ->
            _menhir_run_627 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState619 ->
            _menhir_run_618 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState617 ->
            _menhir_run_618 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState580 ->
            _menhir_run_579 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState576 ->
            _menhir_run_579 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_627 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            _menhir_run_628 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState627
        | _ ->
            _eRR ()
    
    and _menhir_run_618 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState619 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_577 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_199 x in
            _menhir_goto_nclist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_nclist_typed_identifier_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState617 ->
            _menhir_run_621 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState619 ->
            _menhir_run_620 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_621 : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_LPAR -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_026 _1 in
        _menhir_goto_clist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_620 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_typed_identifier -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_200 _2 h t in
        _menhir_goto_nclist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_579 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v_0 ->
                let _menhir_stack = MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, _v) in
                _menhir_run_577 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState580
            | Tokens.RBRACE ->
                let (x, x_inlined1) = (_v, ()) in
                let _v = _menhir_action_216 x x_inlined1 in
                _menhir_goto_ntclist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_215 x in
            _menhir_goto_ntclist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_ntclist_typed_identifier_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState576 ->
            _menhir_run_584 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState580 ->
            _menhir_run_581 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_584 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_310 _1 in
        _menhir_goto_tclist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_581 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_typed_identifier -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_217 _2 h t in
        _menhir_goto_ntclist_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_569 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_ignored_or_identifier (_menhir_stack, _, name) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, x) = ((), _endpos, _v) in
            let _v = _menhir_action_042 _1 _endpos__2_ _startpos__1_ name x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.EQ ->
            let x = _v in
            let _v = _menhir_action_222 x in
            _menhir_goto_option_as_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_468 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_as_ty (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState469 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_339 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_339 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState339
        | Tokens.LBRACE ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState340
            | Tokens.REAL_LIT _v_2 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState340
            | Tokens.RBRACE ->
                let _endpos_3 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LPAR ->
                    let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                    let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                    let _menhir_stack = MenhirCell1_RBRACE (_menhir_stack, MenhirState340, _endpos_3) in
                    _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState341
                | Tokens.AND | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RDIV | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR ->
                    let (_2, _endpos__3_, _3, _endpos_desc_, _startpos_desc_, desc) = ((), _endpos_3, (), _endpos, _startpos, _v) in
                    let _v = _menhir_action_102 _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc in
                    _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
                | _ ->
                    _eRR ())
            | Tokens.NOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState340
            | Tokens.MINUS ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState340
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState340
            | Tokens.INT_LIT _v_4 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState340
            | Tokens.IF ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState340
            | Tokens.IDENTIFIER _v_5 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState340
            | Tokens.COMMA ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                let _menhir_stack = MenhirCell1_COMMA (_menhir_stack, MenhirState340) in
                let _menhir_s = MenhirState343 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.REAL_LIT _v ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INT_LIT _v ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOL_LIT _v ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITVECTOR_LIT _v ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.BOOL_LIT _v_12 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12 MenhirState340
            | Tokens.BNOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState340
            | Tokens.BITVECTOR_LIT _v_13 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13 MenhirState340
            | Tokens.ARBITRARY ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState339, _startpos_0) in
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState340
            | _ ->
                _eRR ())
        | Tokens.AND | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.CONCAT | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EOR | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RDIV | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_060 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_352 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_clist2_IDENTIFIER_ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _, _2_inlined1) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, as_ty) = ((), _endpos, _v) in
            let _v = _menhir_action_290 _1 _2_inlined1 _endpos__2_ _startpos__1_ as_ty in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_337 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_as_ty (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_EQ (_menhir_stack, MenhirState337) in
            let _menhir_s = MenhirState338 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_339 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SEMI_COLON ->
            let _v_6 = _menhir_action_225 () in
            _menhir_run_349 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | _ ->
            _eRR ()
    
    and _menhir_run_126 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_slices _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
        let MenhirCell1_slices (_menhir_stack, _menhir_s, s, _, _) = _menhir_stack in
        let ty = _v in
        let _v = _menhir_action_008 s ty x in
        _menhir_goto_bitfield _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_122 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_ARRAY, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RRBRACKET -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_RRBRACKET (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
        let MenhirCell1_ARRAY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _2, _4, _endpos_t_, t, _5) = ((), (), (), _endpos, _v, ()) in
        let _v = _menhir_action_319 _1 _2 _4 _5 _endpos_t_ _startpos__1_ e t in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _v _menhir_s _tok
    
    and _menhir_run_188 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState188 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_190 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState190 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_192 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_RDIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState192 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_194 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_POW (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState194 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_196 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_PLUS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState196 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_198 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_OR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState198 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_200 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState200
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState200
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_145 _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_e_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState200
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState200
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState200
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState200
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | _ ->
            _eRR ()
    
    and _menhir_run_203 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MUL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState203 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_205 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MOD (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState205 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_207 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState207 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_209 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState209 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_211 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LLBRACKET (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState211 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_214 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState214 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_216 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IN (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState216 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BNOT ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_219 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IMPL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState219 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_221 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_GT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState221 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_223 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState223 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_225 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState225
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState225
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_144 _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_pattern_ expr_pattern in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_expr_pattern_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState225
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState225
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState225
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState225
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | _ ->
            _eRR ()
    
    and _menhir_run_228 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_EOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState228 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_230 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_DOT (_menhir_stack, _menhir_s) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState231 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.IDENTIFIER _v ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, _endpos_x_, x) = ((), _endpos, _v) in
            let _v = _menhir_action_139 _2 _endpos_x_ _startpos_e_ e x in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_235 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIVRM (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState235 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_237 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState237 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_239 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_CONCAT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState239 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_241 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState241 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_243 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState243 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_245 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BEQ (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState245 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_247 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BAND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState247 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_249 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState249 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_252 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState252 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_185 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SLICING ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_203 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_205 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_207 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_209 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_211 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_214 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_216 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_219 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_221 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_223 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_225 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.EOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_235 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_237 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.CONCAT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_239 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_241 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_243 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_245 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_247 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_249 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let (_endpos_expr_pattern_, _startpos_expr_pattern_, expr_pattern) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_231 _endpos_expr_pattern_ _startpos_expr_pattern_ expr_pattern in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_pattern_ _startpos_expr_pattern_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_023 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_value_, _startpos_value_, value) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_059 _endpos_value_ _startpos_value_ value in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_value_ _startpos_value_ _v _menhir_s _tok
    
    let _menhir_run_000 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_opn =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState000 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARROBASE_LOOPLIMIT ->
            _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    let _menhir_run_560 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.VAR ->
            _menhir_run_561 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.TYPE ->
            _menhir_run_571 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.SETTER ->
            _menhir_run_606 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.PRAGMA ->
            _menhir_run_634 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.LET ->
            _menhir_run_638 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.GETTER ->
            _menhir_run_639 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.FUNC ->
            _menhir_run_647 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.CONSTANT ->
            _menhir_run_657 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.CONFIG ->
            _menhir_run_658 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState560
        | Tokens.EOF ->
            let _v = _menhir_action_169 () in
            _menhir_run_660 _menhir_stack _v
        | _ ->
            _eRR ()
    
    let _menhir_run_671 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_stmts =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState671 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_374 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_425 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_438 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_447 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARROBASE_LOOPLIMIT ->
            _menhir_run_450 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
  end
  
  let stmts =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_stmts v = _menhir_run_671 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let spec =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_spec v = _menhir_run_560 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let opn =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_opn v = _menhir_run_000 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
end
